/*
	MPI wrappers which measure time taken.
	Version 12

	Used for analyzing performance of an MPI code.

	Contents
		MPI_Init		C and Fortran
		MPI_Finalize		C and Fortran

		MPI_Comm_rank		C
		MPI_Get_count		C
		MPI_Ibsend		C
		MPI_Probe		C
		MPI_Recv		C and Fortran   (with detail stats)
		MPI_Irecv		C and Fortran
		MPI_Send		C and Fortran
		MPI_Isend		C and Fortran  
		MPI_Ssend		C and Fortran
		MPI_Wait		C and Fortran   (with detail stats)
		MPI_Waitall		C and Fortran   (with detail stats)

		MPI_Barrier		C and Fortran
		MPI_Alltoall		C and Fortran
		MPI_Alltoallv		C and Fortran
		MPI_Reduce		C and Fortran
		MPI_Allreduce		C and Fortran
		MPI_Bcast		C and Fortran
		MPI_Gather		C and Fortran
		MPI_Gatherv		C and Fortran
		MPI_Scatter		C and Fortran
		MPI_Scatterv		C and Fortran

	Version History:
                o version 1-7 from Karl Feind.
		o version 8 from Frank Kampe added MPI_Isend,
		  MPI_Irecv, and MPI_Waitall.
		o version 9 fixes the MPI_Scatter wrapper
		o version 9b fixes the mpi_gatherv_ wrapper
		o version 9c fixes the mpi_gatherv_ wrapper
                o version 10 added size-distribution collection,
                  MPI_SGI_profiling_enable()/MPI_PROFILE_AT_INIT,
		  communication-time % measurement, type-size determination.
		o version 11 added peer-to-peer communication size measurement
                  and calculation of load imbalance.
		o version 12 added MPI_SGI_loadbarrier() code.  Uses -lshm.
                  Not available on Linux until shm ported.    Undefine
		  LOADBARRIER to build on Linux.
		o version 13 adds the ability to verify that single-copy
		  was used on some calls.   You need to have
		  environment variable MPI_BUFFER_MAX set, and either
		  call MPI_SGI_scopy_check_enable() or set environment
		  variable MPI_SCOPY_CHECK_AT_INIT.  Once single-copy
		  checking is enabled, the first MPI_SCOPY_FAILURES_MAX
		  (or 10, if that environment variable isn't set)
		  failures to use single-copy in calls to routines
		  MPI_Send(), MPI_Sendrecv(), and MPI_Isend() in each
		  rank will reported.  On IRIX, a stack traceback is
		  displayed at each failure.  Fixed a bug in reporting
		  average datasize in some routines.

 */

/*  Define LOADBARRIER if you want the MPI_SGI_loadbarrier() code included.
 */
#define LOADBARRIER

/*  But you cannot have loadbarrier for linux, as it requires libsma, which
 *  hasn't been ported yet. 
 */
#ifdef __linux
#undef LOADBARRIER
#endif

#include <malloc.h>
#include <assert.h>
#include <limits.h>
#include <mpi.h>
/*rv*/
#include <mpi_ext.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#ifdef LOADBARRIER
#include <mpp/shmem.h>
static void MPI_SGI_loadbarrier_stats(FILE *f);
#endif

/*
 * CONSTANTS
 */

#define MAXRANKS 1024		/* must be set >= total number of MPI	*/
				/* processes.				*/

/*
 * UTILITIES
 */

#define TIMEIT_NORETURN(func,  var)    {  \
	double t0, t1;                    \
	t0 = MPI_Wtime();                 \
 	rv = func;                        \
	t1=MPI_Wtime();                   \
	var += t1-t0;                     \
	mpi_sum += t1-t0;                 \
}


#define TIMEIT(func,  var)    {     \
	int rv;                     \
	TIMEIT_NORETURN(func,  var) \
	return rv;                  \
}

/*
 * DATA
 */

                     /* MPI_Wtime values when: */
static double tinit;       /* MPI_Init() called */
static double tstart=-1;   /* MPI_SGI_profiling_enable() called */
static double tend;        /* MPI_Finalize() called */

static double mpi_sum=0.0; /* sum of time taken by all profiled routines */

static int enabled=0;     /* should we collect statistics?   Cleared by MPI_Init,
                         * set by mpi_profiling_enable().  If the environment
			 * variable MPI_PROFILE_AT_INIT is set, enabled 
			 * during MPI_Init().
                         */

/*  Defines for making distribution of size collection easier */
#define INFO_SIZE_LIMIT 1000

#define dist_declare(class)                                                  \
	static long class ## _datacnt;                                       \
	static long class ## _min_datacnt = INT_MAX;                         \
	static long class ## _max_datacnt = INT_MIN;                         \
	struct class ## _info_type {                                         \
		long int info_count;                                         \
		     int info_instances;                                     \
		     int info_peer;                                          \
	};                                                                   \
	static struct class ## _info_type class ## _info [INFO_SIZE_LIMIT];  \
	static int class ## _info_index = 0;

#define dist_collect( bytes, peer, class )                                   \
	{                                                                    \
		int __dci, __dclimit;                                        \
		                                                             \
		class ## _datacnt += (bytes);                                \
		if( (bytes) < (class ## _min_datacnt) ) {                    \
			class ## _min_datacnt = (bytes);                     \
		}                                                            \
		if( (bytes) > (class ## _max_datacnt) ) {                    \
			class ## _max_datacnt = (bytes);                     \
		}                                                            \
		                                                             \
		__dclimit = (class ## _info_index) < INFO_SIZE_LIMIT ?       \
			(class ## _info_index) : INFO_SIZE_LIMIT;            \
                                                                             \
		for(__dci=0; __dci < __dclimit; __dci++) {                   \
			if( (class ## _info)[__dci].info_count == (bytes) && \
			    (class ## _info)[__dci].info_peer == (peer) ) {  \
				(class ## _info)[__dci].info_instances ++;   \
				break;                                       \
			}                                                    \
		}                                                            \
		                                                             \
		if( __dci == __dclimit ) {                                   \
			if( (class ## _info_index) < INFO_SIZE_LIMIT )  {    \
				(class ## _info)[__dci].info_count = (bytes);\
				(class ## _info)[__dci].info_instances = 1;  \
				(class ## _info)[__dci].info_peer = peer;    \
			}                                                    \
			(class ## _info_index)++;                            \
		}                                                            \
	}
	
#define dist_print( cnt, class )                                             \
    if( cnt) {                                                               \
	int __dpi, __dpj, __dclimit;                                         \
	int peers[INFO_SIZE_LIMIT];                                          \
        int __peerscount = 0;                                                \
							                     \
	__dclimit = (class ## _info_index) < INFO_SIZE_LIMIT ?               \
		(class ## _info_index) : INFO_SIZE_LIMIT;                    \
							                     \
        if( cnt ) fprintf(f,                                                 \
"            Average data size %g  (min %ld, max %ld)   size:count(peer)",   \
	     (float)class ## _datacnt / (float)cnt,                          \
	     class ## _min_datacnt, class ## _max_datacnt);                  \
							                     \
	if( class ## _info_index > INFO_SIZE_LIMIT ) fprintf(f,              \
"\n            (%d entries omitted)",                                        \
                 class ## _info_index - INFO_SIZE_LIMIT);                    \
							                     \
	for(__dpi=0; __dpi < __dclimit; __dpi++ ) {                          \
		for( __dpj=__dpi+1; __dpj < __dclimit; __dpj++ ) {           \
			if( (class ## _info)[__dpi].info_count <             \
			    (class ## _info)[__dpj].info_count ) {           \
				struct class ## _info_type tmp;              \
				tmp = (class ## _info)[__dpj];               \
				(class ## _info)[__dpj] =                    \
                                                 (class ## _info)[__dpi];    \
				(class ## _info)[__dpi] = tmp;               \
			}                                                    \
		}                                                            \
	}                                                                    \
				                                             \
	for(__dpi=0; __dpi < __dclimit; __dpi++ ) {                          \
		if( (__dpi%4) == 0 ) {                                       \
			fprintf(f, "\n          ");                          \
		}                                                            \
		fprintf(f,"  %5ld:%4dx(%d)",                                 \
		        (class ## _info)[__dpi].info_count,                  \
			(class ## _info)[__dpi].info_instances,              \
			(class ## _info)[__dpi].info_peer);                  \
	}                                                                    \
	if( __dclimit%4 ) fprintf(f, "\n");                                  \
	fprintf(f, "\n            unique peers:");                           \
							                     \
	for(__dpi=0; __dpi < __dclimit; __dpi++ ) {                          \
		for(__dpj=0; __dpj < __peerscount; __dpj++ ) {               \
			if( peers[__dpj]==(class##_info)[__dpi].info_peer) { \
				goto skip ## class;                          \
			}                                                    \
		}                                                            \
		peers[__peerscount++] = (class ## _info) [__dpi].info_peer;  \
                skip ## class: ;                                             \
	}                                                                    \
							                     \
	for(__dpi=0; __dpi < __peerscount; __dpi++ ) {                       \
	        if( __dpi%12 == 0 && __dpi ) {                               \
			fprintf(f, "\n                      ");              \
		}                                                            \
		fprintf(f,"  %4d", peers[__dpi]);                            \
	}                                                                    \
							                     \
	if( __peerscount%12 ) fprintf(f, "\n");                              \
	fprintf(f, "\n");                                                    \
    }


/*reiner---------------------------------------------------------------------*/

static int scopy_buffer_max=0;
static int scopy_check_enabled=0;
static int scopy_check_overflowed=0;
static int scopy_times_checked=0;
static int scopy_failures_detected=0;
static int scopy_failures_max;
static int scopy_len;
static __uint64_t stat_gen;
static __uint64_t stat_value_after;
static __uint64_t stat_value_before;
static unsigned int stats_num=1;
static unsigned int stats_types=MPI_SGI_STATS_SENDS_PT2PT_SINGLE_COPY;

int
MPI_SGI_scopy_check_start(int count, MPI_Datatype type)
{
	int error;

	scopy_len=0;
	error= MPI_Type_size(type, &scopy_len);
	scopy_len=count*scopy_len;

	if ( scopy_len > scopy_buffer_max ) {
#ifndef MPICH_NAME
		MPI_SGI_stat_get(stats_num,&stats_types,&stat_value_before,
			&stat_gen);
#endif
	}
	return(error);
}

#ifndef __linux
#include <libexc.h>
#endif

int
MPI_SGI_scopy_check_end(char *tag)
{
	int error=0;

	if ( scopy_len > scopy_buffer_max ) {
		scopy_times_checked ++;
#ifndef MPICH_NAME
		MPI_SGI_stat_get(stats_num,&stats_types,&stat_value_after,
				&stat_gen);
#endif
		if ( stat_value_after == stat_value_before ) {
#ifdef __linux
			/*  Don't have trace_back_stack_and_print() on linux... */
			printf("Single-copy wasn't used for call to %s\n", tag);
#else
			error=trace_back_stack_and_print();
#endif
			scopy_failures_detected++;
			if( scopy_failures_detected >= scopy_failures_max ) {
				scopy_check_overflowed = 1;
				scopy_check_enabled = 0;
			}
		}
	}
	return(error);
}

static void
scopy_display(FILE *f)
{
	if( scopy_check_overflowed ) {
		fprintf( f,
"           Single-copy checking was turned off after %d failed checks\n", scopy_failures_max );
	}
	if( scopy_check_overflowed || scopy_check_enabled ) {
		fprintf( f,
"           Single copy was not used %d of %d times checked\n", scopy_failures_detected, scopy_times_checked);
	} else {
		fprintf( f, "           Single-copy checking was not enabled.\n");
	}
}
	
/*reiner--------------------------------------------------------------------*/


/* ------------------------------------------------------------------------ */
/* MPI_Comm_rank */

static long comm_rank_cnt;
static double comm_rank_time;

int
MPI_Comm_rank(MPI_Comm comm, int *rank)
{
	if( enabled ) {
		comm_rank_cnt++;
		TIMEIT( PMPI_Comm_rank(comm, rank), comm_rank_time );
	} else {
		return PMPI_Comm_rank(comm, rank);
	}
}

/* ------------------------------------------------------------------------ */
/* MPI_Get_count */

static long get_count_cnt;
static double get_count_time;

int
MPI_Get_count(MPI_Status *status, MPI_Datatype type, int *count)
{
	if( enabled ) {
		get_count_cnt++;
		TIMEIT( PMPI_Get_count(status, type, count), get_count_time);
	} else {
		return PMPI_Get_count(status, type, count);
	}
}


/* ------------------------------------------------------------------------ */
/* MPI_Ibsend */

static long ibsend_cnt, ibsend_cnts[MAXRANKS];
static double ibsend_time, ibsend_times[MAXRANKS];
dist_declare(ibsend)

int
MPI_Ibsend(void *buf, int count, MPI_Datatype type, int des, int tag,
	MPI_Comm comm, MPI_Request *request)
{
	int rv;
	double t0, t1;

	if( enabled ) {
		int bytes;

		ibsend_cnt++;
		if (comm == MPI_COMM_WORLD) {
			ibsend_cnts[des]++;
		}
	
		t0 = MPI_Wtime(); 
		rv = PMPI_Ibsend(buf, count, type, des, tag, comm, request);
		t1 = MPI_Wtime(); 
		ibsend_time += t1-t0;
		mpi_sum += t1-t0;
		if (comm == MPI_COMM_WORLD) {
			ibsend_times[des] += t1-t0;
		}
	
		MPI_Type_size( type, &bytes );
		bytes *= count;
		dist_collect( bytes, des, ibsend );

		return rv;
	} else {
		return PMPI_Ibsend(buf, count, type, des, tag, comm, request);
	}
}

/* ------------------------------------------------------------------------ */
/* MPI_Probe */

static long probe_cnt, probe_cnts_as, probe_cnts[MAXRANKS];
static double probe_time, probe_times_as, probe_times[MAXRANKS];

int
MPI_Probe(int source, int tag, MPI_Comm comm, MPI_Status *status)
{
	int rv;
	double t0, t1;

	if( enabled ) {
		probe_cnt++;
	
		if (source == MPI_ANY_SOURCE) {
			probe_cnts_as++;
		} else if (comm == MPI_COMM_WORLD) {
			probe_cnts[source]++;
		}
	
		t0 = MPI_Wtime(); 
		rv = PMPI_Probe(source, tag, comm, status);
		t1 = MPI_Wtime(); 
		probe_time += t1-t0;
		mpi_sum += t1-t0;
		if (source == MPI_ANY_SOURCE) {
			probe_times_as += t1-t0;
		} else if (comm == MPI_COMM_WORLD) {
			probe_times[source] += t1-t0;
		}
		return rv;
	} else {
		return PMPI_Probe(source, tag, comm, status);
	}
}

/* ------------------------------------------------------------------------ */
/* MPI_Recv */

static long recv_cnt, recv_cnts_as, recv_cnts[MAXRANKS];
static double recv_time, recv_times_as, recv_times[MAXRANKS];

static long recv_nomsg_cnt, recv_nomsg_cnts[MAXRANKS];
static double recv_nomsg_time;
static double recv_nomsg_times[MAXRANKS];
dist_declare(recv)

int
MPI_Recv(void *buf, int count, MPI_Datatype type, int src, int tag, 
	MPI_Comm comm, MPI_Status *status)

{
	int rv, flag;
	double t0, twait, t1;

	if( enabled ) {
		rv = PMPI_Iprobe(src, tag, comm, &flag, status);
	
		t0 = MPI_Wtime(); 
		twait = t0;
		if (!flag) {
			while (!flag) {
				PMPI_Iprobe(src, tag, comm, &flag, status);
			}
			twait = MPI_Wtime(); 
		}
	
		rv = PMPI_Recv(buf, count, type, src, tag, comm, status);
		t1 = MPI_Wtime(); 
	
		recv_cnt++;
		recv_time += t1-t0;
		mpi_sum += t1-t0;
		if (comm == MPI_COMM_WORLD) {
			if (src == MPI_ANY_SOURCE) 
				src = status->MPI_SOURCE;
			recv_cnts[src]++;
			recv_times[src] += t1-t0;
	
			if (twait > t0) {
				recv_nomsg_cnt ++;
				recv_nomsg_cnts[src] ++;
				recv_nomsg_time += twait-t0;
				recv_nomsg_times[src] += twait-t0;
			}
		}
	
#ifdef MPICH_NAME
		dist_collect( (status->count), src, recv );
#else
		dist_collect( (status->size), src, recv );
#endif
	
		return rv;
	} else {
		return PMPI_Recv(buf, count, type, src, tag, comm, status);
	}
}

void
mpi_recv_(void *buf, int *count, MPI_Datatype *type, int *src, int *tag,
        MPI_Comm *comm, MPI_Status *status, int *ierr)
{
	*ierr = MPI_Recv(buf, *count, *type, *src, *tag, *comm, status);
}


/* ------------------------------------------------------------------------ */
/* MPI_Irecv */

static long irecv_cnt, irecv_cnts[MAXRANKS];
static double irecv_time, irecv_times[MAXRANKS];
dist_declare(irecv)

int
MPI_Irecv(void *buf, int count, MPI_Datatype type, int src, int tag, 
	MPI_Comm comm, MPI_Request *req)

{
	int rv, flag;
	double t0, t1;
        MPI_Status status;

	if( enabled ) { 
		rv = PMPI_Iprobe(src, tag, comm, &flag, &status);
	
		t0 = MPI_Wtime(); 
		
		rv = PMPI_Irecv(buf, count, type, src, tag, comm, req);
		t1 = MPI_Wtime(); 
	
		irecv_cnt++;
		irecv_time += t1-t0;
		mpi_sum += t1-t0;
		if ((comm == MPI_COMM_WORLD) && flag) {
			if (src == MPI_ANY_SOURCE) 
				src = status.MPI_SOURCE;
			irecv_cnts[src]++;
			irecv_times[src] += t1-t0;
	
		}
	
		if( flag ) {
#ifdef MPICH_NAME
			dist_collect( (status.count), src, irecv );
#else
			dist_collect( (status.size), src, irecv );
#endif
		} else { 
			int bytes;

			MPI_Type_size( type, &bytes );
			bytes *= count;
			dist_collect(bytes, src, irecv);
		}
	
		return rv;
	} else {
		return PMPI_Irecv(buf, count, type, src, tag, comm, req);
	}
}

void
mpi_irecv_(void *buf, int *count, MPI_Datatype *type, int *src, int *tag,
        MPI_Comm *comm, MPI_Request *req, int *ierr)
{
	*ierr = MPI_Irecv(buf, *count, *type, *src, *tag, *comm, req);
}


/* ------------------------------------------------------------------------ */
/* MPI_Send */

static long send_cnt, send_cnts[MAXRANKS];
static double send_time, send_times[MAXRANKS];
dist_declare(send)

int
MPI_Send(void *buf, int count, MPI_Datatype type, int des, int tag,
	MPI_Comm comm)
{
	int rv;
	int bytes;
	double t0, t1;

	if( enabled ) {
		send_cnt++;
		if (comm == MPI_COMM_WORLD) {
			send_cnts[des]++;
		}
	
                if(scopy_check_enabled) MPI_SGI_scopy_check_start(count,type);
		t0 = MPI_Wtime(); 
		rv = PMPI_Send(buf, count, type, des, tag, comm);
		t1 = MPI_Wtime(); 
	        if(scopy_check_enabled) MPI_SGI_scopy_check_end("MPI_Send");
		send_time += t1-t0;
		mpi_sum += t1-t0;
		if (comm == MPI_COMM_WORLD) {
			send_times[des] += t1-t0;
		}
	
		MPI_Type_size( type, &bytes );
		bytes *= count;
		dist_collect( bytes, des, send );
	
		return rv;
	} else {
		return PMPI_Send(buf, count, type, des, tag, comm);
	}
}

void
mpi_send_(void *buf, int *count, MPI_Datatype *type, int *des, int *tag,
        MPI_Comm *comm, int *ierr)
{
	*ierr = MPI_Send(buf, *count, *type, *des, *tag, *comm);
}

/* reiner------------------------------------------------------------------ */
/* MPI_Sendrecv */

int
MPI_Sendrecv(void *sbuf, int scount, MPI_Datatype stype, int des, int stag,
	void *rbuf,int rcount, MPI_Datatype rtype,int src, int rtag,
        MPI_Comm comm, MPI_Status *status)
{
	int rv;
	int bytes;
	double t0, t1;

	if( enabled ) {
		send_cnt++;
		if (comm == MPI_COMM_WORLD) {
			send_cnts[des]++;
		}
	
                if(scopy_check_enabled) MPI_SGI_scopy_check_start(scount,stype);
		t0 = MPI_Wtime(); 
		rv = PMPI_Sendrecv(sbuf, scount, stype, des, stag,
                                   rbuf, rcount, rtype, src, rtag, comm,status);
		t1 = MPI_Wtime(); 
	        if(scopy_check_enabled) MPI_SGI_scopy_check_end("MPI_Sendrecv");
		send_time += t1-t0;
		mpi_sum += t1-t0;
		if (comm == MPI_COMM_WORLD) {
			send_times[des] += t1-t0;
		}
	
		MPI_Type_size( stype, &bytes );
		bytes *= scount;
		dist_collect( bytes, des, send );

                recv_cnt++;
                recv_time += t1-t0;
                mpi_sum += t1-t0;
                if (comm == MPI_COMM_WORLD) {
                        if (src == MPI_ANY_SOURCE)
                                src = status->MPI_SOURCE;
                        recv_cnts[src]++;
                        recv_times[src] += t1-t0;

                }

#ifdef MPICH_NAME
		dist_collect( (status->count), src, recv );
#else
		dist_collect( (status->size), src, recv );
#endif


	
		return rv;
	} else {

	 return PMPI_Sendrecv(sbuf, scount, stype, des, stag, 
                              rbuf, rcount, rtype, src, rtag, comm, status);
	}
}

void
mpi_sendrecv_(void *sbuf, int *scount, MPI_Datatype *stype, int *des, int *stag,
              void *rbuf, int *rcount, MPI_Datatype *rtype, int *src, int *rtag,
              MPI_Comm *comm, MPI_Status *status, int *ierr)
{
	*ierr = MPI_Sendrecv(sbuf, *scount, *stype, *des, *stag, 
                             rbuf, *rcount, *rtype, *src, *rtag, *comm, status);
}


/* ------------------------------------------------------------------------ */
/* MPI_Isend */

static long isend_cnt, isend_cnts[MAXRANKS];
static double isend_time, isend_times[MAXRANKS];

int
MPI_Isend(void *buf, int count, MPI_Datatype type, int des, int tag,
	MPI_Comm comm, MPI_Request *req)
{
	int rv;
	double t0, t1;

	if( enabled ) {
		isend_cnt++;
		if (comm == MPI_COMM_WORLD) {
			isend_cnts[des]++;
		}
	
                if(scopy_check_enabled) MPI_SGI_scopy_check_start(count,type);
		t0 = MPI_Wtime(); 
		rv = PMPI_Isend(buf, count, type, des, tag, comm, req);
		t1 = MPI_Wtime(); 
                if(scopy_check_enabled) MPI_SGI_scopy_check_end("MPI_Isend");
		isend_time += t1-t0;
		mpi_sum += t1-t0;
		if (comm == MPI_COMM_WORLD) {
			isend_times[des] += t1-t0;
		}
	
		return rv;
	} else {
		return PMPI_Isend(buf, count, type, des, tag, comm, req);
	}
}

void
mpi_isend_(void *buf, int *count, MPI_Datatype *type, int *des, int *tag,
        MPI_Comm *comm, MPI_Request *req, int *ierr)
{
	*ierr = MPI_Isend(buf, *count, *type, *des, *tag, *comm, req);
}


/* ------------------------------------------------------------------------ */
/* MPI_Ssend */

static long ssend_cnt, ssend_cnts[MAXRANKS];
static double ssend_time, ssend_times[MAXRANKS];

int
MPI_Ssend(void *buf, int count, MPI_Datatype type, int des, int tag,
        MPI_Comm comm)
{
        int rv;
        double t0, t1;

	if( enabled ) {
        	ssend_cnt++;
        	if (comm == MPI_COMM_WORLD) {
                	ssend_cnts[des]++;
        	}
	
        	t0 = MPI_Wtime();
        	rv = PMPI_Ssend(buf, count, type, des, tag, comm);
        	t1 = MPI_Wtime();
        	ssend_time += t1-t0;
		mpi_sum += t1-t0;
        	if (comm == MPI_COMM_WORLD) {
                	ssend_times[des] += t1-t0;
        	}
	
        	return rv;
	} else {
		return PMPI_Ssend(buf, count, type, des, tag, comm);
	}
}

void
mpi_ssend_(void *buf, int *count, MPI_Datatype *type, int *des, int *tag,
        MPI_Comm *comm, int *ierr)
{
        *ierr = MPI_Ssend(buf, *count, *type, *des, *tag, *comm);
}


/* ------------------------------------------------------------------------ */
/* MPI_Wait */

static long wait_cnt/*, wait_cnts[MAXRANKS]*/;
static double wait_time/*, wait_times[MAXRANKS]*/;
dist_declare(wait)

int
MPI_Wait(MPI_Request *request, MPI_Status *status)
{
	int rv;

	if( enabled ) {
		wait_cnt++;
		TIMEIT_NORETURN( PMPI_Wait(request, status), wait_time);
	
#ifdef MPICH_NAME
		dist_collect( (status->count), 0, wait );
#else
		dist_collect( (status->size), 0, wait );
#endif
		return rv;
	} else {
		return PMPI_Wait(request, status);
	}
}

void
mpi_wait_(MPI_Request *request, MPI_Status *status, int *ierr)
{
	*ierr = MPI_Wait(request, status);
}


/* ------------------------------------------------------------------------ */
/* MPI_Waitall */

static long waitall_datacnt;
static long waitall_cnt/*, waitall_cnts[MAXRANKS]*/;
static long waitall_numreq;
static double waitall_time/*, waitall_times[MAXRANKS]*/;

int
MPI_Waitall(int numreq, MPI_Request *request, MPI_Status *status)
{
	int rv,i;

	if( enabled ) {
		waitall_cnt++;
		TIMEIT_NORETURN( PMPI_Waitall(numreq, request, status),
                                 waitall_time);

        	for (i = 0; i < numreq; i++,status++)
#ifdef MPICH_NAME
			waitall_datacnt += status->count;
#else
			waitall_datacnt += status->size;
#endif
        	waitall_numreq += numreq;
		return rv;
	} else {
		return PMPI_Waitall(numreq, request, status);
	}
}

void
mpi_waitall_(int *numreq, MPI_Request *request, MPI_Status *status, int *ierr)
{
	*ierr = MPI_Waitall(*numreq, request, status);
}



/* ------------------------------------------------------------------------ */
/* MPI_Barrier */

static long barrier_cnt;
static double barrier_time;

int
MPI_Barrier(MPI_Comm comm)
{
	if( enabled ) {
		barrier_cnt++;
		TIMEIT( PMPI_Barrier(comm), barrier_time );
	} else {
		return PMPI_Barrier(comm);
	}
}

void
mpi_barrier_(MPI_Comm *comm, int *ierr)
{
	*ierr = MPI_Barrier(*comm);
}


/* ------------------------------------------------------------------------ */
/* MPI_Alltoall */

static long alltoall_cnt;
static double alltoall_time;
static long alltoall_datacnt;

int
MPI_Alltoall(void *sendbuf, int sendcount, MPI_Datatype sendtype,
        void *recvbuf, int recvcount, MPI_Datatype recvtype, MPI_Comm comm)
{
	if( enabled ) {
		alltoall_cnt++;
		alltoall_datacnt += sendcount;
		TIMEIT( PMPI_Alltoall(sendbuf, sendcount, sendtype, recvbuf,
			recvcount, recvtype, comm), alltoall_time );
	} else {
		return PMPI_Alltoall(sendbuf, sendcount, sendtype, recvbuf,
                                     recvcount, recvtype, comm);
	}
}

void
mpi_alltoall_( void *sendbuf, int *sendcount, MPI_Datatype *sendtype,
	void *recvbuf, int *recvcnt, MPI_Datatype *recvtype,
	MPI_Comm *comm, int *ierr)
{
        *ierr = MPI_Alltoall(sendbuf, *sendcount, *sendtype, recvbuf,
		*recvcnt, *recvtype, *comm);
}


/* ------------------------------------------------------------------------ */
/* MPI_Alltoallv */

static long alltoallv_cnt;
static double alltoallv_time;

int
MPI_Alltoallv(void *sendbuf, int *sendcnts, int *sdispls,
	MPI_Datatype sendtype, void *recvbuf, int *recvcnts, int *rdispls,
	MPI_Datatype recvtype, MPI_Comm comm)
{
	if( enabled ) {
        	alltoallv_cnt++;
        	TIMEIT( PMPI_Alltoallv(sendbuf, sendcnts, sdispls, sendtype,
                                       recvbuf, recvcnts, rdispls, recvtype,
                                       comm), alltoallv_time );
	} else {
		return PMPI_Alltoallv(sendbuf, sendcnts, sdispls, sendtype,
                                      recvbuf, recvcnts, rdispls, recvtype,
                                      comm);
	}
}

void
mpi_alltoallv_( void *sendbuf, int *sendcnts, int *sdispls,
	MPI_Datatype *sendtype, void *recvbuf, int *recvcnts, int *rdispls,
	MPI_Datatype *recvtype, MPI_Comm *comm, int *ierr)
{
        *ierr = MPI_Alltoallv(sendbuf, sendcnts, sdispls, *sendtype, recvbuf,
		recvcnts, rdispls, *recvtype, *comm);
}


/* ------------------------------------------------------------------------ */
/* MPI_Reduce */

static long reduce_cnt;
static double reduce_time;

int
MPI_Reduce(void *sendbuf, void *recvbuf, int count, MPI_Datatype type,
	MPI_Op op, int root, MPI_Comm comm)
{
	if( enabled ) {
		reduce_cnt++;
		TIMEIT( PMPI_Reduce(sendbuf, recvbuf, count, type, op, root,
                                    comm), reduce_time );
	} else {
		return PMPI_Reduce(sendbuf, recvbuf, count, type, op, root,
                                    comm);
	}
}
 
void
mpi_reduce_( void *sendbuf, void *recvbuf, int *count, MPI_Datatype *type,
	MPI_Op *op, int *root, MPI_Comm *comm, int *ierr)
{
        *ierr = MPI_Reduce(sendbuf, recvbuf, *count, *type, *op, *root, *comm);
}


/* ------------------------------------------------------------------------ */
/* MPI_Allreduce */

static long allreduce_cnt;
static double allreduce_time;

int
MPI_Allreduce(void *sendbuf, void *recvbuf, int count, MPI_Datatype type,
	MPI_Op op, MPI_Comm comm)
{
	if( enabled ) {
		allreduce_cnt++;
		TIMEIT( PMPI_Allreduce(sendbuf, recvbuf, count, type, op, comm),
			allreduce_time );
	} else {
		return PMPI_Allreduce(sendbuf, recvbuf, count, type, op, comm);
	}
}

void
mpi_allreduce_( void *sendbuf, void *recvbuf, int *count,
	MPI_Datatype *type, MPI_Op *op, MPI_Comm *comm,
	int *ierr)
{
        *ierr = MPI_Allreduce(sendbuf, recvbuf, *count, *type, *op,
		*comm);
}


/* ------------------------------------------------------------------------ */
/* MPI_Bcast */

static long bcast_cnt;
static double bcast_time;
dist_declare(bcast)

int
MPI_Bcast(void *buffer, int count, MPI_Datatype type, int root, MPI_Comm comm)
{
	int bytes;
	
	if( enabled ) {
		bcast_cnt++;

		MPI_Type_size( type, &bytes );
		bytes *= count;
		dist_collect( bytes, 0, bcast );

		TIMEIT( PMPI_Bcast(buffer, count, type, root, comm),
			bcast_time );
	} else {
		return PMPI_Bcast(buffer, count, type, root, comm);
	}

}

void
mpi_bcast_( void *buffer, int *count, MPI_Datatype *type, int *root,
	MPI_Comm *comm, int *ierr)
{
	*ierr = MPI_Bcast(buffer, *count, *type, *root, *comm);
}


/* ------------------------------------------------------------------------ */
/* MPI_Gather */

static long gather_cnt;
static double gather_time;

int
MPI_Gather(void *sendbuf, int sendcount, MPI_Datatype sendtype, void *recvbuf,
	int recvcount, MPI_Datatype recvtype, int root, MPI_Comm comm)
{
	if( enabled ) {
		gather_cnt++;
		TIMEIT( PMPI_Gather(sendbuf, sendcount, sendtype, recvbuf,
			recvcount, recvtype, root, comm),
			gather_time );
	} else {
		return PMPI_Gather(sendbuf, sendcount, sendtype, recvbuf,
                        recvcount, recvtype, root, comm);
	}
}

void
mpi_gather_( void *sendbuf, int *sendcnt, MPI_Datatype *sendtype, void *recvbuf,
	int *recvcount, MPI_Datatype *recvtype, int *root, MPI_Comm *comm,
	int *ierr)
{
        *ierr = MPI_Gather(sendbuf, *sendcnt, *sendtype, recvbuf, *recvcount,
		*recvtype, *root, *comm);
}



/* ------------------------------------------------------------------------ */
/* MPI_Gatherv */

static long gatherv_cnt;
static double gatherv_time;

int
MPI_Gatherv(void *sendbuf, int sendcount, MPI_Datatype sendtype, void *recvbuf,
	int *recvcounts, int *displs, MPI_Datatype recvtype, int root,
	MPI_Comm comm)
{
	if( enabled ) {
		gatherv_cnt++;
		TIMEIT( PMPI_Gatherv(sendbuf, sendcount, sendtype, recvbuf,
        		recvcounts, displs, recvtype, root, comm),
			gatherv_time );
	} else {
		return PMPI_Gatherv(sendbuf, sendcount, sendtype, recvbuf,
                        recvcounts, displs, recvtype, root, comm);
	}
}

void
mpi_gatherv_(void *sendbuf, int *sendcnt, MPI_Datatype *sendtype,
	void *recvbuf, int *recvcnts, int *displs, MPI_Datatype *recvtype,
	int *root, MPI_Comm *comm, int *ierr)
{
	*ierr = MPI_Gatherv(sendbuf, *sendcnt, *sendtype, recvbuf, recvcnts,
		displs, *recvtype, *root, *comm);
}


/* ------------------------------------------------------------------------ */
/* MPI_Scatter */

static long scatter_cnt;
static double scatter_time;

int
MPI_Scatter(void *sendbuf, int sendcount, MPI_Datatype sendtype, void *recvbuf,
        int recvcount, MPI_Datatype recvtype, int root, MPI_Comm comm)
{
	if( enabled ) {
        	scatter_cnt++;
        	TIMEIT( PMPI_Scatter(sendbuf, sendcount, sendtype, recvbuf,
                	recvcount, recvtype, root, comm),
                	scatter_time );
	} else {
		return PMPI_Scatter(sendbuf, sendcount, sendtype, recvbuf,
                        recvcount, recvtype, root, comm);
	}
}

void
mpi_scatter_( void *sendbuf, int *sendcnt, MPI_Datatype *sendtype,
	void *recvbuf, int *recvcount, MPI_Datatype *recvtype, int *root,
	MPI_Comm *comm, int *ierr)
{
        *ierr = MPI_Scatter(sendbuf, *sendcnt, *sendtype, recvbuf, *recvcount,
                *recvtype, *root, *comm);
}


/* ------------------------------------------------------------------------ */
/* MPI_Scatterv */

static long scatterv_cnt;
static double scatterv_time;

int
MPI_Scatterv(void *sendbuf, int *sendcounts, int *displs,
	MPI_Datatype sendtype, void *recvbuf, int recvcount,
	MPI_Datatype recvtype, int root, MPI_Comm comm)
{
	if( enabled ) {
		scatterv_cnt++;
        	TIMEIT( PMPI_Scatterv(sendbuf, sendcounts, displs, sendtype,
			recvbuf, recvcount, recvtype, root, comm), 
			scatterv_time );
	} else {
		return PMPI_Scatterv(sendbuf, sendcounts, displs, sendtype,
                        recvbuf, recvcount, recvtype, root, comm);
	}
}

void
mpi_scatterv_( void *sendbuf, int *sendcnts, int *displs,
	MPI_Datatype *sendtype, void *recvbuf, int *recvcnt,
	MPI_Datatype *recvtype, int *root, MPI_Comm *comm, int *ierr)
{
	*ierr = MPI_Scatterv(sendbuf, sendcnts, displs, *sendtype, recvbuf,
		*recvcnt, *recvtype, *root, *comm);
}



/* ------------------------------------------------------------------------ */
/* MPI_Init_thread */

int MPI_Init_thread( int *argc, char ***argv, int required, int *provided )
{
	int rv, nranks, me;

	if( getenv("MPI_PROFILE_AT_INIT") ) {
		MPI_SGI_profiling_enable();
	} else {
		enabled = 0;
	}

	rv = PMPI_Init_thread( argc, argv, required, provided );
	PMPI_Comm_rank(MPI_COMM_WORLD, &me);
	if( getenv("MPI_ABORT_AT_INIT") && me == 0 ) {
		fprintf(stderr,
"Exiting, because environment variable MPI_ABORT_AT_INIT was set.\n");
		exit(47);
	}
	tinit = tstart;
	PMPI_Comm_size(MPI_COMM_WORLD, &nranks);
	if (nranks > MAXRANKS) {
		if (me == 0) {
			fprintf(stderr,
"Profiling library maximum MPI process limit of %d exceeded.\n", MAXRANKS);
			exit(-1);
		}
	}
	return rv;
}


/* ------------------------------------------------------------------------ */
/* MPI_Init */

int
MPI_Init( int  *argc, char ***argv)
{
	int rv, nranks, me;

	if( getenv("MPI_PROFILE_AT_INIT") ) {
		MPI_SGI_profiling_enable();
	} else {
		enabled = 0;
	}
        if( getenv("MPI_SCOPY_CHECK_AT_INIT")) {
		MPI_SGI_scopy_check_enable();
        } else {
		scopy_check_enabled=0;
        }

	rv = PMPI_Init( argc, argv );
	PMPI_Comm_rank(MPI_COMM_WORLD, &me);
	if( getenv("MPI_ABORT_AT_INIT") && me == 0 ) {
		fprintf(stderr,
"Exiting, because environment variable MPI_ABORT_AT_INIT was set.\n");
		exit(47);
	}
	tinit = tstart;
	PMPI_Comm_size(MPI_COMM_WORLD, &nranks);
	if (nranks > MAXRANKS) {
		if (me == 0) {
			fprintf(stderr,
"Profiling library maximum MPI process limit of %d exceeded\n", MAXRANKS);
			exit(-1);
		}
	}
	return rv;
}

void
mpi_init_(int *ier)
{
        int x;
        char ** a;
        x=0;
        MPI_Init(&x, &a);
        *ier = 0;
}

/* ------------------------------------------------------------------------ */
/* MPI_Finalize */


static void
print_rank0_info(FILE *f, int nranks, double *all_timings)
{
	char *dt;
	double tk = MPI_Wtick();
	int i;
	time_t t;

	time(&t);
	dt = ctime(&t);

	fprintf(f,"Summary counts and timings  %s\n", dt);
	fprintf(f,
"Total MPI processes                                  %d\n\n",nranks);
	if( enabled ) {
		double total_tinit   = 0.0;
		double total_tstart  = 0.0;
		double total_tend    = 0.0;
		double total_mpi_sum = 0.0;
		
		int best_rank, worst_rank;
		double best_percent, worst_percent;

		for( i = 0; i < nranks; i++ ) {
			total_tinit   += all_timings[i*4];
			total_tstart  += all_timings[i*4+1];
			total_tend    += all_timings[i*4+2];
			total_mpi_sum += all_timings[i*4+3];
		}
		
		fprintf(f,
"Total MPI job time, avg per rank                    %g sec\n",
			(total_tend-total_tinit)/nranks);
		fprintf(f,
"Profiled job time, avg per rank                     %g sec\n",
			(total_tend-total_tstart)/nranks);
		if(total_tend-total_tinit) fprintf(f,
"Percent job time profiled, avg per rank             %g%%\n\n",
		   ((total_tend-total_tstart)/(total_tend-total_tinit))*100.0);

		fprintf(f,
"Time in all profiled MPI routines, avg per rank     %g sec\n",
			 total_mpi_sum/nranks);

		if(tend-tstart) fprintf(f,
"Percent time in profiled MPI routines, avg per rank %g%%\n\n",
		   (1.0-(((total_tend-total_tstart)-total_mpi_sum)/(total_tend-total_tstart)))*100.0);

		fprintf(f,"Rank:Percent in profiled MPI routines\n\t");
		best_rank = -1;
		best_percent = 10000;
		worst_rank = -1;
		worst_percent = -10000;

		for( i=0; i < nranks; i++ ) {
			double tstart  = all_timings[i*4+1];
			double tend    = all_timings[i*4+2];
			double mpi_sum = all_timings[i*4+3];
			double percent;

			percent = (1.0-(((tend-tstart)-mpi_sum)/
				(tend-tstart)))*100.0;

			fprintf(f,"%d:%g%%\t", i, percent);
			if( i%4 == 3 ) fprintf(f, "\n\t");
			
			if( percent < best_percent ) {
				best_percent = percent;
				best_rank = i;
			}
			if( percent > worst_percent ) {
				worst_percent = percent;
				worst_rank = i;
			}
		}
		fprintf( f, "\n" );
		fprintf( f, "Best:   Rank %d      %g%%\n",
				 best_rank, best_percent);
		fprintf( f, "Worst:  Rank %d      %g%%\n",
				 worst_rank, worst_percent);
		{
			float t1 = 100.0 - best_percent;
			float t2 = 100.0 - worst_percent;
			fprintf( f, "Load Imbalance:  %f%%\n",
				(t1-t2)/(t1+t2)*100.0);
		}
	
	} else {
		fprintf(f, "MPI_SGI_profiling_enable() was never called.\n"); 
		fprintf(f, 
"Consider rerunning with environment variable MPI_PROFILE_AT_INIT set.\n");
	}

	fprintf(f,
"\nWtime resolution is                   %g sec\n",tk);
	fflush(f);
}

#define show(x) fprintf(f,"%10s calls %ld\ttime %g\n", #x, x##_cnt, x##_time);

static void
print_summary_counts(FILE *f, int rank, int nranks) 
{
	int i, j;
	int ncols;

	for (i=0; i<nranks; i++) {
	    if (rank == i) {
		float ave;

		fprintf(f,"\nactivity on process rank %d\n",rank);
	        scopy_display(f);

		if(comm_rank_cnt){
			show(comm_rank);
		}
		if(get_count_cnt){
			show(get_count);
		}
		if(ibsend_cnt){
			show(ibsend);
		}
		if(probe_cnt){ 
			show(probe);
		}
		if( recv_cnt ){
			fprintf(f,
"%10s calls %ld\ttime %g   datacnt %ld  waits %ld  wait time %g\n",
				"recv", recv_cnt, recv_time, recv_datacnt,
					recv_nomsg_cnt, recv_nomsg_time);
			dist_print( recv_cnt, recv );
		}
		if( irecv_cnt ) {
			fprintf(f,
"%10s calls %ld\ttime %g   datacnt %ld\n",
				"irecv", irecv_cnt, irecv_time, irecv_datacnt);
			dist_print( irecv_cnt, irecv );
		}
		if(send_cnt){
			fprintf(f,
"%10s calls %ld\ttime %g   datacnt %ld\n",
				"send", send_cnt, send_time, send_datacnt);
			dist_print( send_cnt, send );
		}
		if(ssend_cnt){
			show(ssend);
		}
		if(isend_cnt){
			show(isend);
		}
		if(wait_cnt){
			ave = wait_datacnt/wait_cnt;
			fprintf(f,
"%10s calls %ld\ttime %g   avg datacnt %g\n", "wait", wait_cnt, wait_time, ave);
			dist_print( wait_cnt, wait );
		}
		if(waitall_cnt){
			ave=waitall_numreq ? waitall_datacnt/waitall_numreq : 0;
			fprintf(f,
"%10s calls %ld\ttime %g   # of reqs %ld  avg datacnt %g\n",
				"waitall",waitall_cnt, waitall_time,
				waitall_numreq, ave);
		}

		if(barrier_cnt){
			show(barrier);
		}
		if(alltoall_cnt){
			ave = alltoall_datacnt/alltoall_cnt;
			fprintf(f,
"%10s calls %ld\ttime %g   avg datacnt %g\n", "alltoall", alltoall_cnt,
 				alltoall_time, ave);
		}
		if(alltoallv_cnt){
			show(alltoallv);
		}
		if(reduce_cnt){
			show(reduce);
		}
		if(allreduce_cnt){
			show(allreduce);
		}
		if(bcast_cnt){
			show(bcast);
			dist_print( bcast_cnt, bcast );
		}
		if(gather_cnt){
			show(gather);
		}
		if(gatherv_cnt){
			show(gatherv);
		}
		if(scatter_cnt){
			show(scatter);
		}
		if(scatterv_cnt){
			show(scatterv);
		}
		fflush(f);
	    }
	    PMPI_Barrier(MPI_COMM_WORLD);
	}

/* 
 * Print point-to-point counts and timings.
 */
	ncols = nranks;

	if (ncols > 10) ncols = 10;

	/******************
	 * recv
	 */
	/* rank 0 prints the title and column headings */
	PMPI_Barrier(MPI_COMM_WORLD);
	if (rank == 0) {
		/* section title */
		fprintf(f,"\n\n");
		fprintf(f,"------------------------------------------------\n");
		fprintf(f,"\nrecv profile\n\n");
		
		/* column headings */
		fprintf(f,"             cnt/sec for all remote ranks\n");

		fprintf(f,"%5s","local");
		fprintf(f,"%13s","ANY_SOURCE");
		for (i=0; i<ncols; i++)
			fprintf(f,"%9d    ",i);
		fprintf(f,"\n");
		fprintf(f,"%5s","rank");

		fprintf(f,"\n");
		fflush(f);
	}
	PMPI_Barrier(MPI_COMM_WORLD);

	for (i=0; i<nranks; i++) {
	    if (rank == i) {
		if (recv_cnt != 0) {
		  fprintf(f,"%5d ",i);
		  if(recv_cnts_as)fprintf(f,"%4ld/%.1g ",
				recv_cnts_as, recv_times_as);
		  else            fprintf(f,"    /        ");
		  for (j=0; j<ncols; j++)
			if(recv_cnts[j])fprintf(f,"%4ld/%.1g ",
				recv_cnts[j],recv_times[j]);
			else            fprintf(f,"    /        ");
		  fprintf(f,"\n");
		  fflush(f);
	         }
	    }
	    PMPI_Barrier(MPI_COMM_WORLD);
	}

        /******************
         * recv waits
         */
        /* rank 0 prints the title and column headings */
        PMPI_Barrier(MPI_COMM_WORLD);
        if (rank == 0) {
                /* section title */
		fprintf(f,"\n\n");
		fprintf(f,"------------------------------------------------\n");
                fprintf(f,"\nrecv wait for data profile\n\n");

                /* column headings */
                fprintf(f,"             cnt/sec for all remote ranks\n");

                fprintf(f,"%5s","local");
                for (i=0; i<ncols; i++)
                        fprintf(f,"%9d    ",i);
                fprintf(f,"\n");
                fprintf(f,"%5s","rank");

                fprintf(f,"\n");
                fflush(f);
        }
        PMPI_Barrier(MPI_COMM_WORLD);

        for (i=0; i<nranks; i++) {
            if (rank == i) {
                if (recv_nomsg_cnt != 0) {
                  fprintf(f,"%5d ",i);
                  for (j=0; j<ncols; j++)
                        if(recv_nomsg_cnts[j])fprintf(f,"%4ld/%.1e ",
                                recv_nomsg_cnts[j],recv_nomsg_times[j]);
			else fprintf(f,"    /        ");
                  fprintf(f,"\n");
                  fflush(f);
                 }
            }
            PMPI_Barrier(MPI_COMM_WORLD);
        }


	/******************
	 * send
	 */
	/* rank 0 prints the title and column headings */
	PMPI_Barrier(MPI_COMM_WORLD);
	if (rank == 0) {
		/* section title */
		fprintf(f,"\n\n");
		fprintf(f,"------------------------------------------------\n");
		fprintf(f,"\nsend profile\n\n");
		
		/* column headings */
		fprintf(f,"             cnt/sec for all destination ranks\n");

		fprintf(f,"%5s","src");
		for (i=0; i<ncols; i++)
			fprintf(f,"%9d    ",i);
		fprintf(f,"\n");
		fprintf(f,"%5s","rank");

		fprintf(f,"\n");
		fflush(f);
	}
	PMPI_Barrier(MPI_COMM_WORLD);

	for (i=0; i<nranks; i++) {
	    if (rank == i) {
		if (send_cnt != 0) {
		  fprintf(f,"%5d ",i);
		  for (j=0; j<ncols; j++)
			if(send_cnts[j])fprintf(f,"%4ld/%.1e ",
				send_cnts[j],send_times[j]);
			else            fprintf(f,"    /         ");
		  fprintf(f,"\n");
		  fflush(f);
	         }
	    }
	    PMPI_Barrier(MPI_COMM_WORLD);
	}

        /******************
         * ssend
         */
        /* rank 0 prints the title and column headings */
        PMPI_Barrier(MPI_COMM_WORLD);
        if (rank == 0) {
                /* section title */
		fprintf(f,"\n\n");
		fprintf(f,"------------------------------------------------\n");
                fprintf(f,"\nssend profile\n\n");

                /* column headings */
                fprintf(f,"             cnt/sec for all destination ranks\n");

                fprintf(f,"%5s","src");
                for (i=0; i<ncols; i++)
                        fprintf(f,"%9d    ",i);
                fprintf(f,"\n");
                fprintf(f,"%5s","rank");

                fprintf(f,"\n");
                fflush(f);
        }
        PMPI_Barrier(MPI_COMM_WORLD);

        for (i=0; i<nranks; i++) {
            if (rank == i) {
                if (ssend_cnt != 0) {
                  fprintf(f,"%5d ",i);
                  for (j=0; j<ncols; j++)
                        if(ssend_cnts[j])fprintf(f,"%4ld/%.1e ",
                                ssend_cnts[j],ssend_times[j]);
			else             fprintf(f,"    /        ");
                  fprintf(f,"\n");
                  fflush(f);
                 }
            }
            PMPI_Barrier(MPI_COMM_WORLD);
        }



	/******************
	 * ibsend
	 */
	/* rank 0 prints the title and column headings */
	PMPI_Barrier(MPI_COMM_WORLD);
	if (rank == 0) {
		/* section title */
		fprintf(f,"\n\n");
		fprintf(f,"------------------------------------------------\n");
		fprintf(f,"\nibsend profile\n\n");
		
		/* column headings */
		fprintf(f,"             cnt/sec for all destination ranks\n");

		fprintf(f,"%5s","src");
		for (i=0; i<ncols; i++)
			fprintf(f,"%9d    ",i);
		fprintf(f,"\n");
		fprintf(f,"%5s","rank");

		fprintf(f,"\n");
		fflush(f);
	}
	PMPI_Barrier(MPI_COMM_WORLD);

	for (i=0; i<nranks; i++) {
	    if (rank == i) {
		if (ibsend_cnt != 0) {
		  fprintf(f,"%5d ",i);
		  for (j=0; j<ncols; j++)
			if(ibsend_cnts[j])fprintf(f,"%4ld/%.1e ",
				ibsend_cnts[j],ibsend_times[j]);
			else              fprintf(f,"    /        ");
		  fprintf(f,"\n");
		  fflush(f);
	         }
	    }
	    PMPI_Barrier(MPI_COMM_WORLD);
	}

}


int
MPI_Finalize()
{
	FILE *f = fopen("MPI_PROFILING_STATS", "a");	
	double timings[4], *all_timings;
	int rank, nranks, rv;

	if ( f ) {
		int save_enabled;

		MPI_Comm_rank(MPI_COMM_WORLD, &rank);
		MPI_Comm_size(MPI_COMM_WORLD, &nranks);
	
		PMPI_Barrier(MPI_COMM_WORLD);
		tend = MPI_Wtime();
		
		/* send mpi_sum values to rank zero */
		if( rank == 0 ) all_timings =
			(double *)malloc( sizeof(timings) * nranks );
	
		timings[0] = tinit;
		timings[1] = tstart;
		timings[2] = tend;
		timings[3] = mpi_sum;
	
		save_enabled = enabled;
		enabled = 0;
		MPI_Gather( timings, 4, MPI_DOUBLE, all_timings, 4, MPI_DOUBLE, 0,
			MPI_COMM_WORLD );
		enabled = save_enabled;
	
		/* rank 0 prints the title and column headings */
		if (rank == 0) {
			print_rank0_info(f, nranks, all_timings);
		}
        	PMPI_Barrier(MPI_COMM_WORLD);
	
		if( enabled ) {
			print_summary_counts(f, rank, nranks);
		}
#ifdef LOADBARRIER
		MPI_SGI_loadbarrier_stats(f);
#endif
		fclose(f);
	}
		
	rv = PMPI_Finalize();
	return rv;
}

void
mpi_finalize_(int *ierr)
{
	*ierr = MPI_Finalize();
}


/* ------------------------------------------------------------------------ */
/* MPI_SGI_profiling_enable */



int
MPI_SGI_profiling_enable()
{
	tstart = MPI_Wtime();
	enabled = 1;
	return(0);
}

void
mpi_sgi_profiling_enable_(int *ierr)
{
        *ierr = MPI_SGI_profiling_enable();
}

/*reiner--------------------------------------------------------------------*/
int
MPI_SGI_scopy_check_enable()
{
	char *s;
        if( s = getenv("MPI_BUFFER_MAX") ) {
		sscanf(s, "%d", &scopy_buffer_max);

		if ( !(s=getenv("SCOPY_FAILURES_MAX")) ) {
			scopy_failures_max = 10;
		} else {
			sscanf(s, "%d", &scopy_failures_max);
		}
#ifndef MPICH_NAME
		scopy_check_enabled=1;
#endif
	} else {
		scopy_check_enabled=0;
        }
	return(0);
}

void
mpi_sgi_scopy_check_enable_(int *ierr)
{
	*ierr = MPI_SGI_scopy_check_enable();
}

/*reiner--------------------------------------------------------------------*/

#ifdef LOADBARRIER
/////////////////////////////////////////////////////////////////////
//
// MPI_SGI_loadbarrier()
//
// This function is callable from Fortran or C.  It
// tracks statistics about variances in arrival times at this
// barrier.  A barrier synchronization occurs at this call.

struct loadbarrier_data {
	
	double first_start_time;	// MPI_Wtime at first call to 
					// longbarrier()

	double interval_end_time;	// MPI_Wtime when this loadbarrier()
					// was arrived at.   Clock is sampled
					// before synchronizing.

	double total_wait;		// Time (sec) this rank waits. 
					// Computed as the difference between
					// interval_end_time on my rank
					// and interval_end_time on
					// the latest arriving rank at
					// each barrier.

	double program_wait;		// The sum across all barrier intervals
					// of the difference between the
					// latest and earliest barrier
					// arrival times.

	long numfirst;			// Number of times this process
					// was the first to arrive.

	long numlast;			// Number of times this process
					// was the last to arrive.

	long num_barriers;		// Count the synchronzations

	double max; 

};

static struct loadbarrier_data *lb_data;

static double program_total_wait = 0.0;

static int firstcall = 1;		// 1 if first call to loadbarrier

void
MPI_SGI_loadbarrier(void)
{
    static double min;
    static int minrank, maxrank;
    int i;
    static int myrank, numranks;


    if (firstcall) {
	firstcall = 0;
	lb_data =
		 (struct loadbarrier_data *)shmalloc(sizeof(struct loadbarrier_data));
	assert( lb_data );
	MPI_Comm_rank(MPI_COMM_WORLD, &myrank);
	MPI_Comm_size(MPI_COMM_WORLD, &numranks);
	lb_data->first_start_time = MPI_Wtime();
        lb_data->num_barriers++;
	return;
    }

    lb_data->num_barriers++;
    lb_data->interval_end_time = MPI_Wtime();

    MPI_Barrier(MPI_COMM_WORLD);	// synchronize all 

    // Rank 0 does all the work.
    if (myrank == 0) {
	min = -1;
	lb_data->max = 0;

	for (i=0; i<numranks; i++) {
	    double nexttime = shmem_double_g(&(lb_data->interval_end_time), i);
	    if (min == -1 || min > nexttime) {
		min = nexttime;
		minrank = i;
	    }
	    if (lb_data->max < nexttime) {
		lb_data->max = nexttime;
		maxrank = i;
	    }
	    program_total_wait += nexttime;
	}
	lb_data->program_wait += lb_data->max - min;
	program_total_wait -= numranks * min;

	// bump the count of slow arrivals for the last arriver.
	shmem_long_inc(&(lb_data->numlast), maxrank);
	shmem_long_inc(&(lb_data->numfirst), minrank);
    }

    MPI_Barrier(MPI_COMM_WORLD);	// synchronize all 

    if (myrank != 0) {
	lb_data->max =
		shmem_double_g(&(lb_data->max), 0);    	// get from rank 0
    }

    lb_data->total_wait += lb_data->max - lb_data->interval_end_time;

    MPI_Barrier(MPI_COMM_WORLD);	// synchronize the exit 

}

/////////////////////////////////////////////////////////////////////
//
// MPI_SGI_loadbarrier_stats(FILE *)
//
// Print out stats about the barrier synchronizations and load
// imbalance statistics.  Implies a barrier sync.  Called from MPI_Finalize().
//
void
MPI_SGI_loadbarrier_stats(FILE *f)
{
    int i, myrank, numranks;
    double endtime;
    MPI_Comm_rank(MPI_COMM_WORLD, &myrank);
    MPI_Comm_size(MPI_COMM_WORLD, &numranks);

    MPI_Barrier(MPI_COMM_WORLD);   // wait for all to arrive
    if (firstcall) {
	if (myrank == 0) fprintf(f,"\n\nMPI_SGI_loadbarrier() was never called.\n");
	return;   
    }
    endtime = MPI_Wtime();

    if (myrank == 0) {
	double diff = endtime - lb_data->first_start_time;
	fprintf(f,"\n\n\n*** MPI_SGI_loadbarrier() stats ****\n\n");
	fprintf(f,
" %ld MPI_SGI_loadbarrier()s in %g sec\n\t%g MPI_SGI_loadbarrier()s per sec\n\t%g sec per MPI_SGI_loadbarrier()\n",
		lb_data->num_barriers, diff, lb_data->num_barriers/diff, 
		diff/lb_data->num_barriers );
	fprintf(f,"\nTime during which cpus were under-utilized: %g sec\n", 
		lb_data->program_wait);
	fprintf(f,"CPU-seconds by which CPUs were under-utilized: %g sec\n", 
		program_total_wait);
	fprintf(f,"Average CPUs waiting: %g\n", 
		program_total_wait/lb_data->program_wait);

	fprintf(f,"\nRank     Totalwait    Times First    Times Last\n");
	for (i=0; i<numranks; i++) {
	    fprintf(f,"%4d%14g%14ld%14ld\n",
		i,
		shmem_double_g(&(lb_data->total_wait), i), 
		shmem_long_g(&(lb_data->numfirst), i), 
		shmem_long_g(&(lb_data->numlast), i)); 
	}
	fprintf(f,"\n*** end MPI_SGI_loadbarrier() stats ****\n");
    }
}



/////////////////////////////////////////////////////////////////////
//
// Fortran entry point is mpi_sgi_loadbarrier_ 
//
void mpi_sgi_loadbarrier_(void)       { MPI_SGI_loadbarrier(); }
#endif
