// TITLE:	Wrapper for calling MPI with 64-bit INTEGER and/or REAL data.
// AUTHOR:	Parallel Communication Team at SGI
// UPDATED:	May, 2002
// USAGE:	See the accompanying README & Makefile.
// STATUS:	Updated for new features added with MPT 1.6.
//		Incomplete handling of array_of_commands and argv
//		parameters in: mpi_comm_spawn_multiple_() and
//		mpi_comm_spawn_().  Grep this file for 'INCOMPLETE'.
//		Additional information is provided in section
//		entitled NOTES at the end of this file.
// 

#include <assert.h>
#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


// Macros
#define UNSUPPORTED(functionname) \
void functionname() \
{ \
printf("Wrapper function " ## #functionname ## " is not yet supported...\n"); \
MPI_Finalize(); \
abort(); \
}

// Typedefs

#if SIZES_I8_R8
// Definitions for -i8 -r8 -d16 bindings:
typedef long long       fortran_int; 
typedef double          fortran_real; 
typedef long double     fortran_double; 

#elif SIZES_I4_R8
// Definitions for -i4 -r8 d16 bindings:
typedef int             fortran_int; 
typedef double          fortran_real; 
typedef long double     fortran_double; 

#else	// SIZES_I4_R4
// Definitions for -i4 -r4 d8 bindings:
typedef int             fortran_int; 
typedef float           fortran_real; 
typedef double          fortran_double; 
#endif

// Need to special-case the MPI_Status structure...
// From mpi.h:
typedef struct {
        fortran_int MPI_SOURCE;
        fortran_int MPI_TAG;
        fortran_int MPI_ERROR;
        fortran_int size;
        fortran_int reserved[2];
} fortran_MPI_Status;

enum {
        MPI_WRAPPER_TO_FORTRAN  = 0,
        FORTRAN_TO_MPI_WRAPPER  = 1
};

// Prototypes
static int convert_type(int ftype);
void       copy_status_struct(fortran_MPI_Status *status, MPI_Status *x_status, int direction);

//
// Utility function for converting type.
//
static int 
convert_type(int ftype)
{
        int x_type;
#ifdef DEBUG
printf("in convert_type, ftype = %i\n",ftype);
#endif
        if (ftype == MPI_REAL) {
            if (sizeof(fortran_real) == sizeof(float)) {
#ifdef DEBUG
		printf ("x_type = MPI_REAL = %i\n",MPI_REAL);
#endif
                x_type = MPI_REAL;
            } else if (sizeof(fortran_real) == 4) {
#ifdef DEBUG
		printf ("x_type = MPI_REAL4 = %i\n",MPI_REAL4);
#endif
                x_type = MPI_REAL4;
            } else if (sizeof(fortran_real) == 8) {
#ifdef DEBUG
		printf ("x_type = MPI_REAL8 = %i\n",MPI_REAL8);
#endif
                x_type = MPI_REAL8;
            } else if (sizeof(fortran_real) == 16) {
#ifdef DEBUG
		printf ("x_type = MPI_REAL16 = %i\n",MPI_REAL16);
#endif
                x_type = MPI_REAL16;
            } else
                assert(0);
        }
        else if (ftype == MPI_INTEGER) {
            if (sizeof(fortran_int) == sizeof(int)) {
                x_type = MPI_INTEGER;
#ifdef DEBUG
		printf ("x_type = MPI_INTEGER = %i\n",MPI_INTEGER);
#endif
            } else if (sizeof(fortran_int) == 4) {
                x_type = MPI_INTEGER4;
#ifdef DEBUG
		printf ("x_type = MPI_INTEGER4 = %i\n",MPI_INTEGER4);
#endif
            } else if (sizeof(fortran_int) == 8) {
                x_type = MPI_INTEGER8;
#ifdef DEBUG
		printf ("x_type = MPI_INTEGER8 = %i\n",MPI_INTEGER8);
#endif
            } else
                assert(0);
        }
        else {
                x_type = ftype;
#ifdef DEBUG
		printf ("x_type hasn't changed = %i\n",x_type);
#endif
        }

        return x_type;
}

//
// Utility function for copying a C-style MPI_Status struct to a
// fortran_MPI_Status typedef for return to the Fortran calling program.
// This function will also do the opposite, namely, copy a fortran_MPI_Status
// struct to the C-style MPI_Status for input into an MPI routine.
//
void
copy_status_struct(
fortran_MPI_Status *status,
MPI_Status *x_status,
int direction)
{
   if (direction == MPI_WRAPPER_TO_FORTRAN) {
#ifdef DEBUG
printf("copying C to Fortran in copy_status_struct\n");
#endif
     status->MPI_SOURCE = x_status->MPI_SOURCE;
     status->MPI_TAG = x_status->MPI_TAG;
     status->MPI_ERROR = x_status->MPI_ERROR;
     status->size = x_status->size;
     status->reserved[0] = x_status->reserved[0];
     status->reserved[1] = x_status->reserved[1];
   } else if (direction == FORTRAN_TO_MPI_WRAPPER) {
#ifdef DEBUG
printf("copying Fortran to C in copy_status_struct\n");
#endif
     x_status->MPI_SOURCE = status->MPI_SOURCE;
     x_status->MPI_TAG = status->MPI_TAG;
     x_status->MPI_ERROR = status->MPI_ERROR;
     x_status->size = status->size;
     x_status->reserved[0] = status->reserved[0];
     x_status->reserved[1] = status->reserved[1];
   }
}

// Support both extant linux fortran name mangling schemes
#ifdef __linux
#pragma weak mpi_abort__ = pmpi_abort_
#pragma weak mpi_address__ = pmpi_address_
#pragma weak mpi_allgather__ = pmpi_allgather_
#pragma weak mpi_allgatherv__ = pmpi_allgatherv_
#pragma weak mpi_allreduce__ = pmpi_allreduce_
#pragma weak mpi_alltoall__ = pmpi_alltoall_
#pragma weak mpi_alltoallv__ = pmpi_alltoallv_
#pragma weak mpi_attr_delete__ = pmpi_attr_delete_
#pragma weak mpi_attr_get__ = pmpi_attr_get_
#pragma weak mpi_attr_put__ = pmpi_attr_put_
#pragma weak mpi_barrier__ = pmpi_barrier_
#pragma weak mpi_bcast__ = pmpi_bcast_
#pragma weak mpi_bsend_init__ = pmpi_bsend_init_
#pragma weak mpi_bsend__ = pmpi_bsend_
#pragma weak mpi_buffer_attach__ = pmpi_buffer_attach_
#pragma weak mpi_buffer_detach__ = pmpi_buffer_detach_
#pragma weak mpi_cancel__ = pmpi_cancel_
#pragma weak mpi_cart_coords__ = pmpi_cart_coords_
#pragma weak mpi_cart_create__ = pmpi_cart_create_
#pragma weak mpi_cartdim_get__ = pmpi_cartdim_get_
#pragma weak mpi_cart_get__ = pmpi_cart_get_
#pragma weak mpi_cart_map__ = pmpi_cart_map_
#pragma weak mpi_cart_rank__ = pmpi_cart_rank_
#pragma weak mpi_cart_shift__ = pmpi_cart_shift_
#pragma weak mpi_cart_sub__ = pmpi_cart_sub_
#pragma weak mpi_comm_compare__ = pmpi_comm_compare_
#pragma weak mpi_comm_create__ = pmpi_comm_create_
#pragma weak mpi_comm_dup__ = pmpi_comm_dup_
#pragma weak mpi_comm_free__ = pmpi_comm_free_
#pragma weak mpi_comm_group__ = pmpi_comm_group_
#pragma weak mpi_comm_rank__ = pmpi_comm_rank_
#pragma weak mpi_comm_remote_group__ = pmpi_comm_remote_group_
#pragma weak mpi_comm_remote_size__ = pmpi_comm_remote_size_
#pragma weak mpi_comm_size__ = pmpi_comm_size_
#pragma weak mpi_comm_split__ = pmpi_comm_split_
#pragma weak mpi_comm_test_inter__ = pmpi_comm_test_inter_
#pragma weak mpi_dims_create__ = pmpi_dims_create_
#pragma weak mpi_errhandler_create__ = pmpi_errhandler_create_
#pragma weak mpi_errhandler_free__ = pmpi_errhandler_free_
#pragma weak mpi_errhandler_get__ = pmpi_errhandler_get_
#pragma weak mpi_errhandler_set__ = pmpi_errhandler_set_
#pragma weak mpi_error_class__ = pmpi_error_class_
#pragma weak mpi_error_string__ = pmpi_error_string_
#pragma weak mpi_finalize__ = pmpi_finalize_
#pragma weak mpi_gather__ = pmpi_gather_
#pragma weak mpi_gatherv__ = pmpi_gatherv_
#pragma weak mpi_get_count__ = pmpi_get_count_
#pragma weak mpi_get_elements__ = pmpi_get_elements_
#pragma weak mpi_get_processor_name__ = pmpi_get_processor_name_
#pragma weak mpi_graph_create__ = pmpi_graph_create_
#pragma weak mpi_graphdims_get__ = pmpi_graphdims_get_
#pragma weak mpi_graph_get__ = pmpi_graph_get_
#pragma weak mpi_graph_map__ = pmpi_graph_map_
#pragma weak mpi_graph_neighbors_count__ = pmpi_graph_neighbors_count_
#pragma weak mpi_graph_neighbors__ = pmpi_graph_neighbors_
#pragma weak mpi_group_compare__ = pmpi_group_compare_
#pragma weak mpi_group_difference__ = pmpi_group_difference_
#pragma weak mpi_group_excl__ = pmpi_group_excl_
#pragma weak mpi_group_free__ = pmpi_group_free_
#pragma weak mpi_group_incl__ = pmpi_group_incl_
#pragma weak mpi_group_intersection__ = pmpi_group_intersection_
#pragma weak mpi_group_range_excl__ = pmpi_group_range_excl_
#pragma weak mpi_group_range_incl__ = pmpi_group_range_incl_
#pragma weak mpi_group_rank__ = pmpi_group_rank_
#pragma weak mpi_group_size__ = pmpi_group_size_
#pragma weak mpi_group_translate_ranks__ = pmpi_group_translate_ranks_
#pragma weak mpi_group_union__ = pmpi_group_union_
#pragma weak mpi_ibsend__ = pmpi_ibsend_
#pragma weak mpi_initialized__ = pmpi_initialized_
#pragma weak mpi_init__ = pmpi_init_
#pragma weak mpi_intercomm_create__ = pmpi_intercomm_create_
#pragma weak mpi_intercomm_merge__ = pmpi_intercomm_merge_
#pragma weak mpi_iprobe__ = pmpi_iprobe_
#pragma weak mpi_irecv__ = pmpi_irecv_
#pragma weak mpi_irsend__ = pmpi_irsend_
#pragma weak mpi_isend__ = pmpi_isend_
#pragma weak mpi_issend__ = pmpi_issend_
#pragma weak mpi_keyval_create__ = pmpi_keyval_create_
#pragma weak mpi_keyval_free__ = pmpi_keyval_free_
#pragma weak mpi_op_create__ = pmpi_op_create_
#pragma weak mpi_op_free__ = pmpi_op_free_
#pragma weak mpi_pack__ = pmpi_pack_
#pragma weak mpi_pack_size__ = pmpi_pack_size_
#pragma weak mpi_pcontrol__ = pmpi_pcontrol_
#pragma weak mpi_probe__ = pmpi_probe_
#pragma weak mpi_recv_init__ = pmpi_recv_init_
#pragma weak mpi_recv__ = pmpi_recv_
#pragma weak mpi_reduce__ = pmpi_reduce_
#pragma weak mpi_reduce_scatter__ = pmpi_reduce_scatter_
#pragma weak mpi_request_free__ = pmpi_request_free_
#pragma weak mpi_rsend_init__ = pmpi_rsend_init_
#pragma weak mpi_rsend__ = pmpi_rsend_
#pragma weak mpi_scan__ = pmpi_scan_
#pragma weak mpi_scatter__ = pmpi_scatter_
#pragma weak mpi_scatterv__ = pmpi_scatterv_
#pragma weak mpi_send_init__ = pmpi_send_init_
#pragma weak mpi_send__ = pmpi_send_
#pragma weak mpi_sendrecv__ = pmpi_sendrecv_
#pragma weak mpi_sendrecv_replace__ = pmpi_sendrecv_replace_
#pragma weak mpi_ssend_init__ = pmpi_ssend_init_
#pragma weak mpi_ssend__ = pmpi_ssend_
#pragma weak mpi_startall__ = pmpi_startall_
#pragma weak mpi_start__ = pmpi_start_
#pragma weak mpi_testall__ = pmpi_testall_
#pragma weak mpi_testany__ = pmpi_testany_
#pragma weak mpi_test_cancelled__ = pmpi_test_cancelled_
#pragma weak mpi_test__ = pmpi_test_
#pragma weak mpi_testsome__ = ppmpi_testsome_
#pragma weak mpi_topo_test__ = pmpi_topo_test_
#pragma weak mpi_type_commit__ = pmpi_type_commit_
#pragma weak mpi_type_contiguous__ = pmpi_type_contiguous_
#pragma weak mpi_type_extent__ = pmpi_type_extent_
#pragma weak mpi_type_free__ = pmpi_type_free_
#pragma weak mpi_type_hindexed__ = pmpi_type_hindexed_
#pragma weak mpi_type_hvector__ = pmpi_type_hvector_
#pragma weak mpi_type_indexed__ = pmpi_type_indexed_
#pragma weak mpi_type_lb__ = pmpi_type_lb_
#pragma weak mpi_type_size__ = pmpi_type_size_
#pragma weak mpi_type_struct__ = pmpi_type_struct_
#pragma weak mpi_type_ub__ = pmpi_type_ub_
#pragma weak mpi_type_vector__ = pmpi_type_vector_
#pragma weak mpi_unpack__ = pmpi_unpack_
#pragma weak mpi_waitall__ = pmpi_waitall_
#pragma weak mpi_waitany__ = pmpi_waitany_
#pragma weak mpi_wait__ = pmpi_wait_
#pragma weak mpi_waitsome__ = ppmpi_waitsome_
#pragma weak mpi_wtick__ = pmpi_wtick_
#pragma weak mpi_wtime__ = pmpi_wtime_
#endif


//
// Wrapper routines from the MPI-1.2 standard
//

//   
//   MPI_ABORT( comm, errorcode )
//   [ IN comm] communicator of tasks to abort
//   [ IN errorcode] error code to return to invoking environment
//   
//   int MPI_Abort(MPI_Comm comm, int errorcode)
//   
//   MPI_ABORT(COMM, ERRORCODE, IERROR)
//   INTEGER COMM, ERRORCODE, IERROR
//   
void
mpi_abort_(
fortran_int  *comm,
fortran_int  *errorcode,
fortran_int  *ierr)
{

#ifdef DEBUG
  printf("wrapper.c... input to	PMPI_Abort: *comm = %i...\n",(int) *comm);
#endif
        *ierr = (fortran_int) PMPI_Abort((MPI_Comm) *comm, (int) *errorcode);
}

//
// MPI_BARRIER( comm )
// [ IN comm] communicator (handle)
//
// int MPI_Barrier(MPI_Comm comm )
//
// MPI_BARRIER(COMM, IERROR)
// INTEGER COMM, IERROR
//
void
mpi_barrier_(
fortran_int  *comm,
fortran_int  *ierr)
{
#ifdef DEBUG
  printf("wrapper.c... input to PMPI_Barrier: *comm = %i\n",(int) *comm);
#endif
        *ierr = (fortran_int) PMPI_Barrier((MPI_Comm) *comm);
}

//
//   MPI_SEND(buf, count, datatype, dest, tag, comm)
//   [ IN buf] initial address of send buffer (choice)
//   [ IN count] number of elements in send buffer (nonnegative integer)
//   [ IN datatype] datatype of each send buffer element (handle)
//   [ IN dest] rank of destination (integer)
//   [ IN tag] message tag (integer)
//   [ IN comm] communicator (handle)
//   
//   int MPI_Send(void* buf, int count, MPI_Datatype datatype, int dest,
//   int tag, MPI_Comm comm)
//   
//   MPI_SEND(BUF, COUNT, DATATYPE, DEST, TAG, COMM, IERROR)
//   <type> BUF(*)
//   INTEGER COUNT, DATATYPE, DEST, TAG, COMM, IERROR
//   
void
mpi_send_(
void         *buf,
fortran_int  *count,
fortran_int  *type,
fortran_int  *dest,
fortran_int  *tag,
fortran_int  *comm,
fortran_int  *ierr)
{
MPI_Datatype x_type = convert_type(*type);
#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Send with *type = %i, x_type = %i\n",(int) *type, x_type);
#endif
        *ierr = (fortran_int) PMPI_Send( (void *) buf, (int) *count, (MPI_Datatype) x_type, (int) *dest, (int) *tag, (MPI_Comm) *comm);
}

//   
//   MPI_RECV (buf, count, datatype, source, tag, comm, status)
//   [ OUT buf] initial address of receive buffer (choice)
//   [ IN count] number of elements in receive buffer (integer)
//   [ IN datatype] datatype of each receive buffer element (handle)
//   [ IN source] rank of source (integer)
//   [ IN tag] message tag (integer)
//   [ IN comm] communicator (handle)
//   [ OUT status] status object (Status)
//   
//   int MPI_Recv(void* buf, int count, MPI_Datatype datatype, int source,
//   int tag, MPI_Comm comm, MPI_Status *status)
//   
//   MPI_RECV(BUF, COUNT, DATATYPE, SOURCE, TAG, COMM, STATUS, IERROR)
//   <type> BUF(*)
//   INTEGER COUNT, DATATYPE, SOURCE, TAG, COMM, STATUS(MPI_STATUS_SIZE),
//   IERROR
void
mpi_recv_(
void         *buf,
fortran_int  *count,
fortran_int  *type,
fortran_int  *source,
fortran_int  *tag,
fortran_int  *comm,
fortran_MPI_Status  *status,
fortran_int  *ierr)
{
MPI_Status x_status;

MPI_Datatype x_type = convert_type(*type);
#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Recv with *type = %i, x_type = %i\n",(int) *type, x_type);
#endif
        *ierr = (fortran_int) PMPI_Recv( (void *) buf, (int) *count, (MPI_Datatype) x_type, (int) *source, (int) *tag, (MPI_Comm) *comm, &x_status);

        copy_status_struct(status, &x_status, MPI_WRAPPER_TO_FORTRAN);
}

//   
//   MPI_GET_COUNT(status, datatype, count)
//   [ IN status] return status of receive operation (Status)
//   [ IN datatype] datatype of each receive buffer entry (handle)
//   [ OUT count] number of received entries (integer)
//   
//   int MPI_Get_count(MPI_Status *status, MPI_Datatype datatype, int
//   *count)
//   
//   MPI_GET_COUNT(STATUS, DATATYPE, COUNT, IERROR)
//   INTEGER STATUS(MPI_STATUS_SIZE), DATATYPE, COUNT, IERROR
//   
void
mpi_get_count_(
fortran_MPI_Status  *status,
fortran_int  *type,
fortran_int  *count,
fortran_int  *ierr)
{
int x_count ;
MPI_Datatype x_type = convert_type(*type);

MPI_Status x_status;
        copy_status_struct(status, &x_status, FORTRAN_TO_MPI_WRAPPER);

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Get_count with *type = %i, x_type = %i\n",(int) *type, x_type);
#endif
        *ierr = (fortran_int) PMPI_Get_count( &x_status, (MPI_Datatype) x_type, &x_count);

        copy_status_struct(status, &x_status, MPI_WRAPPER_TO_FORTRAN);
        *count = (fortran_int) x_count ;
#ifdef DEBUG
  printf("wrapper.c... output from PMPI_Get_count: *count = %i...\n",(int) *count);
#endif
}

//   
//   MPI_BSEND (buf, count, datatype, dest, tag, comm)
//   [ IN buf] initial address of send buffer (choice)
//   [ IN count] number of elements in send buffer (integer)
//   [ IN datatype] datatype of each send buffer element (handle)
//   [ IN dest] rank of destination (integer)
//   [ IN tag] message tag (integer)
//   [ IN comm] communicator (handle)
//   
//   int MPI_Bsend(void* buf, int count, MPI_Datatype datatype, int dest,
//   int tag, MPI_Comm comm)
//   
//   MPI_BSEND(BUF, COUNT, DATATYPE, DEST, TAG, COMM, IERROR)
//   <type> BUF(*)
//   INTEGER COUNT, DATATYPE, DEST, TAG, COMM, IERROR
//   
void
mpi_bsend_(
void         *buf,
fortran_int  *count,
fortran_int  *type,
fortran_int  *dest,
fortran_int  *tag,
fortran_int  *comm,
fortran_int  *ierr)
{
MPI_Datatype x_type = convert_type(*type);
#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Bsend with *type = %i, x_type = %i\n",(int) *type, x_type);
#endif
        *ierr = (fortran_int) PMPI_Bsend( (void *) buf, (int) *count, (MPI_Datatype) x_type, (int) *dest, (int) *tag, (MPI_Comm) *comm);
}

//   
//   MPI_SSEND (buf, count, datatype, dest, tag, comm)
//   [ IN buf] initial address of send buffer (choice)
//   [ IN count] number of elements in send buffer (integer)
//   [ IN datatype] datatype of each send buffer element (handle)
//   [ IN dest] rank of destination (integer)
//   [ IN tag] message tag (integer)
//   [ IN comm] communicator (handle)
//   
//   int MPI_Ssend(void* buf, int count, MPI_Datatype datatype, int dest,
//   int tag, MPI_Comm comm)
//   
//   MPI_SSEND(BUF, COUNT, DATATYPE, DEST, TAG, COMM, IERROR)
//   <type> BUF(*)
//   INTEGER COUNT, DATATYPE, DEST, TAG, COMM, IERROR
//   
void
mpi_ssend_(
void         *buf,
fortran_int  *count,
fortran_int  *type,
fortran_int  *dest,
fortran_int  *tag,
fortran_int  *comm,
fortran_int  *ierr)
{
MPI_Datatype x_type = convert_type(*type);
#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Ssend with *type = %i, x_type = %i\n",(int) *type, x_type);
#endif
        *ierr = (fortran_int) PMPI_Ssend( (void *) buf, (int) *count, (MPI_Datatype) x_type, (int) *dest, (int) *tag, (MPI_Comm) *comm);
}

//   
//   MPI_RSEND (buf, count, datatype, dest, tag, comm)
//   [ IN buf] initial address of send buffer (choice)
//   [ IN count] number of elements in send buffer (integer)
//   [ IN datatype] datatype of each send buffer element (handle)
//   [ IN dest] rank of destination (integer)
//   [ IN tag] message tag (integer)
//   [ IN comm] communicator (handle)
//   
//   int MPI_Rsend(void* buf, int count, MPI_Datatype datatype, int dest,
//   int tag, MPI_Comm comm)
//   
//   MPI_RSEND(BUF, COUNT, DATATYPE, DEST, TAG, COMM, IERROR)
//   <type> BUF(*)
//   INTEGER COUNT, DATATYPE, DEST, TAG, COMM, IERROR
//   
void
mpi_rsend_(
void         *buf,
fortran_int  *count,
fortran_int  *type,
fortran_int  *dest,
fortran_int  *tag,
fortran_int  *comm,
fortran_int  *ierr)
{
MPI_Datatype x_type = convert_type(*type);
#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Rsend with *type = %i, x_type = %i\n",(int) *type, x_type);
#endif
        *ierr = (fortran_int) PMPI_Rsend( (void *) buf, (int) *count, (MPI_Datatype) x_type, (int) *dest, (int) *tag, (MPI_Comm) *comm);
}

//   
//   MPI_BUFFER_ATTACH( buffer_addr, size)
//   [ IN buffer] initial buffer address (choice)
//   [ IN size] buffer size, in bytes (integer)
//   
//   int MPI_Buffer_attach( void* buffer, int size)
//   
//   MPI_BUFFER_ATTACH( BUFFER, SIZE, IERROR)
//   <type> BUFFER(*)
//   INTEGER SIZE, IERROR
//   
void
mpi_buffer_attach_(
void         *buffer_addr,
fortran_int  *size,
fortran_int  *ierr)
{
#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Buffer_attach...\n");
#endif
        *ierr = (fortran_int) PMPI_Buffer_attach( (void *) buffer_addr, (int) *size);
}

//   
//   MPI_BUFFER_DETACH( buffer_addr, size)
//   [ OUT buffer_addr] initial buffer address (choice)
//   [ OUT size] buffer size, in bytes (integer)
//   
//   int MPI_Buffer_detach( void* buffer_addr, int* size)
//   
//   MPI_BUFFER_DETACH( BUFFER_ADDR, SIZE, IERROR)
//   <type> BUFFER_ADDR(*)
//   INTEGER SIZE, IERROR
//   
void
mpi_buffer_detach_(
void         *buffer_addr,
fortran_int  *size,
fortran_int  *ierr)
{
int x_size;
#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Buffer_detach...\n");
#endif
        *ierr = (fortran_int) PMPI_Buffer_detach( (void *) buffer_addr, &x_size);
        *size = (fortran_int) x_size ;
}

//   
//   MPI_ISEND(buf, count, datatype, dest, tag, comm, request)
//   [ IN buf] initial address of send buffer (choice)
//   [ IN count] number of elements in send buffer (integer)
//   [ IN datatype] datatype of each send buffer element (handle)
//   [ IN dest] rank of destination (integer)
//   [ IN tag] message tag (integer)
//   [ IN comm] communicator (handle)
//   [ OUT request] communication request (handle)
//   
//   int MPI_Isend(void* buf, int count, MPI_Datatype datatype, int dest,
//   int tag, MPI_Comm comm, MPI_Request *request)
//   
//   MPI_ISEND(BUF, COUNT, DATATYPE, DEST, TAG, COMM, REQUEST, IERROR)
//   <type> BUF(*)
//   INTEGER COUNT, DATATYPE, DEST, TAG, COMM, REQUEST, IERROR
//   
void
mpi_isend_(
void         *buf,
fortran_int  *count,
fortran_int  *type,
fortran_int  *dest,
fortran_int  *tag,
fortran_int  *comm,
fortran_int  *request,
fortran_int  *ierr)
{
MPI_Datatype x_type = convert_type(*type);
MPI_Request x_request;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Isend with *type = %i, x_type = %i\n",(int) *type, x_type);
#endif
        *ierr = (fortran_int) PMPI_Isend( (void *) buf, (int) *count, (MPI_Datatype) x_type, (int) *dest, (int) *tag, (MPI_Comm) *comm, (MPI_Request * ) &x_request);

        *request = (fortran_int) x_request;
#ifdef DEBUG
  printf("wrapper.c... returning from PMPI_Isend with *request = %i, x_request = %u\n", (int) *request, x_request);
#endif
}

//   
//   MPI_IBSEND(buf, count, datatype, dest, tag, comm, request)
//   [ IN buf] initial address of send buffer (choice)
//   [ IN count] number of elements in send buffer (integer)
//   [ IN datatype] datatype of each send buffer element (handle)
//   [ IN dest] rank of destination (integer)
//   [ IN tag] message tag (integer)
//   [ IN comm] communicator (handle)
//   [ OUT request] communication request (handle)
//   
//   int MPI_Ibsend(void* buf, int count, MPI_Datatype datatype, int dest,
//   int tag, MPI_Comm comm, MPI_Request *request)
//   
//   MPI_IBSEND(BUF, COUNT, DATATYPE, DEST, TAG, COMM, REQUEST, IERROR)
//   <type> BUF(*)
//   INTEGER COUNT, DATATYPE, DEST, TAG, COMM, REQUEST, IERROR
//   
void
mpi_ibsend_(
void         *buf,
fortran_int  *count,
fortran_int  *type,
fortran_int  *dest,
fortran_int  *tag,
fortran_int  *comm,
fortran_int  *request,
fortran_int  *ierr)
{
MPI_Datatype x_type = convert_type(*type);
MPI_Request x_request;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Ibsend with *type = %i, x_type = %i\n",(int) *type, x_type);
#endif
        *ierr = (fortran_int) PMPI_Ibsend( (void *) buf, (int) *count, (MPI_Datatype) x_type, (int) *dest, (int) *tag, (MPI_Comm) *comm, (MPI_Request * ) &x_request);

        *request = (fortran_int) x_request;
#ifdef DEBUG
  printf("wrapper.c... returning from PMPI_Ibsend with *request = %i, x_request = %u\n", (int) *request, x_request);
#endif
}

//   
//   MPI_ISSEND(buf, count, datatype, dest, tag, comm, request)
//   [ IN buf] initial address of send buffer (choice)
//   [ IN count] number of elements in send buffer (integer)
//   [ IN datatype] datatype of each send buffer element (handle)
//   [ IN dest] rank of destination (integer)
//   [ IN tag] message tag (integer)
//   [ IN comm] communicator (handle)
//   [ OUT request] communication request (handle)
//   
//   int MPI_Issend(void* buf, int count, MPI_Datatype datatype, int dest,
//   int tag, MPI_Comm comm, MPI_Request *request)
//   
//   MPI_ISSEND(BUF, COUNT, DATATYPE, DEST, TAG, COMM, REQUEST, IERROR)
//   <type> BUF(*)
//   INTEGER COUNT, DATATYPE, DEST, TAG, COMM, REQUEST, IERROR
//   
void
mpi_issend_(
void         *buf,
fortran_int  *count,
fortran_int  *type,
fortran_int  *dest,
fortran_int  *tag,
fortran_int  *comm,
fortran_int  *request,
fortran_int  *ierr)
{
MPI_Datatype x_type = convert_type(*type);
MPI_Request x_request;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Issend with *type = %i, x_type = %i\n",(int) *type, x_type);
#endif
        *ierr = (fortran_int) PMPI_Issend( (void *) buf, (int) *count, (MPI_Datatype) x_type, (int) *dest, (int) *tag, (MPI_Comm) *comm, (MPI_Request * ) &x_request);

        *request = (fortran_int) x_request;
#ifdef DEBUG
  printf("wrapper.c... returning from PMPI_Issend with *request = %i, x_request = %u\n", (int) *request, x_request);
#endif
}

//   
//   MPI_IRSEND(buf, count, datatype, dest, tag, comm, request)
//   [ IN buf] initial address of send buffer (choice)
//   [ IN count] number of elements in send buffer (integer)
//   [ IN datatype] datatype of each send buffer element (handle)
//   [ IN dest] rank of destination (integer)
//   [ IN tag] message tag (integer)
//   [ IN comm] communicator (handle)
//   [ OUT request] communication request (handle)
//   
//   int MPI_Irsend(void* buf, int count, MPI_Datatype datatype, int dest,
//   int tag, MPI_Comm comm, MPI_Request *request)
//   
//   MPI_IRSEND(BUF, COUNT, DATATYPE, DEST, TAG, COMM, REQUEST, IERROR)
//   <type> BUF(*)
//   INTEGER COUNT, DATATYPE, DEST, TAG, COMM, REQUEST, IERROR
//   
void
mpi_irsend_(
void         *buf,
fortran_int  *count,
fortran_int  *type,
fortran_int  *dest,
fortran_int  *tag,
fortran_int  *comm,
fortran_int  *request,
fortran_int  *ierr)
{
MPI_Datatype x_type = convert_type(*type);
MPI_Request x_request;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Irsend with *type = %i, x_type = %i\n",(int) *type, x_type);
#endif
        *ierr = (fortran_int) PMPI_Irsend( (void *) buf, (int) *count, (MPI_Datatype) x_type, (int) *dest, (int) *tag, (MPI_Comm) *comm, (MPI_Request * ) &x_request);

        *request = (fortran_int) x_request;
#ifdef DEBUG
  printf("wrapper.c... returning from PMPI_Irsend with *request = %i, x_request = %u\n", (int) *request, x_request);
#endif
}

//   
//   MPI_IRECV (buf, count, datatype, source, tag, comm, request)
//   [ OUT buf] initial address of receive buffer (choice)
//   [ IN count] number of elements in receive buffer (integer)
//   [ IN datatype] datatype of each receive buffer element (handle)
//   [ IN source] rank of source (integer)
//   [ IN tag] message tag (integer)
//   [ IN comm] communicator (handle)
//   [ OUT request] communication request (handle)
//   
//   int MPI_Irecv(void* buf, int count, MPI_Datatype datatype, int source,
//   int tag, MPI_Comm comm, MPI_Request *request)
//   
//   MPI_IRECV(BUF, COUNT, DATATYPE, SOURCE, TAG, COMM, REQUEST, IERROR)
//   <type> BUF(*)
//   INTEGER COUNT, DATATYPE, SOURCE, TAG, COMM, REQUEST, IERROR
//   
void
mpi_irecv_(
void         *buf,
fortran_int  *count,
fortran_int  *type,
fortran_int  *source,
fortran_int  *tag,
fortran_int  *comm,
fortran_int  *request,
fortran_int  *ierr)
{
MPI_Datatype x_type = convert_type(*type);
MPI_Request  x_request;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Irecv with *type = %i, x_type = %i\n",(int) *type, x_type);
#endif
        *ierr = (fortran_int) PMPI_Irecv( (void *) buf, (int) *count, (MPI_Datatype) x_type, (int) *source, (int) *tag, (MPI_Comm) *comm, &x_request);

        *request = (fortran_int) x_request;
#ifdef DEBUG
  printf("wrapper.c... returning from PMPI_Irecv with *request = %i, x_request = %u\n", (int) *request, x_request);
#endif
}

//   
//   MPI_WAIT(request, status)
//   [ INOUT request] request (handle)
//   [ OUT status] status object (Status)
//   
//   int MPI_Wait(MPI_Request *request, MPI_Status *status)
//   
//   MPI_WAIT(REQUEST, STATUS, IERROR)
//   INTEGER REQUEST, STATUS(MPI_STATUS_SIZE), IERROR
//   
void
mpi_wait_(
fortran_int *request,
fortran_MPI_Status  *status,
fortran_int  *ierr)
{
MPI_Status x_status;
MPI_Request x_request = *request;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Wait\n");
  printf("wrapper.c... calling PMPI_Wait with request = %i\n",(int) *request);
#endif
        *ierr = (fortran_int) PMPI_Wait( (MPI_Request *) &x_request, &x_status);

        *request = (fortran_int) x_request;
        copy_status_struct(status, &x_status, MPI_WRAPPER_TO_FORTRAN);
}

//
//   MPI_TEST(request, flag, status)
//   [ INOUT request] communication request (handle)
//   [ OUT flag] true if operation completed (logical)
//   [ OUT status] status object (Status)
//   
//   int MPI_Test(MPI_Request *request, int *flag, MPI_Status *status)
//   
//   MPI_TEST(REQUEST, FLAG, STATUS, IERROR)
//   LOGICAL FLAG
//   INTEGER REQUEST, STATUS(MPI_STATUS_SIZE), IERROR
//   
void
mpi_test_(
fortran_int *request,
fortran_int *flag,
fortran_MPI_Status  *status,
fortran_int  *ierr)
{
MPI_Request x_request = *request;
int x_flag;
MPI_Status x_status;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Test\n");
  printf("wrapper.c... calling PMPI_Test with request = %i\n",(int) *request);
#endif
        *ierr = (fortran_int) PMPI_Test( &x_request, &x_flag, &x_status);

        copy_status_struct(status, &x_status, MPI_WRAPPER_TO_FORTRAN);
        *flag = (fortran_int) x_flag ;
        *request = (fortran_int) x_request ;
}

//   
//   MPI_REQUEST_FREE(request)
//   [ INOUT request] communication request (handle)
//   
//   int MPI_Request_free(MPI_Request *request)
//   
//   MPI_REQUEST_FREE(REQUEST, IERROR)
//   INTEGER REQUEST, IERROR
//   
void
mpi_request_free_(
fortran_int *request,
fortran_int  *ierr)
{
MPI_Request x_request = *request;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Request_free\n");
#endif
        *ierr = (fortran_int) PMPI_Request_free( &x_request);

        *request = (fortran_int) x_request ;
}

//   
//   MPI_WAITANY (count, array_of_requests, index, status)
//   [ IN count] list length (integer)
//   [ INOUT array_of_requests] array of requests (array of handles)
//   [ OUT index] index of handle for operation that completed (integer)
//   [ OUT status] status object (Status)
//   
//   int MPI_Waitany(int count, MPI_Request *array_of_requests, int *index,
//   MPI_Status *status)
//   
//   MPI_WAITANY(COUNT, ARRAY_OF_REQUESTS, INDEX, STATUS, IERROR)
//   INTEGER COUNT, ARRAY_OF_REQUESTS(*), INDEX, STATUS(MPI_STATUS_SIZE),
//   IERROR
//   
void
mpi_waitany_(
fortran_int  *count,
fortran_int  *request,
fortran_int  *index,
fortran_MPI_Status  *status,
fortran_int  *ierr)
{
MPI_Status x_status[*count];
MPI_Request x_request[*count];
int x_index;
int i;

	for (i=0;i<*count;i++) {
          x_request[i] = request[i];
#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Waitany with x_request[%i] = %i\n",i,x_request[i]);
#endif
	}

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Waitany\n");
  printf("wrapper.c... calling PMPI_Waitany with count = %i\n",(int) *count);
#endif
        *ierr = (fortran_int) PMPI_Waitany( (int) *count, x_request, &x_index, x_status);

	for (i=0;i<*count;i++) {
          copy_status_struct(&status[i], &x_status[i], MPI_WRAPPER_TO_FORTRAN);
	}

	for (i=0;i<*count;i++) {
          request[i] = x_request[i];
	}

	*index = (fortran_int) x_index;
}

//   
//   MPI_TESTANY(count, array_of_requests, index, flag, status)
//   [ IN count] list length (integer)
//   [ INOUT array_of_requests] array of requests (array of handles)
//   [ OUT index] index of operation that completed, or MPI_UNDEFINED if
//   none completed (integer)
//   [ OUT flag] true if one of the operations is complete (logical)
//   [ OUT status] status object (Status)
//   
//   int MPI_Testany(int count, MPI_Request *array_of_requests, int *index,
//   int *flag, MPI_Status *status)
//   
//   MPI_TESTANY(COUNT, ARRAY_OF_REQUESTS, INDEX, FLAG, STATUS, IERROR)
//   LOGICAL FLAG
//   INTEGER COUNT, ARRAY_OF_REQUESTS(*), INDEX, STATUS(MPI_STATUS_SIZE),
//   IERROR
//   
void
mpi_testany_(
fortran_int  *count,
fortran_int  *request,
fortran_int  *index,
fortran_int  *flag,
fortran_MPI_Status  *status,
fortran_int  *ierr)
{
MPI_Status x_status[*count];

MPI_Request x_request[*count];
int x_flag;
int x_index;
int i;

	for (i=0;i<*count;i++) {
          x_request[i] = request[i];
#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Testany with x_request[%i] = %i\n",i,x_request[i]);
#endif
	}

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Testany\n");
  printf("wrapper.c... calling PMPI_Testany with count = %i\n",(int) *count);
#endif
        *ierr = (fortran_int) PMPI_Testany( (int) *count, x_request, &x_index, &x_flag, x_status);

	for (i=0;i<*count;i++) {
          copy_status_struct(&status[i], &x_status[i], MPI_WRAPPER_TO_FORTRAN);
	}

	for (i=0;i<*count;i++) {
          request[i] = x_request[i];
#ifdef DEBUG
  printf("wrapper.c... returning from PMPI_Testany with request[%i] = %i\n",i, (int) request[i]);
#endif
	}

	*index = (fortran_int) x_index;
}

//   
//   MPI_WAITALL( count, array_of_requests, array_of_statuses)
//   [ IN count] lists length (integer)
//   [ INOUT array_of_requests] array of requests (array of handles)
//   [ OUT array_of_statuses] array of status objects (array of Status)
//   
//   int MPI_Waitall(int count, MPI_Request *array_of_requests, MPI_Status
//   *array_of_statuses)
//   
//   MPI_WAITALL(COUNT, ARRAY_OF_REQUESTS, ARRAY_OF_STATUSES, IERROR)
//   INTEGER COUNT, ARRAY_OF_REQUESTS(*)
//   INTEGER ARRAY_OF_STATUSES(MPI_STATUS_SIZE,*), IERROR
//   
void
mpi_waitall_(
fortran_int  *count,
fortran_int  *request,
fortran_MPI_Status  *status,
fortran_int  *ierr)
{
MPI_Status x_status[*count];
MPI_Request x_request[*count];
int i;

	for (i=0;i<*count;i++) {
          x_request[i] = request[i];
#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Waitall with x_request[%i] = %i\n",i,x_request[i]);
#endif
	}

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Waitall\n");
  printf("wrapper.c... calling PMPI_Waitall with count = %i\n",(int) *count);
#endif
        *ierr = (fortran_int) PMPI_Waitall( (int) *count, x_request, x_status);

	for (i=0;i<*count;i++) {
          copy_status_struct(&status[i], &x_status[i], MPI_WRAPPER_TO_FORTRAN);
	}

	for (i=0;i<*count;i++) {
          request[i] = x_request[i];
#ifdef DEBUG
  printf("wrapper.c... returning from PMPI_Waitall with request[%i] = %i\n",i, (int) request[i]);
#endif
	}
}

//   
//   MPI_TESTALL(count, array_of_requests, flag, array_of_statuses)
//   [ IN count] lists length (integer)
//   [ INOUT array_of_requests] array of requests (array of handles)
//   [ OUT flag] (logical)
//   [ OUT array_of_statuses] array of status objects (array of Status)
//   
//   int MPI_Testall(int count, MPI_Request *array_of_requests, int *flag,
//   MPI_Status *array_of_statuses)
//   
//   MPI_TESTALL(COUNT, ARRAY_OF_REQUESTS, FLAG, ARRAY_OF_STATUSES, IERROR)
//   LOGICAL FLAG
//   INTEGER COUNT, ARRAY_OF_REQUESTS(*),
//   ARRAY_OF_STATUSES(MPI_STATUS_SIZE,*), IERROR
//   
void
mpi_testall_(
fortran_int  *count,
fortran_int  *flag,
fortran_int  *request,
fortran_MPI_Status  *status,
fortran_int  *ierr)
{
MPI_Status x_status[*count];
MPI_Request x_request[*count];
int x_flag;
int i;

	for (i=0;i<*count;i++) {
          x_request[i] = request[i];
#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Testall with x_request[%i] = %i\n",i,x_request[i]);
#endif
	}

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Testall\n");
  printf("wrapper.c... calling PMPI_Testall with count = %i\n",(int) *count);
#endif
        *ierr = (fortran_int) PMPI_Testall( (int) *count, x_request, &x_flag , x_status);

	for (i=0;i<*count;i++) {
          copy_status_struct(&status[i], &x_status[i], MPI_WRAPPER_TO_FORTRAN);
	}
        *flag = (fortran_int) x_flag ;
}

//   
//   MPI_WAITSOME(incount, array_of_requests, outcount, array_of_indices,
//   array_of_statuses)
//   [ IN incount] length of array_of_requests (integer)
//   [ INOUT array_of_requests] array of requests (array of handles)
//   [ OUT outcount] number of completed requests (integer)
//   [ OUT array_of_indices] array of indices of operations that completed
//   (array of integers)
//   [ OUT array_of_statuses] array of status objects for operations that
//   completed (array of Status)
//   
//   int MPI_Waitsome(int incount, MPI_Request *array_of_requests, int
//   *outcount, int *array_of_indices, MPI_Status *array_of_statuses)
//   
//   MPI_WAITSOME(INCOUNT, ARRAY_OF_REQUESTS, OUTCOUNT, ARRAY_OF_INDICES,
//   ARRAY_OF_STATUSES, IERROR)
//   INTEGER INCOUNT, ARRAY_OF_REQUESTS(*), OUTCOUNT, ARRAY_OF_INDICES(*),
//   ARRAY_OF_STATUSES(MPI_STATUS_SIZE,*), IERROR
//   
void
mpi_waitsome_(
fortran_int *incount,
fortran_int array_of_requests[],
fortran_int *outcount,
fortran_int array_of_indices[],
fortran_MPI_Status array_of_statuses[],
fortran_int *ierr)
{
int x_outcount;
MPI_Request x_array_of_requests[ (int) *incount ] ;
int x_array_of_indices[ (int) *incount ] ;
MPI_Status x_array_of_statuses[ (int) *incount ];
int i;

	for (i=0;i<*incount;i++) {
	  x_array_of_requests[i] = array_of_requests[i];
	}

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Waitsome\n");
#endif
	*ierr = (fortran_int) PMPI_Waitsome( (int) *incount, x_array_of_requests, &x_outcount, x_array_of_indices, x_array_of_statuses ) ;

	for (i=0;i<*incount;i++) {
	  array_of_requests[i] = (fortran_int) x_array_of_requests[i];
	}

	*outcount = (fortran_int) x_outcount;
	for (i=0;i<*outcount;i++) {
	  array_of_indices[i] = (fortran_int) x_array_of_indices[i];
	}

	for (i=0;i<*outcount;i++) {
          copy_status_struct(&array_of_statuses[i], &x_array_of_statuses[i], MPI_WRAPPER_TO_FORTRAN);
	}

}

//   
//   MPI_TESTSOME(incount, array_of_requests, outcount, array_of_indices,
//   array_of_statuses)
//   [ IN incount] length of array_of_requests (integer)
//   [ INOUT array_of_requests] array of requests (array of handles)
//   [ OUT outcount] number of completed requests (integer)
//   [ OUT array_of_indices] array of indices of operations that completed
//   (array of integers)
//   [ OUT array_of_statuses] array of status objects for operations that
//   completed (array of Status)
//   
//   int MPI_Testsome(int incount, MPI_Request *array_of_requests, int
//   *outcount, int *array_of_indices, MPI_Status *array_of_statuses)
//   
//   MPI_TESTSOME(INCOUNT, ARRAY_OF_REQUESTS, OUTCOUNT, ARRAY_OF_INDICES,
//   ARRAY_OF_STATUSES, IERROR)
//   INTEGER INCOUNT, ARRAY_OF_REQUESTS(*), OUTCOUNT, ARRAY_OF_INDICES(*),
//   ARRAY_OF_STATUSES(MPI_STATUS_SIZE,*), IERROR
//   
void
mpi_testsome_(
fortran_int *incount,
fortran_int array_of_requests[],
fortran_int *outcount,
fortran_int array_of_indices[],
fortran_MPI_Status array_of_statuses[],
fortran_int *ierr)
{
int x_outcount;
MPI_Request x_array_of_requests[ (int) *incount ] ;
int x_array_of_indices[ (int) *incount ] ;
MPI_Status x_array_of_statuses[ (int) *incount ];
int i;

	for (i=0;i<*incount;i++) {
	  x_array_of_requests[i] = array_of_requests[i];
	}

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Testsome\n");
#endif
	*ierr = (fortran_int) PMPI_Testsome( (int) *incount, x_array_of_requests, &x_outcount, x_array_of_indices, x_array_of_statuses ) ;

	for (i=0;i<*incount;i++) {
	  array_of_requests[i] = (fortran_int) x_array_of_requests[i];
	}

	*outcount = (fortran_int) x_outcount;
	for (i=0;i<*outcount;i++) {
	  array_of_indices[i] = (fortran_int) x_array_of_indices[i];
	}

	for (i=0;i<*outcount;i++) {
          copy_status_struct(&array_of_statuses[i], &x_array_of_statuses[i], MPI_WRAPPER_TO_FORTRAN);
	}

}

//   
//   MPI_IPROBE(source, tag, comm, flag, status)
//   [ IN source] source rank, or MPI_ANY_SOURCE (integer)
//   [ IN tag] tag value or MPI_ANY_TAG (integer)
//   [ IN comm] communicator (handle)
//   [ OUT flag] (logical)
//   [ OUT status] status object (Status)
//   
//   int MPI_Iprobe(int source, int tag, MPI_Comm comm, int *flag,
//   MPI_Status *status)
//   
//   MPI_IPROBE(SOURCE, TAG, COMM, FLAG, STATUS, IERROR)
//   LOGICAL FLAG
//   INTEGER SOURCE, TAG, COMM, STATUS(MPI_STATUS_SIZE), IERROR
//   
void
mpi_iprobe_(
fortran_int  *source,
fortran_int  *tag,
fortran_int  *comm,
fortran_int  *flag,
fortran_MPI_Status  *status,
fortran_int  *ierr)
{
MPI_Status x_status;
int x_flag ;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Iprobe...\n");
#endif
        *ierr = (fortran_int) PMPI_Iprobe( (int) *source, (int) *tag, (MPI_Comm) *comm, &x_flag, &x_status);

        copy_status_struct(status, &x_status, MPI_WRAPPER_TO_FORTRAN);
        *flag = (fortran_int) x_flag ;
}

//   
//   MPI_PROBE(source, tag, comm, status)
//   [ IN source] source rank, or MPI_ANY_SOURCE (integer)
//   [ IN tag] tag value, or MPI_ANY_TAG (integer)
//   [ IN comm] communicator (handle)
//   [ OUT status] status object (Status)
//   
//   int MPI_Probe(int source, int tag, MPI_Comm comm, MPI_Status *status)
//   
//   MPI_PROBE(SOURCE, TAG, COMM, STATUS, IERROR)
//   INTEGER SOURCE, TAG, COMM, STATUS(MPI_STATUS_SIZE), IERROR
//   
void
mpi_probe_(
fortran_int  *source,
fortran_int  *tag,
fortran_int  *comm,
fortran_MPI_Status  *status,
fortran_int  *ierr)
{
MPI_Status x_status;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Probe...\n");
#endif
        *ierr = (fortran_int) PMPI_Probe( (int) *source, (int) *tag, (MPI_Comm) *comm, &x_status);

        copy_status_struct(status, &x_status, MPI_WRAPPER_TO_FORTRAN);
}

//   
//   MPI_CANCEL(request)
//   [ IN request] communication request (handle)
//   
//   int MPI_Cancel(MPI_Request *request)
//   
//   MPI_CANCEL(REQUEST, IERROR)
//   INTEGER REQUEST, IERROR
//   
void
mpi_cancel_(
fortran_int *request,
fortran_int  *ierr)
{

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Cancel\n");
#endif
        *ierr = (fortran_int) PMPI_Cancel( (MPI_Request *) request );
}

//   
//   MPI_TEST_CANCELLED(status, flag)
//   [ IN status] status object (Status)
//   [ OUT flag] (logical)
//   
//   int MPI_Test_cancelled(MPI_Status *status, int *flag)
//   
//   MPI_TEST_CANCELLED(STATUS, FLAG, IERROR)
//   LOGICAL FLAG
//   INTEGER STATUS(MPI_STATUS_SIZE), IERROR
//   
void
mpi_test_cancelled_(
fortran_MPI_Status  *status,
fortran_int  *flag,
fortran_int  *ierr)
{
int x_flag ;

MPI_Status x_status;
        copy_status_struct(status, &x_status, FORTRAN_TO_MPI_WRAPPER);

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Test_cancelled...\n");
#endif
        *ierr = (fortran_int) PMPI_Test_cancelled( &x_status, &x_flag);

        copy_status_struct(status, &x_status, MPI_WRAPPER_TO_FORTRAN);
        *flag = (fortran_int) x_flag ;
}

//   
//   MPI_SEND_INIT(buf, count, datatype, dest, tag, comm, request)
//   [ IN buf] initial address of send buffer (choice)
//   [ IN count] number of elements sent (integer)
//   [ IN datatype] type of each element (handle)
//   [ IN dest] rank of destination (integer)
//   [ IN tag] message tag (integer)
//   [ IN comm] communicator (handle)
//   [ OUT request] communication request (handle)
//   
//   int MPI_Send_init(void* buf, int count, MPI_Datatype datatype, int
//   dest, int tag, MPI_Comm comm, MPI_Request *request)
//   
//   MPI_SEND_INIT(BUF, COUNT, DATATYPE, DEST, TAG, COMM, REQUEST, IERROR)
//   <type> BUF(*)
//   INTEGER REQUEST, COUNT, DATATYPE, DEST, TAG, COMM, REQUEST, IERROR
//   
void
mpi_send_init_(
void         *buf,
fortran_int  *count,
fortran_int  *type,
fortran_int  *dest,
fortran_int  *tag,
fortran_int  *comm,
fortran_int  *request,
fortran_int  *ierr)
{
MPI_Datatype x_type = convert_type(*type);
MPI_Request x_request;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_send_init with *type = %i, x_type = %i\n",(int) *type, x_type);
#endif
        *ierr = (fortran_int) PMPI_Send_init( (void *) buf, (int) *count, (MPI_Datatype) x_type, (int) *dest, (int) *tag, (MPI_Comm) *comm, (MPI_Request * ) &x_request);

        *request = (fortran_int) x_request;
#ifdef DEBUG
  printf("wrapper.c... returning from PMPI_Send_init with *request = %i, x_request = %u\n", (int) *request, x_request);
#endif
}

//   
//   MPI_BSEND_INIT(buf, count, datatype, dest, tag, comm, request)
//   [ IN buf] initial address of send buffer (choice)
//   [ IN count] number of elements sent (integer)
//   [ IN datatype] type of each element (handle)
//   [ IN dest] rank of destination (integer)
//   [ IN tag] message tag (integer)
//   [ IN comm] communicator (handle)
//   [ OUT request] communication request (handle)
//   
//   int MPI_Bsend_init(void* buf, int count, MPI_Datatype datatype, int
//   dest, int tag, MPI_Comm comm, MPI_Request *request)
//   
//   MPI_BSEND_INIT(BUF, COUNT, DATATYPE, DEST, TAG, COMM, REQUEST, IERROR)
//   <type> BUF(*)
//   INTEGER REQUEST, COUNT, DATATYPE, DEST, TAG, COMM, REQUEST, IERROR
//   
void
mpi_bsend_init_(
void         *buf,
fortran_int  *count,
fortran_int  *type,
fortran_int  *dest,
fortran_int  *tag,
fortran_int  *comm,
fortran_int  *request,
fortran_int  *ierr)
{
MPI_Datatype x_type = convert_type(*type);
MPI_Request x_request;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_bsend_init with *type = %i, x_type = %i\n",(int) *type, x_type);
#endif
        *ierr = (fortran_int) PMPI_Bsend_init( (void *) buf, (int) *count, (MPI_Datatype) x_type, (int) *dest, (int) *tag, (MPI_Comm) *comm, (MPI_Request * ) &x_request);

        *request = (fortran_int) x_request;
#ifdef DEBUG
  printf("wrapper.c... returning from PMPI_Bsend_init with *request = %i, x_request = %u\n", (int) *request, x_request);
#endif
}

//   
//   MPI_SSEND_INIT(buf, count, datatype, dest, tag, comm, request)
//   [ IN buf] initial address of send buffer (choice)
//   [ IN count] number of elements sent (integer)
//   [ IN datatype] type of each element (handle)
//   [ IN dest] rank of destination (integer)
//   [ IN tag] message tag (integer)
//   [ IN comm] communicator (handle)
//   [ OUT request] communication request (handle)
//   
//   int MPI_Ssend_init(void* buf, int count, MPI_Datatype datatype, int
//   dest, int tag, MPI_Comm comm, MPI_Request *request)
//   
//   MPI_SSEND_INIT(BUF, COUNT, DATATYPE, DEST, TAG, COMM, REQUEST, IERROR)
//   <type> BUF(*)
//   INTEGER COUNT, DATATYPE, DEST, TAG, COMM, REQUEST, IERROR
//
void
mpi_ssend_init_(
void         *buf,
fortran_int  *count,
fortran_int  *type,
fortran_int  *dest,
fortran_int  *tag,
fortran_int  *comm,
fortran_int  *request,
fortran_int  *ierr)
{
MPI_Datatype x_type = convert_type(*type);
MPI_Request x_request;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_ssend_init with *type = %i, x_type = %i\n",(int) *type, x_type);
#endif
        *ierr = (fortran_int) PMPI_Ssend_init( (void *) buf, (int) *count, (MPI_Datatype) x_type, (int) *dest, (int) *tag, (MPI_Comm) *comm, (MPI_Request * ) &x_request);

        *request = (fortran_int) x_request;
#ifdef DEBUG
  printf("wrapper.c... returning from PMPI_Ssend_init with *request = %i, x_request = %u\n", (int) *request, x_request);
#endif
}

//   
//   MPI_RSEND_INIT(buf, count, datatype, dest, tag, comm, request)
//   [ IN buf] initial address of send buffer (choice)
//   [ IN count] number of elements sent (integer)
//   [ IN datatype] type of each element (handle)
//   [ IN dest] rank of destination (integer)
//   [ IN tag] message tag (integer)
//   [ IN comm] communicator (handle)
//   [ OUT request] communication request (handle)
//   
//   int MPI_Rsend_init(void* buf, int count, MPI_Datatype datatype, int
//   dest, int tag, MPI_Comm comm, MPI_Request *request)
//   
//   MPI_RSEND_INIT(BUF, COUNT, DATATYPE, DEST, TAG, COMM, REQUEST, IERROR)
//   <type> BUF(*)
//   INTEGER COUNT, DATATYPE, DEST, TAG, COMM, REQUEST, IERROR
//
void
mpi_rsend_init_(
void         *buf,
fortran_int  *count,
fortran_int  *type,
fortran_int  *dest,
fortran_int  *tag,
fortran_int  *comm,
fortran_int  *request,
fortran_int  *ierr)
{
MPI_Datatype x_type = convert_type(*type);
MPI_Request x_request;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Rsend_init with *type = %i, x_type = %i\n",(int) *type, x_type);
#endif
        *ierr = (fortran_int) PMPI_Rsend_init( (void *) buf, (int) *count, (MPI_Datatype) x_type, (int) *dest, (int) *tag, (MPI_Comm) *comm, (MPI_Request * ) &x_request);

        *request = (fortran_int) x_request;
#ifdef DEBUG
  printf("wrapper.c... returning from PMPI_Rsend_init with *request = %i, x_request = %u\n", (int) *request, x_request);
#endif
}

//   
//   MPI_RECV_INIT(buf, count, datatype, source, tag, comm, request)
//   [ OUT buf] initial address of receive buffer (choice)
//   [ IN count] number of elements received (integer)
//   [ IN datatype] type of each element (handle)
//   [ IN source] rank of source or MPI_ANY_SOURCE (integer)
//   [ IN tag] message tag or MPI_ANY_TAG (integer)
//   [ IN comm] communicator (handle)
//   [ OUT request] communication request (handle)
//   
//   int MPI_Recv_init(void* buf, int count, MPI_Datatype datatype, int
//   source, int tag, MPI_Comm comm, MPI_Request *request)
//   
//   MPI_RECV_INIT(BUF, COUNT, DATATYPE, SOURCE, TAG, COMM, REQUEST,
//   IERROR)
//   <type> BUF(*)
//   INTEGER COUNT, DATATYPE, SOURCE, TAG, COMM, REQUEST, IERROR
//   
void
mpi_recv_init_(
void         *buf,
fortran_int  *count,
fortran_int  *type,
fortran_int  *source,
fortran_int  *tag,
fortran_int  *comm,
fortran_int  *request,
fortran_int  *ierr)
{
MPI_Datatype x_type = convert_type(*type);
MPI_Request  x_request;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Recv_init with *type = %i, x_type = %i\n",(int) *type, x_type);
#endif
        *ierr = (fortran_int) PMPI_Recv_init( (void *) buf, (int) *count, (MPI_Datatype) x_type, (int) *source, (int) *tag, (MPI_Comm) *comm, &x_request);

        *request = (fortran_int) x_request;
#ifdef DEBUG
  printf("wrapper.c... returning from PMPI_Recv_init with *request = %i, x_request = %u\n", (int) *request, x_request);
#endif
}

//   
//   MPI_START(request)
//   [ INOUT request] communication request (handle)
//   
//   int MPI_Start(MPI_Request *request)
//   
//   MPI_START(REQUEST, IERROR)
//   INTEGER REQUEST, IERROR
//   
void
mpi_start_(
fortran_int  *request,
fortran_int  *ierr)
{
MPI_Request x_request;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Start...\n");
#endif
        *ierr = (fortran_int) PMPI_Start( (MPI_Request * ) &x_request);

        *request = (fortran_int) x_request;
#ifdef DEBUG
  printf("wrapper.c... returning from PMPI_Start with *request = %i, x_request = %u\n", (int) *request, x_request);
#endif
}

//   
//   MPI_STARTALL(count, array_of_requests)
//   [ IN count] list length (integer)
//   [ INOUT array_of_requests] array of requests (array of handle)
//   
//   int MPI_Startall(int count, MPI_Request *array_of_requests)
//   
//   MPI_STARTALL(COUNT, ARRAY_OF_REQUESTS, IERROR)
//   INTEGER COUNT, ARRAY_OF_REQUESTS(*), IERROR
//   
void
mpi_startall_(
fortran_int  *count,
fortran_int  *request,
fortran_int  *ierr)
{
MPI_Request x_request[*count];
int i;

	for (i=0;i<*count;i++) {
          x_request[i] = request[i];
#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Startall with x_request[%i] = %i\n",i,x_request[i]);
#endif
	}

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Startall\n");
  printf("wrapper.c... calling PMPI_Startall with count = %i\n",(int) *count);
#endif
        *ierr = (fortran_int) PMPI_Startall( (int) *count, x_request );
}

//   
//   MPI_SENDRECV(sendbuf, sendcount, sendtype, dest, sendtag, recvbuf,
//   recvcount, recvtype, source, recvtag, comm, status)
//   [ IN sendbuf] initial address of send buffer (choice)
//   [ IN sendcount] number of elements in send buffer (integer)
//   [ IN sendtype] type of elements in send buffer (handle)
//   [ IN dest] rank of destination (integer)
//   [ IN sendtag] send tag (integer)
//   [ OUT recvbuf] initial address of receive buffer (choice)
//   [ IN recvcount] number of elements in receive buffer (integer)
//   [ IN recvtype] type of elements in receive buffer (handle)
//   [ IN source] rank of source (integer)
//   [ IN recvtag] receive tag (integer)
//   [ IN comm] communicator (handle)
//   [ OUT status] status object (Status)
//   
//   int MPI_Sendrecv(void *sendbuf, int sendcount, MPI_Datatype sendtype,
//   int dest, int sendtag, void *recvbuf, int recvcount, MPI_Datatype
//   recvtype, int source, MPI_Datatype recvtag, MPI_Comm comm, MPI_Status
//   *status)
//   
//   MPI_SENDRECV(SENDBUF, SENDCOUNT, SENDTYPE, DEST, SENDTAG, RECVBUF,
//   RECVCOUNT, RECVTYPE, SOURCE, RECVTAG, COMM, STATUS, IERROR)
//   <type> SENDBUF(*), RECVBUF(*)
//   INTEGER SENDCOUNT, SENDTYPE, DEST, SENDTAG, RECVCOUNT, RECVTYPE,
//   SOURCE, RECV TAG, COMM, STATUS(MPI_STATUS_SIZE), IERROR
//   
void
mpi_sendrecv_(
void         *sendbuf,
fortran_int  *sendcount,
fortran_int  *sendtype,
fortran_int  *dest,
fortran_int  *sendtag,
void         *recvbuf,
fortran_int  *recvcount,
fortran_int  *recvtype,
fortran_int  *source,
fortran_int  *recvtag,
fortran_int  *comm,
fortran_MPI_Status  *status,
fortran_int  *ierr)
{
MPI_Datatype x_sendtype = convert_type(*sendtype);
MPI_Datatype x_recvtype = convert_type(*recvtype);
MPI_Status x_status;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Sendrecv with *sendtype = %i, x_sendtype = %i\n",(int) *sendtype, x_sendtype);
  printf("wrapper.c... calling PMPI_Sendrecv with *recvtype = %i, x_recvtype = %i\n",(int) *recvtype, x_recvtype);
#endif
        *ierr = (fortran_int) PMPI_Sendrecv( (void *) sendbuf, (int) *sendcount, (MPI_Datatype) x_sendtype, (int) *dest, (int) *sendtag, (void *) recvbuf, (int) *recvcount, (MPI_Datatype) x_recvtype, (int) *source, (int) *recvtag, (MPI_Comm) *comm, &x_status);

        copy_status_struct(status, &x_status, MPI_WRAPPER_TO_FORTRAN);
}

//   
//   MPI_SENDRECV_REPLACE(buf, count, datatype, dest, sendtag, source,
//   recvtag, comm, status)
//   [ INOUT buf] initial address of send and receive buffer (choice)
//   [ IN count] number of elements in send and receive buffer (integer)
//   [ IN datatype] type of elements in send and receive buffer (handle)
//   [ IN dest] rank of destination (integer)
//   [ IN sendtag] send message tag (integer)
//   [ IN source] rank of source (integer)
//   [ IN recvtag] receive message tag (integer)
//   [ IN comm] communicator (handle)
//   [ OUT status] status object (Status)
//   
//   int MPI_Sendrecv_replace(void* buf, int count, MPI_Datatype datatype,
//   int dest, int sendtag, int source, int recvtag, MPI_Comm comm,
//   MPI_Status *status)
//   
//   MPI_SENDRECV_REPLACE(BUF, COUNT, DATATYPE, DEST, SENDTAG, SOURCE,
//   RECVTAG, COMM, STATUS, IERROR)
//   <type> BUF(*)
//   INTEGER COUNT, DATATYPE, DEST, SENDTAG, SOURCE, RECVTAG, COMM,
//   STATUS(MPI_STATUS_SIZE), IERROR
//   
void
mpi_sendrecv_replace_(
void         *buf,
fortran_int  *count,
fortran_int  *type,
fortran_int  *dest,
fortran_int  *sendtag,
fortran_int  *source,
fortran_int  *recvtag,
fortran_int  *comm,
fortran_MPI_Status  *status,
fortran_int  *ierr)
{
MPI_Datatype x_type = convert_type(*type);
MPI_Status x_status;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Sendrecv_replace with *type = %i, x_type = %i\n",(int) *type, x_type);
#endif
        *ierr = (fortran_int) PMPI_Sendrecv_replace( (void *) buf, (int) *count, (MPI_Datatype) x_type, (int) *dest, (int) *sendtag, (int) *source, (int) *recvtag, (MPI_Comm) *comm, &x_status);

        copy_status_struct(status, &x_status, MPI_WRAPPER_TO_FORTRAN);
}

//   
//   MPI_TYPE_CONTIGUOUS(count, oldtype, newtype)
//   [ IN count] replication count (nonnegative integer)
//   [ IN oldtype] old datatype (handle)
//   [ OUT newtype] new datatype (handle)
//   
//   int MPI_Type_contiguous(int count, MPI_Datatype oldtype, MPI_Datatype
//   *newtype)
//   
//   MPI_TYPE_CONTIGUOUS(COUNT, OLDTYPE, NEWTYPE, IERROR)
//   INTEGER COUNT, OLDTYPE, NEWTYPE, IERROR
//   
void
mpi_type_contiguous_(
fortran_int  *count,
fortran_int  *oldtype,
fortran_int  *newtype,
fortran_int  *ierr)
{
MPI_Datatype x_newtype;
MPI_Datatype x_oldtype = convert_type(*oldtype);

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Type_contiguous with *oldtype = %i, x_oldtype = %i\n",(int) *oldtype, x_oldtype);
#endif
        *ierr = (fortran_int) PMPI_Type_contiguous( (int) *count, (MPI_Datatype) x_oldtype, &x_newtype);

        *newtype = (fortran_int) x_newtype ;
}

//   
//   MPI_TYPE_VECTOR( count, blocklength, stride, oldtype, newtype)
//   [ IN count] number of blocks (nonnegative integer)
//   [ IN blocklength] number of elements in each block (nonnegative
//   integer)
//   [ IN stride] number of elements between start of each block (integer)
//   [ IN oldtype] old datatype (handle)
//   [ OUT newtype] new datatype (handle)
//   
//   int MPI_Type_vector(int count, int blocklength, int stride,
//   MPI_Datatype oldtype, MPI_Datatype *newtype)
//   
//   MPI_TYPE_VECTOR(COUNT, BLOCKLENGTH, STRIDE, OLDTYPE, NEWTYPE, IERROR)
//   INTEGER COUNT, BLOCKLENGTH, STRIDE, OLDTYPE, NEWTYPE, IERROR
//   
void
mpi_type_vector_(
fortran_int  *count,
fortran_int  *blocklength,
fortran_int  *stride,
fortran_int  *oldtype,
fortran_int  *newtype,
fortran_int  *ierr)
{
MPI_Datatype x_newtype;
MPI_Datatype x_oldtype = convert_type(*oldtype);

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Type_vector with *oldtype = %i, x_oldtype = %i\n",(int) *oldtype, x_oldtype);
#endif
        *ierr = (fortran_int) PMPI_Type_vector( (int) *count, (int) *blocklength, (int) *stride, (MPI_Datatype) x_oldtype, &x_newtype);

        *newtype = (fortran_int) x_newtype ;
}

//   
//   MPI_TYPE_HVECTOR( count, blocklength, stride, oldtype, newtype)
//   [ IN count] number of blocks (nonnegative integer)
//   [ IN blocklength] number of elements in each block (nonnegative
//   integer)
//   [ IN stride] number of bytes between start of each block (integer)
//   [ IN oldtype] old datatype (handle)
//   [ OUT newtype] new datatype (handle)
//   
//   int MPI_Type_hvector(int count, int blocklength, MPI_Aint stride,
//   MPI_Datatype oldtype, MPI_Datatype *newtype)
//   
//   MPI_TYPE_HVECTOR(COUNT, BLOCKLENGTH, STRIDE, OLDTYPE, NEWTYPE, IERROR)
//   INTEGER COUNT, BLOCKLENGTH, STRIDE, OLDTYPE, NEWTYPE, IERROR
//   
void
mpi_type_hvector_(
fortran_int  *count,
fortran_int  *blocklength,
fortran_int  *stride,
fortran_int  *oldtype,
fortran_int  *newtype,
fortran_int  *ierr)
{
MPI_Datatype x_newtype;
MPI_Datatype x_oldtype = convert_type(*oldtype);

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Type_hvector with *oldtype = %i, x_oldtype = %i\n",(int) *oldtype, x_oldtype);
#endif
        *ierr = (fortran_int) PMPI_Type_hvector( (int) *count, (int) *blocklength, (MPI_Aint) *stride, (MPI_Datatype) x_oldtype, &x_newtype);

        *newtype = (fortran_int) x_newtype ;
}

//   
//   MPI_TYPE_INDEXED( count, array_of_blocklengths,
//   array_of_displacements, oldtype, newtype)
//   [ IN count] number of blocks -- also number of entries in
//   array_of_displacements and array_of_blocklengths (nonnegative integer)
//   [ IN array_of_blocklengths] number of elements per block (array of
//   nonnegative integers)
//   [ IN array_of_displacements] displacement for each block, in multiples
//   of oldtype extent (array of integer)
//   [ IN oldtype] old datatype (handle)
//   [ OUT newtype] new datatype (handle)
//   
//   int MPI_Type_indexed(int count, int *array_of_blocklengths, int
//   *array_of_displacements, MPI_Datatype oldtype, MPI_Datatype *newtype)
//   
//   MPI_TYPE_INDEXED(COUNT, ARRAY_OF_BLOCKLENGTHS, ARRAY_OF_DISPLACEMENTS,
//   OLDTYPE, NEWTYPE, IERROR)
//   INTEGER COUNT, ARRAY_OF_BLOCKLENGTHS(*), ARRAY_OF_DISPLACEMENTS(*),
//   OLDTYPE, NEWTYPE, IERROR
//   
void
mpi_type_indexed_(
fortran_int  *count,
fortran_int  array_of_blocklengths[],
fortran_int  array_of_displacements[],
fortran_int  *oldtype,
fortran_int  *newtype,
fortran_int  *ierr)
{
int x_array_of_displacements[ (int) *count];
int x_array_of_blocklengths[ (int) *count];
MPI_Datatype x_newtype;
MPI_Datatype x_oldtype = convert_type(*oldtype);
int i;

	for (i=0;i< (int) *count;i++) {
	  x_array_of_blocklengths[i] = array_of_blocklengths[i];
	}
	for (i=0;i< (int) *count;i++) {
	  x_array_of_displacements[i] = array_of_displacements[i];
	}

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Type_indexed...\n");
#endif
        *ierr = (fortran_int) PMPI_Type_indexed( (int) *count, x_array_of_blocklengths, x_array_of_displacements, (MPI_Datatype) x_oldtype, &x_newtype);

        *newtype = (fortran_int) x_newtype ;
	free(x_array_of_displacements);
	free(x_array_of_blocklengths);
}

//   
//   MPI_TYPE_HINDEXED( count, array_of_blocklengths,
//   array_of_displacements, oldtype, newtype)
//   [ IN count] number of blocks -- also number of entries in
//   array_of_displacements and array_of_blocklengths (integer)
//   [ IN array_of_blocklengths] number of elements in each block (array of
//   nonnegative integers)
//   [ IN array_of_displacements] byte displacement of each block (array of
//   integer)
//   [ IN oldtype] old datatype (handle)
//   [ OUT newtype] new datatype (handle)
//   
//   int MPI_Type_hindexed(int count, int *array_of_blocklengths, MPI_Aint
//   *array_of_displacements, MPI_Datatype oldtype, MPI_Datatype *newtype)
//   
//   MPI_TYPE_HINDEXED(COUNT, ARRAY_OF_BLOCKLENGTHS,
//   ARRAY_OF_DISPLACEMENTS, OLDTYPE, NEWTYPE, IERROR)
//   INTEGER COUNT, ARRAY_OF_BLOCKLENGTHS(*), ARRAY_OF_DISPLACEMENTS(*),
//   OLDTYPE, NEWTYPE, IERROR
//   
void
mpi_type_hindexed_(
fortran_int  *count,
fortran_int  array_of_blocklengths[],
fortran_int  array_of_displacements[],
fortran_int  *oldtype,
fortran_int  *newtype,
fortran_int  *ierr)
{
MPI_Aint x_array_of_displacements[ (int) *count];
int x_array_of_blocklengths[ (int) *count];
MPI_Datatype x_newtype;
MPI_Datatype x_oldtype = convert_type(*oldtype);
int i;

	for (i=0;i< (int) *count;i++) {
	  x_array_of_blocklengths[i] = array_of_blocklengths[i];
	}
	for (i=0;i< (int) *count;i++) {
	  x_array_of_displacements[i] = array_of_displacements[i];
	}

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Type_hindexed...\n");
#endif
        *ierr = (fortran_int) PMPI_Type_hindexed( (int) *count, x_array_of_blocklengths, x_array_of_displacements, (MPI_Datatype) x_oldtype, &x_newtype);

        *newtype = (fortran_int) x_newtype ;
	free(x_array_of_displacements);
	free(x_array_of_blocklengths);
}

//   
//   MPI_TYPE_STRUCT(count, array_of_blocklengths, array_of_displacements,
//   array_of_types, newtype)
//   [ IN count] number of blocks (integer) -- also number of entries in
//   arrays array_of_types, array_of_displacements and
//   array_of_blocklengths
//   [ IN array_of_blocklength] number of elements in each block (array of
//   integer)
//   [ IN array_of_displacements] byte displacement of each block (array of
//   integer)
//   [ IN array_of_types] type of elements in each block (array of handles
//   to datatype objects)
//   [ OUT newtype] new datatype (handle)
//   
//   int MPI_Type_struct(int count, int *array_of_blocklengths, MPI_Aint
//   *array_of_displacements, MPI_Datatype *array_of_types, MPI_Datatype
//   *newtype)
//   
//   MPI_TYPE_STRUCT(COUNT, ARRAY_OF_BLOCKLENGTHS, ARRAY_OF_DISPLACEMENTS,
//   ARRAY_OF_TYPES, NEWTYPE, IERROR)
//   INTEGER COUNT, ARRAY_OF_BLOCKLENGTHS(*), ARRAY_OF_DISPLACEMENTS(*),
//   ARRAY_OF_TYPES(*), NEWTYPE, IERROR
//   
void
mpi_type_struct_(
fortran_int  *count,
fortran_int  array_of_blocklengths[],
fortran_int  array_of_displacements[],
fortran_int  *array_of_types[],
fortran_int  *newtype,
fortran_int  *ierr)
{
MPI_Aint x_array_of_displacements[ (int) *count];
int x_array_of_blocklengths[ (int) *count];
MPI_Datatype x_newtype;
MPI_Datatype x_array_of_types[ (int) *count ];
int i;

	for (i=0;i< (int) *count;i++) {
	  x_array_of_blocklengths[i] = array_of_blocklengths[i];
	}
	for (i=0;i< (int) *count;i++) {
	  x_array_of_displacements[i] = array_of_displacements[i];
	}
	for (i=0;i< (int) *count;i++) {
	  x_array_of_types[i] = convert_type( x_array_of_types[i] );
	}

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Type_struct...\n");
#endif
        *ierr = (fortran_int) PMPI_Type_struct( (int) *count, x_array_of_blocklengths, x_array_of_displacements, x_array_of_types, &x_newtype);

        *newtype = (fortran_int) x_newtype ;
	free(x_array_of_displacements);
	free(x_array_of_blocklengths);
}

//   
//   MPI_ADDRESS(location, address)
//   [ IN location] location in caller memory (choice)
//   [ OUT address] address of location (integer)
//   
//   int MPI_Address(void* location, MPI_Aint *address)
//   
//   MPI_ADDRESS(LOCATION, ADDRESS, IERROR)
//   <type> LOCATION(*)
//   INTEGER ADDRESS, IERROR
//   
void
mpi_address_(
void         *location,
fortran_int  *address,
fortran_int  *ierr)
{
MPI_Aint x_address ;
#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Address...\n");
#endif
        *ierr = (fortran_int) PMPI_Address( (void *) location, &x_address);

        *address = (fortran_int) x_address ;
}

//   
//   MPI_TYPE_EXTENT(datatype, extent)
//   [ IN datatype] datatype (handle)
//   [ OUT extent] datatype extent (integer)
//   
//   int MPI_Type_extent(MPI_Datatype datatype, MPI_Aint *extent)
//   
//   MPI_TYPE_EXTENT(DATATYPE, EXTENT, IERROR)
//   INTEGER DATATYPE, EXTENT, IERROR
//   
void
mpi_type_extent_(
fortran_int  *datatype,
fortran_int  *extent,
fortran_int  *ierr)
{
MPI_Aint x_extent;
MPI_Datatype x_datatype = convert_type(*datatype);

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Type_extent with *datatype = %i, x_datatype = %i\n",(int) *datatype, x_datatype);
#endif
        *ierr = (fortran_int) PMPI_Type_extent( (MPI_Datatype) x_datatype, &x_extent);

        *extent = (fortran_int) x_extent ;
}

//   
//   MPI_TYPE_SIZE(datatype, size)
//   [ IN datatype] datatype (handle)
//   [ OUT size] datatype size (integer)
//   
//   int MPI_Type_size(MPI_Datatype datatype, int *size)
//   
//   MPI_TYPE_SIZE(DATATYPE, SIZE, IERROR)
//   INTEGER DATATYPE, SIZE, IERROR
//   
void
mpi_type_size_(
fortran_int  *datatype,
fortran_int  *size,
fortran_int  *ierr)
{
int x_size;
MPI_Datatype x_datatype = convert_type(*datatype);

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Type_size with *datatype = %i, x_datatype = %i\n",(int) *datatype, x_datatype);
#endif
        *ierr = (fortran_int) PMPI_Type_size( (MPI_Datatype) x_datatype, &x_size);

        *size = (fortran_int) x_size ;
}

//   
//   MPI_TYPE_LB( datatype, displacement)
//   [ IN datatype] datatype (handle)
//   [ OUT displacement] displacement of lower bound from origin, in bytes
//   (integer)
//   
//   int MPI_Type_lb(MPI_Datatype datatype, MPI_Aint* displacement)
//   
//   MPI_TYPE_LB( DATATYPE, DISPLACEMENT, IERROR)
//   INTEGER DATATYPE, DISPLACEMENT, IERROR
//   
void
mpi_type_lb_(
fortran_int  *datatype,
fortran_int  *displacement,
fortran_int  *ierr)
{
MPI_Aint x_displacement;
MPI_Datatype x_datatype = convert_type(*datatype);

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Type_lb with *datatype = %i, x_datatype = %i\n",(int) *datatype, x_datatype);
#endif
        *ierr = (fortran_int) PMPI_Type_lb( (MPI_Datatype) x_datatype, &x_displacement);

        *displacement = (fortran_int) x_displacement ;
}

//   
//   MPI_TYPE_UB( datatype, displacement)
//   [ IN datatype] datatype (handle)
//   [ OUT displacement] displacement of upper bound from origin, in bytes
//   (integer)
//   
//   int MPI_Type_ub(MPI_Datatype datatype, MPI_Aint* displacement)
//   
//   MPI_TYPE_UB( DATATYPE, DISPLACEMENT, IERROR)
//   INTEGER DATATYPE, DISPLACEMENT, IERROR
//   
void
mpi_type_ub_(
fortran_int  *datatype,
fortran_int  *displacement,
fortran_int  *ierr)
{
MPI_Aint x_displacement;
MPI_Datatype x_datatype = convert_type(*datatype);

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Type_ub with *datatype = %i, x_datatype = %i\n",(int) *datatype, x_datatype);
#endif
        *ierr = (fortran_int) PMPI_Type_ub( (MPI_Datatype) x_datatype, &x_displacement);

        *displacement = (fortran_int) x_displacement ;
}

//   
//   MPI_TYPE_COMMIT(datatype)
//   [ INOUT datatype] datatype that is committed (handle)
//   
//   int MPI_Type_commit(MPI_Datatype *datatype)
//   
//   MPI_TYPE_COMMIT(DATATYPE, IERROR)
//   INTEGER DATATYPE, IERROR
//   
void
mpi_type_commit_(
fortran_int  *datatype,
fortran_int  *ierr)
{
MPI_Datatype x_datatype = convert_type(*datatype);

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Type_commit with *datatype = %i, x_datatype = %i\n",(int) *datatype, x_datatype);
#endif
        *ierr = (fortran_int) PMPI_Type_commit( &x_datatype );

        *datatype = (fortran_int) x_datatype ;
}

//
//   MPI_TYPE_FREE(datatype)
//   [ INOUT datatype] datatype that is freed (handle)
//   
//   int MPI_Type_free(MPI_Datatype *datatype)
//   
//   MPI_TYPE_FREE(DATATYPE, IERROR)
//   INTEGER DATATYPE, IERROR
//   
void
mpi_type_free_(
fortran_int  *datatype,
fortran_int  *ierr)
{
MPI_Datatype x_datatype = convert_type(*datatype);

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Type_free with *datatype = %i, x_datatype = %i\n",(int) *datatype, x_datatype);
#endif
        *ierr = (fortran_int) PMPI_Type_free( &x_datatype );

        *datatype = (fortran_int) x_datatype ;
}

//   
//   MPI_GET_ELEMENTS( status, datatype, count)
//   [ IN status] return status of receive operation (Status)
//   [ IN datatype] datatype used by receive operation (handle)
//   [ OUT count] number of received basic elements (integer)
//   
//   int MPI_Get_elements(MPI_Status *status, MPI_Datatype datatype, int
//   *count)
//   
//   MPI_GET_ELEMENTS(STATUS, DATATYPE, COUNT, IERROR)
//   INTEGER STATUS(MPI_STATUS_SIZE), DATATYPE, COUNT, IERROR
//   
void
mpi_get_elements_(
fortran_MPI_Status  *status,
fortran_int  *type,
fortran_int  *count,
fortran_int  *ierr)
{
MPI_Datatype x_type = convert_type(*type);
int x_count ;

MPI_Status x_status;
        copy_status_struct(status, &x_status, FORTRAN_TO_MPI_WRAPPER);

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Get_elements with *type = %i, x_type = %i\n",(int) *type, x_type);
#endif
        *ierr = (fortran_int) PMPI_Get_elements( &x_status, (MPI_Datatype) x_type, &x_count);

        copy_status_struct(status, &x_status, MPI_WRAPPER_TO_FORTRAN);
        *count = (fortran_int) x_count ;
#ifdef DEBUG
  printf("wrapper.c... output from PMPI_Get_elements: *count = %i...\n",(int) *count);
#endif
}

//   
//   MPI_PACK(inbuf, incount, datatype, outbuf, outsize, position, comm)
//   [ IN inbuf] input buffer start (choice)
//   [ IN incount] number of input data items (integer)
//   [ IN datatype] datatype of each input data item (handle)
//   [ OUT outbuf] output buffer start (choice)
//   [ IN outsize] output buffer size, in bytes (integer)
//   [ INOUT position] current position in buffer, in bytes (integer)
//   [ IN comm] communicator for packed message (handle)
//   
//   int MPI_Pack(void* inbuf, int incount, MPI_Datatype datatype, void
//   *outbuf, int outsize, int *position, MPI_Comm comm)
//   
//   MPI_PACK(INBUF, INCOUNT, DATATYPE, OUTBUF, OUTSIZE, POSITION, COMM,
//   IERROR)
//   <type> INBUF(*), OUTBUF(*)
//   INTEGER INCOUNT, DATATYPE, OUTSIZE, POSITION, COMM, IERROR
//   
void
mpi_pack_(
void         *inbuf,
fortran_int  *incount,
fortran_int  *datatype,
void         *outbuf,
fortran_int  *outsize,
fortran_int  *position,
fortran_int  *comm,
fortran_int  *ierr)
{
int x_position = *position;
MPI_Datatype x_datatype = convert_type(*datatype);

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Pack with *datatype = %i, x_datatype = %i\n",(int) *datatype, x_datatype);
#endif
        *ierr = (fortran_int) PMPI_Pack( (void *) inbuf, (int) *incount, (MPI_Datatype) x_datatype, (void *) outbuf, (int) *outsize, &x_position, (MPI_Comm) *comm);

        *position = (fortran_int) x_position ;
}

//   
//   MPI_UNPACK(inbuf, insize, position, outbuf, outcount, datatype, comm)
//   [ IN inbuf] input buffer start (choice)
//   [ IN insize] size of input buffer, in bytes (integer)
//   [ INOUT position] current position in bytes (integer)
//   [ OUT outbuf] output buffer start (choice)
//   [ IN outcount] number of items to be unpacked (integer)
//   [ IN datatype] datatype of each output data item (handle)
//   [ IN comm] communicator for packed message (handle)
//   
//   int MPI_Unpack(void* inbuf, int insize, int *position, void *outbuf,
//   int outcount, MPI_Datatype datatype, MPI_Comm comm)
//   
//   MPI_UNPACK(INBUF, INSIZE, POSITION, OUTBUF, OUTCOUNT, DATATYPE, COMM,
//   IERROR)
//   <type> INBUF(*), OUTBUF(*)
//   INTEGER INSIZE, POSITION, OUTCOUNT, DATATYPE, COMM, IERROR
//   
void
mpi_unpack_(
void         *inbuf,
fortran_int  *insize,
fortran_int  *position,
void         *outbuf,
fortran_int  *outcount,
fortran_int  *datatype,
fortran_int  *comm,
fortran_int  *ierr)
{
int x_position = *position;
MPI_Datatype x_datatype = convert_type(*datatype);

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Unpack with *datatype = %i, x_datatype = %i\n",(int) *datatype, x_datatype);
#endif
        *ierr = (fortran_int) PMPI_Unpack( (void *) inbuf, (int) *insize, &x_position, (void *) outbuf, (int) *outcount, (MPI_Datatype) x_datatype, (MPI_Comm) *comm);

        *position = (fortran_int) x_position ;
}

//   
//   MPI_PACK_SIZE(incount, datatype, comm, size)
//   [ IN incount] count argument to packing call (integer)
//   [ IN datatype] datatype argument to packing call (handle)
//   [ IN comm] communicator argument to packing call (handle)
//   [ OUT size] upper bound on size of packed message, in bytes (integer)
//   
//   int MPI_Pack_size(int incount, MPI_Datatype datatype, MPI_Comm comm,
//   int *size)
//   
//   MPI_PACK_SIZE(INCOUNT, DATATYPE, COMM, SIZE, IERROR)
//   INTEGER INCOUNT, DATATYPE, COMM, SIZE, IERROR
//
void
mpi_pack_size_(
fortran_int  *incount,
fortran_int  *datatype,
fortran_int  *comm,
fortran_int  *size,
fortran_int  *ierr)
{
int x_size;
MPI_Datatype x_datatype = convert_type(*datatype);

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Pack_size\n");
#endif
        *ierr = (fortran_int) PMPI_Pack_size( (int) *incount, (MPI_Datatype) x_datatype, (MPI_Comm) *comm, &x_size);

        *size = (fortran_int) x_size ;
}

//   
//   MPI_BCAST( buffer, count, datatype, root, comm )
//   [ INOUT buffer] starting address of buffer (choice)
//   [ IN count] number of entries in buffer (integer)
//   [ IN datatype] data type of buffer (handle)
//   [ IN root] rank of broadcast root (integer)
//   [ IN comm] communicator (handle)
//   
//   int MPI_Bcast(void* buffer, int count, MPI_Datatype datatype, int
//   root, MPI_Comm comm )
//   
//   MPI_BCAST(BUFFER, COUNT, DATATYPE, ROOT, COMM, IERROR)
//   <type> BUFFER(*)
//   INTEGER COUNT, DATATYPE, ROOT, COMM, IERROR
//   
void
mpi_bcast_(
void         *buf,
fortran_int  *count,
fortran_int  *type,
fortran_int  *root,
fortran_int  *comm,
fortran_int  *ierr)
{
MPI_Datatype x_type = convert_type(*type);
#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Bcast with *type = %i, x_type = %i\n",(int) *type, x_type);
#endif
        *ierr = (fortran_int) PMPI_Bcast( (void *) buf, (int) *count, (MPI_Datatype) x_type, (int) *root, (MPI_Comm) *comm);
}

//
// MPI_FINALIZE()
//
// int MPI_Finalize(void)
//
// MPI_FINALIZE(IERROR)
// INTEGER IERROR
//
void
mpi_finalize_(
fortran_int *ierr)
{
#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Finalize...\n");
#endif
        *ierr = (fortran_int) PMPI_Finalize();
}

//   
//   MPI_GATHER( sendbuf, sendcount, sendtype, recvbuf, recvcount,
//   recvtype, root, comm)
//   [ IN sendbuf] starting address of send buffer (choice)
//   [ IN sendcount] number of elements in send buffer (integer)
//   [ IN sendtype] data type of send buffer elements (handle)
//   [ OUT recvbuf] address of receive buffer (choice, significant only at
//   root)
//   [ IN recvcount] number of elements for any single receive (integer,
//   significant only at root)
//   [ IN recvtype] data type of recv buffer elements (significant only at
//   root) (handle)
//   [ IN root] rank of receiving process (integer)
//   [ IN comm] communicator (handle)
//   
//   int MPI_Gather(void* sendbuf, int sendcount, MPI_Datatype sendtype,
//   void* recvbuf, int recvcount, MPI_Datatype recvtype, int root,
//   MPI_Comm comm)
//   
//   MPI_GATHER(SENDBUF, SENDCOUNT, SENDTYPE, RECVBUF, RECVCOUNT, RECVTYPE,
//   ROOT, COMM, IERROR)
//   <type> SENDBUF(*), RECVBUF(*)
//   INTEGER SENDCOUNT, SENDTYPE, RECVCOUNT, RECVTYPE, ROOT, COMM, IERROR
//   
void
mpi_gather_(
void         *sendbuf,
fortran_int  *sendcount,
fortran_int  *sendtype,
void         *recvbuf,
fortran_int  *recvcount,
fortran_int  *recvtype,
fortran_int  *root,
fortran_int  *comm,
fortran_int  *ierr)
{
MPI_Datatype x_sendtype = convert_type(*sendtype);
MPI_Datatype x_recvtype = convert_type(*recvtype);

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Gather with *sendtype = %i, x_sendtype = %i\n",(int) *sendtype, x_sendtype);
  printf("wrapper.c... calling PMPI_Gather with *recvtype = %i, x_recvtype = %i\n",(int) *recvtype, x_recvtype);
#endif
        *ierr = (fortran_int) PMPI_Gather( (void *) sendbuf, (int) *sendcount, (MPI_Datatype) x_sendtype, (void *) recvbuf, (int) *recvcount, (MPI_Datatype) x_recvtype, (int) *root, (MPI_Comm) *comm);

}

//
//   MPI_GATHERV( sendbuf, sendcount, sendtype, recvbuf, recvcounts,
//   displs, recvtype, root, comm)
//   [ IN sendbuf] starting address of send buffer (choice)
//   [ IN sendcount] number of elements in send buffer (integer)
//   [ IN sendtype] data type of send buffer elements (handle)
//   [ OUT recvbuf] address of receive buffer (choice, significant only at
//   root)
//   [ IN recvcounts] integer array (of length group size) containing the
//   number of elements that are received from each process (significant
//   only at root)
//   [ IN displs] integer array (of length group size). Entry i specifies
//   the displacement relative to recvbuf at which to place the incoming
//   data from process i (significant only at root)
//   [ IN recvtype] data type of recv buffer elements (significant only at
//   root) (handle)
//   [ IN root] rank of receiving process (integer)
//   [ IN comm] communicator (handle)
//   
//   int MPI_Gatherv(void* sendbuf, int sendcount, MPI_Datatype sendtype,
//   void* recvbuf, int *recvcounts, int *displs, MPI_Datatype recvtype,
//   int root, MPI_Comm comm)
//   
//   MPI_GATHERV(SENDBUF, SENDCOUNT, SENDTYPE, RECVBUF, RECVCOUNTS, DISPLS,
//   RECVTYPE, ROOT, COMM, IERROR)
//   <type> SENDBUF(*), RECVBUF(*)
//   INTEGER SENDCOUNT, SENDTYPE, RECVCOUNTS(*), DISPLS(*), RECVTYPE, ROOT,
//   COMM, IERROR
//   
void
mpi_gatherv_(
void         *sendbuf,
fortran_int  *sendcount,
fortran_int  *sendtype,
void         *recvbuf,
fortran_int  recvcounts[],
fortran_int  array_of_displs[],
fortran_int  *recvtype,
fortran_int  *root,
fortran_int  *comm,
fortran_int  *ierr)
{
MPI_Datatype x_sendtype = convert_type(*sendtype);
MPI_Datatype x_recvtype = convert_type(*recvtype);
int *x_array_of_displs;
int *x_recvcounts;
int i;

MPI_Group group;
int groupsize;
        *ierr = PMPI_Comm_group((MPI_Comm) *comm, &group);
        *ierr = PMPI_Group_size((MPI_Group) group, &groupsize);

	x_array_of_displs = malloc (groupsize * sizeof(int) ) ;
	x_recvcounts = malloc (groupsize * sizeof(int) ) ;

	for (i=0;i<groupsize;i++) {
	  x_array_of_displs[i] = array_of_displs[i];
	}
	for (i=0;i<groupsize;i++) {
	  x_recvcounts[i] = recvcounts[i];
	}

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Gatherv with *sendtype = %i, x_sendtype = %i\n",(int) *sendtype, x_sendtype);
  printf("wrapper.c... calling PMPI_Gatherv with *recvtype = %i, x_recvtype = %i\n",(int) *recvtype, x_recvtype);
#endif
        *ierr = (fortran_int) PMPI_Gatherv( (void *) sendbuf, (int) *sendcount, (MPI_Datatype) x_sendtype, (void *) recvbuf, x_array_of_displs, x_recvcounts, (MPI_Datatype) x_recvtype, (int) *root, (MPI_Comm) *comm);

	free(x_array_of_displs);
	free(x_recvcounts);
}

//   
//   MPI_SCATTER( sendbuf, sendcount, sendtype, recvbuf, recvcount,
//   recvtype, root, comm)
//   [ IN sendbuf] address of send buffer (choice, significant only at
//   root)
//   [ IN sendcount] number of elements sent to each process (integer,
//   significant only at root)
//   [ IN sendtype] data type of send buffer elements (significant only at
//   root) (handle)
//   [ OUT recvbuf] address of receive buffer (choice)
//   [ IN recvcount] number of elements in receive buffer (integer)
//   [ IN recvtype] data type of receive buffer elements (handle)
//   [ IN root] rank of sending process (integer)
//   [ IN comm] communicator (handle)
//   
//   int MPI_Scatter(void* sendbuf, int sendcount, MPI_Datatype sendtype,
//   void* recvbuf, int recvcount, MPI_Datatype recvtype, int root,
//   MPI_Comm comm)
//   
//   MPI_SCATTER(SENDBUF, SENDCOUNT, SENDTYPE, RECVBUF, RECVCOUNT,
//   RECVTYPE, ROOT, COMM, IERROR)
//   <type> SENDBUF(*), RECVBUF(*)
//   INTEGER SENDCOUNT, SENDTYPE, RECVCOUNT, RECVTYPE, ROOT, COMM, IERROR
//
void
mpi_scatter_(
void         *sendbuf,
fortran_int  *sendcount,
fortran_int  *sendtype,
void         *recvbuf,
fortran_int  *recvcount,
fortran_int  *recvtype,
fortran_int  *root,
fortran_int  *comm,
fortran_int  *ierr)
{
MPI_Datatype x_sendtype = convert_type(*sendtype);
MPI_Datatype x_recvtype = convert_type(*recvtype);

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Scatter with *sendtype = %i, x_sendtype = %i\n",(int) *sendtype, x_sendtype);
  printf("wrapper.c... calling PMPI_Scatter with *recvtype = %i, x_recvtype = %i\n",(int) *recvtype, x_recvtype);
#endif
        *ierr = (fortran_int) PMPI_Scatter( (void *) sendbuf, (int) *sendcount, (MPI_Datatype) x_sendtype, (void *) recvbuf, (int) *recvcount, (MPI_Datatype) x_recvtype, (int) *root, (MPI_Comm) *comm);

}

//
//   MPI_SCATTERV( sendbuf, sendcounts, displs, sendtype, recvbuf,
//   recvcount, recvtype, root, comm)
//   [ IN sendbuf] address of send buffer (choice, significant only at
//   root)
//   [ IN sendcounts] integer array (of length group size) specifying the
//   number of elements to send to each processor
//   [ IN displs] integer array (of length group size). Entry i specifies
//   the displacement (relative to sendbuf from which to take the outgoing
//   data to process i
//   [ IN sendtype] data type of send buffer elements (handle)
//   [ OUT recvbuf] address of receive buffer (choice)
//   [ IN recvcount] number of elements in receive buffer (integer)
//   [ IN recvtype] data type of receive buffer elements (handle)
//   [ IN root] rank of sending process (integer)
//   [ IN comm] communicator (handle)
//   
//   int MPI_Scatterv(void* sendbuf, int *sendcounts, int *displs,
//   MPI_Datatype sendtype, void* recvbuf, int recvcount, MPI_Datatype
//   recvtype, int root, MPI_Comm comm)
//   
//   MPI_SCATTERV(SENDBUF, SENDCOUNTS, DISPLS, SENDTYPE, RECVBUF,
//   RECVCOUNT, RECVTYPE, ROOT, COMM, IERROR)
//   <type> SENDBUF(*), RECVBUF(*)
//   INTEGER SENDCOUNTS(*), DISPLS(*), SENDTYPE, RECVCOUNT, RECVTYPE, ROOT,
//   COMM, IERROR
//   
void
mpi_scatterv_(
void         *sendbuf,
fortran_int  sendcounts[],
fortran_int  array_of_displs[],
fortran_int  *sendtype,
void         *recvbuf,
fortran_int  *recvcount,
fortran_int  *recvtype,
fortran_int  *root,
fortran_int  *comm,
fortran_int  *ierr)
{
MPI_Datatype x_sendtype = convert_type(*sendtype);
MPI_Datatype x_recvtype = convert_type(*recvtype);
int *x_array_of_displs;
int *x_sendcounts;
int i;

MPI_Group group;
int groupsize;
        *ierr = PMPI_Comm_group((MPI_Comm) *comm, &group);
        *ierr = PMPI_Group_size((MPI_Group) group, &groupsize);

	x_array_of_displs = malloc (groupsize * sizeof(int) ) ;
	x_sendcounts = malloc (groupsize * sizeof(int) ) ;

	for (i=0;i<groupsize;i++) {
	  x_array_of_displs[i] = array_of_displs[i];
	}
	for (i=0;i<groupsize;i++) {
	  x_sendcounts[i] = sendcounts[i];
	}

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Scatterv with *sendtype = %i, x_sendtype = %i\n",(int) *sendtype, x_sendtype);
  printf("wrapper.c... calling PMPI_Scatterv with *recvtype = %i, x_recvtype = %i\n",(int) *recvtype, x_recvtype);
#endif
        *ierr = (fortran_int) PMPI_Scatterv( (void *) sendbuf, x_sendcounts, x_array_of_displs, (MPI_Datatype) x_sendtype, (void *) recvbuf, (int) *recvcount, (MPI_Datatype) x_recvtype, (int) *root, (MPI_Comm) *comm);

	free(x_array_of_displs);
	free(x_sendcounts);
}

//   
//   MPI_ALLGATHER( sendbuf, sendcount, sendtype, recvbuf, recvcount,
//   recvtype, comm)
//   [ IN sendbuf] starting address of send buffer (choice)
//   [ IN sendcount] number of elements in send buffer (integer)
//   [ IN sendtype] data type of send buffer elements (handle)
//   [ OUT recvbuf] address of receive buffer (choice)
//   [ IN recvcount] number of elements received from any process (integer)
//   [ IN recvtype] data type of receive buffer elements (handle)
//   [ IN comm] communicator (handle)
//   
//   int MPI_Allgather(void* sendbuf, int sendcount, MPI_Datatype sendtype,
//   void* recvbuf, int recvcount, MPI_Datatype recvtype, MPI_Comm comm)
//   
//   MPI_ALLGATHER(SENDBUF, SENDCOUNT, SENDTYPE, RECVBUF, RECVCOUNT,
//   RECVTYPE, COMM, IERROR)
//   <type> SENDBUF(*), RECVBUF(*)
//   INTEGER SENDCOUNT, SENDTYPE, RECVCOUNT, RECVTYPE, COMM, IERROR
//   
void
mpi_allgather_(
void         *sendbuf,
fortran_int  *sendcount,
fortran_int  *sendtype,
void         *recvbuf,
fortran_int  *recvcount,
fortran_int  *recvtype,
fortran_int  *comm,
fortran_int  *ierr)
{
MPI_Datatype x_sendtype = convert_type(*sendtype);
MPI_Datatype x_recvtype = convert_type(*recvtype);

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Allgather with *sendtype = %i, x_sendtype = %i\n",(int) *sendtype, x_sendtype);
  printf("wrapper.c... calling PMPI_Allgather with *recvtype = %i, x_recvtype = %i\n",(int) *recvtype, x_recvtype);
#endif
        *ierr = (fortran_int) PMPI_Allgather( (void *) sendbuf, (int) *sendcount, (MPI_Datatype) x_sendtype, (void *) recvbuf, (int) *recvcount, (MPI_Datatype) x_recvtype, (MPI_Comm) *comm);

}

//
//   MPI_ALLGATHERV( sendbuf, sendcount, sendtype, recvbuf, recvcounts,
//   displs, recvtype, comm)
//   [ IN sendbuf] starting address of send buffer (choice)
//   [ IN sendcount] number of elements in send buffer (integer)
//   [ IN sendtype] data type of send buffer elements (handle)
//   [ OUT recvbuf] address of receive buffer (choice)
//   [ IN recvcounts] integer array (of length group size) containing the
//   number of elements that are received from each process
//   [ IN displs] integer array (of length group size). Entry i specifies
//   the displacement (relative to recvbuf) at which to place the incoming
//   data from process i
//   [ IN recvtype] data type of receive buffer elements (handle)
//   [ IN comm] communicator (handle)
//   
//   int MPI_Allgatherv(void* sendbuf, int sendcount, MPI_Datatype
//   sendtype, void* recvbuf, int *recvcounts, int *displs, MPI_Datatype
//   recvtype, MPI_Comm comm)
//   
//   MPI_ALLGATHERV(SENDBUF, SENDCOUNT, SENDTYPE, RECVBUF, RECVCOUNTS,
//   DISPLS, RECVTYPE, COMM, IERROR)
//   <type> SENDBUF(*), RECVBUF(*)
//   INTEGER SENDCOUNT, SENDTYPE, RECVCOUNTS(*), DISPLS(*), RECVTYPE, COMM,
//   IERROR
//   
void
mpi_allgatherv_(
void         *sendbuf,
fortran_int  *sendcount,
fortran_int  *sendtype,
void         *recvbuf,
fortran_int  recvcounts[],
fortran_int  array_of_displs[],
fortran_int  *recvtype,
fortran_int  *comm,
fortran_int  *ierr)
{
MPI_Datatype x_sendtype = convert_type(*sendtype);
MPI_Datatype x_recvtype = convert_type(*recvtype);
int *x_array_of_displs;
int *x_recvcounts;
int i;

MPI_Group group;
int groupsize;
        *ierr = PMPI_Comm_group((MPI_Comm) *comm, &group);
        *ierr = PMPI_Group_size((MPI_Group) group, &groupsize);

	x_array_of_displs = malloc (groupsize * sizeof(int) ) ;
	x_recvcounts = malloc (groupsize * sizeof(int) ) ;

	for (i=0;i<groupsize;i++) {
	  x_array_of_displs[i] = array_of_displs[i];
	}
	for (i=0;i<groupsize;i++) {
	  x_recvcounts[i] = recvcounts[i];
	}

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Allgatherv with *sendtype = %i, x_sendtype = %i\n",(int) *sendtype, x_sendtype);
  printf("wrapper.c... calling PMPI_Allgatherv with *recvtype = %i, x_recvtype = %i\n",(int) *recvtype, x_recvtype);
#endif
        *ierr = (fortran_int) PMPI_Allgatherv( (void *) sendbuf, (int) *sendcount, (MPI_Datatype) x_sendtype, (void *) recvbuf, x_array_of_displs, x_recvcounts, (MPI_Datatype) x_recvtype, (MPI_Comm) *comm);

	free(x_array_of_displs);
	free(x_recvcounts);
}

//   
//   MPI_ALLTOALL(sendbuf, sendcount, sendtype, recvbuf, recvcount,
//   recvtype, comm)
//   [ IN sendbuf] starting address of send buffer (choice)
//   [ IN sendcount] number of elements sent to each process (integer)
//   [ IN sendtype] data type of send buffer elements (handle)
//   [ OUT recvbuf] address of receive buffer (choice)
//   [ IN recvcount] number of elements received from any process (integer)
//   [ IN recvtype] data type of receive buffer elements (handle)
//   [ IN comm] communicator (handle)
//   
//   int MPI_Alltoall(void* sendbuf, int sendcount, MPI_Datatype sendtype,
//   void* recvbuf, int recvcount, MPI_Datatype recvtype, MPI_Comm comm)
//   
//   MPI_ALLTOALL(SENDBUF, SENDCOUNT, SENDTYPE, RECVBUF, RECVCOUNT,
//   RECVTYPE, COMM, IERROR)
//   <type> SENDBUF(*), RECVBUF(*)
//   INTEGER SENDCOUNT, SENDTYPE, RECVCOUNT, RECVTYPE, COMM, IERROR
//   
void
mpi_alltoall_(
void         *sendbuf,
fortran_int  *sendcount,
fortran_int  *sendtype,
void         *recvbuf,
fortran_int  *recvcount,
fortran_int  *recvtype,
fortran_int  *comm,
fortran_int  *ierr)
{
MPI_Datatype x_sendtype = convert_type(*sendtype);
MPI_Datatype x_recvtype = convert_type(*recvtype);

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Alltoall with *sendtype = %i, x_sendtype = %i\n",(int) *sendtype, x_sendtype);
  printf("wrapper.c... calling PMPI_Alltoall with *recvtype = %i, x_recvtype = %i\n",(int) *recvtype, x_recvtype);
#endif
        *ierr = (fortran_int) PMPI_Alltoall( (void *) sendbuf, (int) *sendcount, (MPI_Datatype) x_sendtype, (void *) recvbuf, (int) *recvcount, (MPI_Datatype) x_recvtype, (MPI_Comm) *comm);

}

//   
//   MPI_ALLTOALLV(sendbuf, sendcounts, sdispls, sendtype, recvbuf,
//   recvcounts, rdispls, recvtype, comm)
//   [ IN sendbuf] starting address of send buffer (choice)
//   [ IN sendcounts] integer array equal to the group size specifying the
//   number of elements to send to each processor
//   [ IN sdispls] integer array (of length group size). Entry j specifies
//   the displacement (relative to sendbuf from which to take the outgoing
//   data destined for process j
//   [ IN sendtype] data type of send buffer elements (handle)
//   [ OUT recvbuf] address of receive buffer (choice)
//   [ IN recvcounts] integer array equal to the group size specifying the
//   number of elements that can be received from each processor
//   [ IN rdispls] integer array (of length group size). Entry i specifies
//   the displacement (relative to recvbuf at which to place the incoming
//   data from process i
//   [ IN recvtype] data type of receive buffer elements (handle)
//   [ IN comm] communicator (handle)
//   
//   int MPI_Alltoallv(void* sendbuf, int *sendcounts, int *sdispls,
//   MPI_Datatype sendtype, void* recvbuf, int *recvcounts, int *rdispls,
//   MPI_Datatype recvtype, MPI_Comm comm)
//   
//   MPI_ALLTOALLV(SENDBUF, SENDCOUNTS, SDISPLS, SENDTYPE, RECVBUF,
//   RECVCOUNTS, RDISPLS, RECVTYPE, COMM, IERROR)
//   <type> SENDBUF(*), RECVBUF(*)
//   INTEGER SENDCOUNTS(*), SDISPLS(*), SENDTYPE, RECVCOUNTS(*),
//   RDISPLS(*), RECVTYPE, COMM, IERROR
//   
void
mpi_alltoallv_(
void         *sendbuf,
fortran_int  sendcounts[],
fortran_int  sdispls[],
fortran_int  *sendtype,
void         *recvbuf,
fortran_int  recvcounts[],
fortran_int  rdispls[],
fortran_int  *recvtype,
fortran_int  *comm,
fortran_int  *ierr)
{
MPI_Datatype x_sendtype = convert_type(*sendtype);
MPI_Datatype x_recvtype = convert_type(*recvtype);
int *x_rdispls;
int *x_sdispls;
int *x_recvcounts;
int *x_sendcounts;
int i;

MPI_Group group;
int groupsize;
        *ierr = PMPI_Comm_group((MPI_Comm) *comm, &group);
        *ierr = PMPI_Group_size((MPI_Group) group, &groupsize);

	x_rdispls = malloc (groupsize * sizeof(int) ) ;
	x_sdispls = malloc (groupsize * sizeof(int) ) ;
	x_recvcounts = malloc (groupsize * sizeof(int) ) ;
	x_sendcounts = malloc (groupsize * sizeof(int) ) ;

	for (i=0;i<groupsize;i++) {
	  x_rdispls[i] = rdispls[i];
	}
	for (i=0;i<groupsize;i++) {
	  x_sdispls[i] = sdispls[i];
	}
	for (i=0;i<groupsize;i++) {
	  x_recvcounts[i] = recvcounts[i];
	}
	for (i=0;i<groupsize;i++) {
	  x_sendcounts[i] = sendcounts[i];
	}

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Alltoallv with *sendtype = %i, x_sendtype = %i\n",(int) *sendtype, x_sendtype);
  printf("wrapper.c... calling PMPI_Alltoallv with *recvtype = %i, x_recvtype = %i\n",(int) *recvtype, x_recvtype);
#endif
        *ierr = (fortran_int) PMPI_Alltoallv( (void *) sendbuf, x_sendcounts, x_sdispls, (MPI_Datatype) x_sendtype, (void *) recvbuf, x_rdispls, x_recvcounts, (MPI_Datatype) x_recvtype, (MPI_Comm) *comm);

	free(x_rdispls);
	free(x_sdispls);
	free(x_recvcounts);
	free(x_sendcounts);
}

//   
//   MPI_REDUCE( sendbuf, recvbuf, count, datatype, op, root, comm)
//   [ IN sendbuf] address of send buffer (choice)
//   [ OUT recvbuf] address of receive buffer (choice, significant only at
//   root)
//   [ IN count] number of elements in send buffer (integer)
//   [ IN datatype] data type of elements of send buffer (handle)
//   [ IN op] reduce operation (handle)
//   [ IN root] rank of root process (integer)
//   [ IN comm] communicator (handle)
//   
//   int MPI_Reduce(void* sendbuf, void* recvbuf, int count, MPI_Datatype
//   datatype, MPI_Op op, int root, MPI_Comm comm)
//   
//   MPI_REDUCE(SENDBUF, RECVBUF, COUNT, DATATYPE, OP, ROOT, COMM, IERROR)
//   <type> SENDBUF(*), RECVBUF(*)
//   INTEGER COUNT, DATATYPE, OP, ROOT, COMM, IERROR
//   
void
mpi_reduce_(
void         *sendbuf,
void         *recbuf,
fortran_int  *count,
fortran_int  *type,
fortran_int  *op,	// MPI_Op
fortran_int  *root,
fortran_int  *comm,
fortran_int  *ierr)
{
MPI_Datatype x_type = convert_type(*type);

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Reduce with *type = %i, x_type = %i\n",(int) *type, x_type);
#endif
        *ierr = (fortran_int) PMPI_Reduce( (void *) sendbuf, (void *) recbuf, (int) *count, (MPI_Datatype) x_type, (MPI_Op) *op, (int) *root, (MPI_Comm) *comm);

}

//   
//   MPI_OP_CREATE( function, commute, op)
//   [ IN function] user defined function (function)
//   [ IN commute] true if commutative; false otherwise.
//   [ OUT op] operation (handle)
//   
//   int MPI_Op_create(MPI_User_function *function, int commute, MPI_Op
//   *op)
//   
//   MPI_OP_CREATE( FUNCTION, COMMUTE, OP, IERROR)
//   EXTERNAL FUNCTION
//   LOGICAL COMMUTE
//   INTEGER OP, IERROR
//   
void
mpi_op_create_(
void *function,
fortran_int  *commute,
fortran_int  *op,	// MPI_Op
fortran_int  *ierr)
{
MPI_Op x_op;

#ifdef DEBUG
  printf("wrapper.c... input to PMPI_Op_create...\n");
#endif
        *ierr = (fortran_int) PMPI_Op_create( (MPI_User_function *) function, (int) *commute, &x_op );
        *op = (fortran_int) x_op ;

}

//
// MPI_OP_FREE( op)
// [ IN op] operation (handle)
//
// int MPI_Op_free( MPI_Op *op)
//
// MPI_OP_FREE( OP, IERROR)
// INTEGER OP, IERROR
//
void
mpi_op_free_(
fortran_int *op,
fortran_int *ierr)
{
#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Op...\n");
#endif
        *ierr = (fortran_int) PMPI_Op_free( (MPI_Op *) op);
}

//   
//   MPI_ALLREDUCE( sendbuf, recvbuf, count, datatype, op, comm)
//   [ IN sendbuf] starting address of send buffer (choice)
//   [ OUT recvbuf] starting address of receive buffer (choice)
//   [ IN count] number of elements in send buffer (integer)
//   [ IN datatype] data type of elements of send buffer (handle)
//   [ IN op] operation (handle)
//   [ IN comm] communicator (handle)
//   
//   int MPI_Allreduce(void* sendbuf, void* recvbuf, int count,
//   MPI_Datatype datatype, MPI_Op op, MPI_Comm comm)
//   
//   MPI_ALLREDUCE(SENDBUF, RECVBUF, COUNT, DATATYPE, OP, COMM, IERROR)
//   <type> SENDBUF(*), RECVBUF(*)
//   INTEGER COUNT, DATATYPE, OP, COMM, IERROR
//   
void
mpi_allreduce_(
void         *sendbuf,
void         *recbuf,
fortran_int  *count,
fortran_int  *type,
fortran_int  *op,	// MPI_Op
fortran_int  *comm,
fortran_int  *ierr)
{
MPI_Datatype x_type = convert_type(*type);
#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Allreduce with *type = %i, x_type = %i\n",(int) *type, x_type);
#endif
        *ierr = (fortran_int) PMPI_Allreduce( (void *) sendbuf, (void *) recbuf, (int) *count, (MPI_Datatype) x_type, (MPI_Op) *op, (MPI_Comm) *comm);
}

//   
//   MPI_REDUCE_SCATTER( sendbuf, recvbuf, recvcounts, datatype, op, comm)
//   [ IN sendbuf] starting address of send buffer (choice)
//   [ OUT recvbuf] starting address of receive buffer (choice)
//   [ IN recvcounts] integer array specifying the number of elements in
//   result distributed to each process. Array must be identical on all
//   calling processes.
//   [ IN datatype] data type of elements of input buffer (handle)
//   [ IN op] operation (handle)
//   [ IN comm] communicator (handle)
//   
//   int MPI_Reduce_scatter(void* sendbuf, void* recvbuf, int *recvcounts,
//   MPI_Datatype datatype, MPI_Op op, MPI_Comm comm)
//   
//   MPI_REDUCE_SCATTER(SENDBUF, RECVBUF, RECVCOUNTS, DATATYPE, OP, COMM,
//   IERROR)
//   <type> SENDBUF(*), RECVBUF(*)
//   INTEGER RECVCOUNTS(*), DATATYPE, OP, COMM, IERROR
//   
void
mpi_reduce_scatter_(
void         *sendbuf,
void         *recvbuf,
fortran_int  recvcounts[],
fortran_int  *datatype,
fortran_int  *op,
fortran_int  *comm,
fortran_int  *ierr)
{
MPI_Datatype x_datatype = convert_type(*datatype);
int *x_recvcounts;
int i;

MPI_Group group;
int groupsize;
        *ierr = PMPI_Comm_group((MPI_Comm) *comm, &group);
        *ierr = PMPI_Group_size((MPI_Group) group, &groupsize);

	x_recvcounts = malloc (groupsize * sizeof(int) ) ;

	for (i=0;i<groupsize;i++) {
	  x_recvcounts[i] = recvcounts[i];
	}

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Reduce_scatter with *datatype = %i, x_datatype = %i\n",(int) *datatype, x_datatype);
#endif
        *ierr = (fortran_int) PMPI_Reduce_scatter( (void *) sendbuf, (void *) recvbuf, x_recvcounts, (MPI_Datatype) x_datatype, (MPI_Op) *op, (MPI_Comm) *comm);

	free(x_recvcounts);
}

//   
//   MPI_SCAN( sendbuf, recvbuf, count, datatype, op, comm )
//   [ IN sendbuf] starting address of send buffer (choice)
//   [ OUT recvbuf] starting address of receive buffer (choice)
//   [ IN count] number of elements in input buffer (integer)
//   [ IN datatype] data type of elements of input buffer (handle)
//   [ IN op] operation (handle)
//   [ IN comm] communicator (handle)
//   
//   int MPI_Scan(void* sendbuf, void* recvbuf, int count, MPI_Datatype
//   datatype, MPI_Op op, MPI_Comm comm )
//   
//   MPI_SCAN(SENDBUF, RECVBUF, COUNT, DATATYPE, OP, COMM, IERROR)
//   <type> SENDBUF(*), RECVBUF(*)
//   INTEGER COUNT, DATATYPE, OP, COMM, IERROR
//   
void
mpi_scan_(
void         *sendbuf,
void         *recbuf,
fortran_int  *count,
fortran_int  *type,
fortran_int  *op,	// MPI_Op
fortran_int  *comm,
fortran_int  *ierr)
{
MPI_Datatype x_type = convert_type(*type);
#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Scan with *type = %i, x_type = %i\n",(int) *type, x_type);
#endif
        *ierr = (fortran_int) PMPI_Scan( (void *) sendbuf, (void *) recbuf, (int) *count, (MPI_Datatype) x_type, (MPI_Op) *op, (MPI_Comm) *comm);
}

//   
//   MPI_GROUP_SIZE(group, size)
//   [ IN group] group (handle)
//   [ OUT size] number of processes in the group (integer)
//   
//   int MPI_Group_size(MPI_Group group, int *size)
//   
//   MPI_GROUP_SIZE(GROUP, SIZE, IERROR)
//   INTEGER GROUP, SIZE, IERROR
//   
void
mpi_group_size_(
fortran_int  *group,
fortran_int  *size,
fortran_int  *ierr)
{
int x_size;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Group_size\n");
#endif
        *ierr = (fortran_int) PMPI_Group_size((MPI_Group) *group, &x_size);
        *size = (fortran_int) x_size ;
}

//   
//   MPI_GROUP_RANK(group, rank)
//   [ IN group] group (handle)
//   [ OUT rank] rank of the calling process in group, or
//   MPI_UNDEFINED if the process is not a member (integer)
//   
//   int MPI_Group_rank(MPI_Group group, int *rank)
//   
//   MPI_GROUP_RANK(GROUP, RANK, IERROR)
//   INTEGER GROUP, RANK, IERROR
//   
void
mpi_group_rank_(
fortran_int  *group,
fortran_int  *rank,
fortran_int  *ierr)
{
int x_rank;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Group_rank\n");
#endif
        *ierr = (fortran_int) PMPI_Group_rank((MPI_Group) *group, &x_rank);
        *rank = (fortran_int) x_rank ;
}

//
//   MPI_GROUP_TRANSLATE_RANKS (group1, n, ranks1, group2, ranks2)
//   [ IN group1] group1 (handle)
//   [ IN n] number of ranks in ranks1 and ranks2 arrays (integer)
//   [ IN ranks1] array of zero or more valid ranks in group1
//   [ IN group2] group2 (handle)
//   [ OUT ranks2] array of corresponding ranks in group2, MPI_UNDE- FINED
//   when no correspondence exists.
//   
//   int MPI_Group_translate_ranks (MPI_Group group1, int n, int *ranks1,
//   MPI_Group group2, int *ranks2)
//   
//   MPI_GROUP_TRANSLATE_RANKS(GROUP1, N, RANKS1, GROUP2, RANKS2, IERROR)
//   INTEGER GROUP1, N, RANKS1(*), GROUP2, RANKS2(*), IERROR
//   
void
mpi_group_translate_ranks_(
fortran_int  *group1,
fortran_int  *n,
fortran_int  ranks1[],
fortran_int  *group2,
fortran_int  ranks2[],
fortran_int  *ierr)
{
int x_ranks1[ (int) *n];
int x_ranks2[ (int) *n];
int i;

	for (i=0;i< (int) *n;i++) {
	  x_ranks1[i] = ranks1[i];
	}

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Group_translate_ranks\n");
#endif

        *ierr = (fortran_int) PMPI_Group_translate_ranks((MPI_Group) *group1, (int) *n, x_ranks1, (MPI_Group) *group2, x_ranks2);

	for (i=0;i< (int) *n;i++) {
	  ranks2[i] = x_ranks2[i];
	}
}

//
//   MPI_GROUP_COMPARE(group1, group2, result)
//   [ IN group1] first group (handle)
//   [ IN group2] second group (handle)
//   [ OUT result] result (integer)
//   
//   int MPI_Group_compare(MPI_Group group1,MPI_Group group2, int *result)
//   
//   MPI_GROUP_COMPARE(GROUP1, GROUP2, RESULT, IERROR)
//   INTEGER GROUP1, GROUP2, RESULT, IERROR
//   
void
mpi_group_compare_(
fortran_int  *group1,
fortran_int  *group2,
fortran_int  *result,
fortran_int  *ierr)
{
int x_result;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Group_compare\n");
#endif
        *ierr = (fortran_int) PMPI_Group_compare((MPI_Group) *group1, (MPI_Group) *group2, &x_result);
        *result = (fortran_int) x_result ;
}

//   
//   MPI_COMM_GROUP(comm, group)
//   [ IN comm] communicator (handle)
//   [ OUT group] group corresponding to comm (handle)
//   
//   int MPI_Comm_group(MPI_Comm comm, MPI_Group *group)
//   
//   MPI_COMM_GROUP(COMM, GROUP, IERROR)
//   INTEGER COMM, GROUP, IERROR
//   
void
mpi_comm_group_(
fortran_int  *comm,
fortran_int  *group,
fortran_int  *ierr)
{
MPI_Group x_group;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Comm_group\n");
#endif
        *ierr = (fortran_int) PMPI_Comm_group((MPI_Comm) *comm, &x_group);
        *group = (fortran_int) x_group ;
}

//   
//   MPI_GROUP_UNION(group1, group2, newgroup)
//   [ IN group1] first group (handle)
//   [ IN group2] second group (handle)
//   [ OUT newgroup] union group (handle)
//   
//   int MPI_Group_union(MPI_Group group1, MPI_Group group2, MPI_Group
//   *newgroup)
//   
//   MPI_GROUP_UNION(GROUP1, GROUP2, NEWGROUP, IERROR)
//   INTEGER GROUP1, GROUP2, NEWGROUP, IERROR
//   
void
mpi_group_union_(
fortran_int  *group1,
fortran_int  *group2,
fortran_int  *newgroup,
fortran_int  *ierr)
{
MPI_Group x_newgroup;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Group_union\n");
#endif
        *ierr = (fortran_int) PMPI_Group_union((MPI_Group) *group1, (MPI_Group) *group2, &x_newgroup);
        *newgroup = (fortran_int) x_newgroup ;
}

//   
//   MPI_GROUP_INTERSECTION(group1, group2, newgroup)
//   [ IN group1] first group (handle)
//   [ IN group2] second group (handle)
//   [ OUT newgroup] intersection group (handle)
//   
//   int MPI_Group_intersection(MPI_Group group1, MPI_Group group2,
//   MPI_Group *newgroup)
//   
//   MPI_GROUP_INTERSECTION(GROUP1, GROUP2, NEWGROUP, IERROR)
//   INTEGER GROUP1, GROUP2, NEWGROUP, IERROR
//   
void
mpi_group_intersection_(
fortran_int  *group1,
fortran_int  *group2,
fortran_int  *newgroup,
fortran_int  *ierr)
{
MPI_Group x_newgroup;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Group_intersection\n");
#endif
        *ierr = (fortran_int) PMPI_Group_intersection((MPI_Group) *group1, (MPI_Group) *group2, &x_newgroup);
        *newgroup = (fortran_int) x_newgroup ;
}

//
//   MPI_GROUP_DIFFERENCE(group1, group2, newgroup)
//   [ IN group1] first group (handle)
//   [ IN group2] second group (handle)
//   [ OUT newgroup] difference group (handle)
//   
//   int MPI_Group_difference(MPI_Group group1, MPI_Group group2, MPI_Group
//   *newgroup)
//   
//   MPI_GROUP_DIFFERENCE(GROUP1, GROUP2, NEWGROUP, IERROR)
//   INTEGER GROUP1, GROUP2, NEWGROUP, IERROR
//   
void
mpi_group_difference_(
fortran_int  *group1,
fortran_int  *group2,
fortran_int  *newgroup,
fortran_int  *ierr)
{
MPI_Group x_newgroup;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Group_difference\n");
#endif
        *ierr = (fortran_int) PMPI_Group_difference((MPI_Group) *group1, (MPI_Group) *group2, &x_newgroup);
        *newgroup = (fortran_int) x_newgroup ;
}

//   
//   MPI_GROUP_INCL(group, n, ranks, newgroup)
//   [ IN group] group (handle)
//   [ IN n] number of elements in array ranks (and size of newgroup)
//   (integer)
//   [ IN ranks] ranks of processes in group to appear in newgroup (array
//   of integers)
//   [ OUT newgroup] new group derived from above, in the order defined by
//   ranks (handle)
//   
//   int MPI_Group_incl(MPI_Group group, int n, int *ranks, MPI_Group
//   *newgroup)
//   
//   MPI_GROUP_INCL(GROUP, N, RANKS, NEWGROUP, IERROR)
//   INTEGER GROUP, N, RANKS(*), NEWGROUP, IERROR
//   
void
mpi_group_incl_(
fortran_int  *group,
fortran_int  *n,
fortran_int  ranks[],
fortran_int  *newgroup,
fortran_int  *ierr)
{
MPI_Group x_newgroup;
int x_ranks[ (int) *n];
int i;

	for (i=0;i<*n;i++) {
	  x_ranks[i] = ranks[i];
	}

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Group_incl\n");
#endif
	*ierr = (fortran_int) PMPI_Group_incl( (MPI_Group) *group, (int) *n, x_ranks, &x_newgroup) ;
        *newgroup = (fortran_int) x_newgroup ;

	for (i=0;i<*n;i++) {
	  ranks[i] = (fortran_int) x_ranks[i];
	}

}

//   
//   MPI_GROUP_EXCL(group, n, ranks, newgroup)
//   [ IN group] group (handle)
//   [ IN n] number of elements in array ranks (integer)
//   [ IN ranks] array of integer ranks in group not to appear in newgroup
//   [ OUT newgroup] new group derived from above, preserving the order
//   defined by group (handle)
//   
//   int MPI_Group_excl(MPI_Group group, int n, int *ranks, MPI_Group
//   *newgroup)
//   
//   MPI_GROUP_EXCL(GROUP, N, RANKS, NEWGROUP, IERROR)
//   INTEGER GROUP, N, RANKS(*), NEWGROUP, IERROR
//   
void
mpi_group_excl_(
fortran_int  *group,
fortran_int  *n,
fortran_int  ranks[],
fortran_int  *newgroup,
fortran_int  *ierr)
{
MPI_Group x_newgroup;
int x_ranks[ (int) *n];
int i;

	for (i=0;i<*n;i++) {
	  x_ranks[i] = ranks[i];
	}

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Group_excl\n");
#endif
	*ierr = (fortran_int) PMPI_Group_excl( (MPI_Group) *group, (int) *n, x_ranks, &x_newgroup) ;
        *newgroup = (fortran_int) x_newgroup ;

	for (i=0;i<*n;i++) {
	  ranks[i] = (fortran_int) x_ranks[i];
	}

}

//   
//   MPI_GROUP_RANGE_INCL(group, n, ranges, newgroup)
//   [ IN group] group (handle)
//   [ IN n] number of triplets in array ranges (integer)
//   [ IN ranges] an array of integer triplets, of the form (first rank,
//   last rank, stride) indicating ranks in group of processes to be
//   included in newgroup
//   [ OUT newgroup] new group derived from above, in the order defined by
//   ranges (handle)
//   
//   int MPI_Group_range_incl(MPI_Group group, int n, int ranges[][3],
//   MPI_Group *newgroup)
//   
//   MPI_GROUP_RANGE_INCL(GROUP, N, RANGES, NEWGROUP, IERROR)
//   INTEGER GROUP, N, RANGES(3,*), NEWGROUP, IERROR
//
void
mpi_group_range_incl_(
fortran_int  *group,
fortran_int  *n,
fortran_int  ranges[][3],
fortran_int  *newgroup,
fortran_int  *ierr)
{
MPI_Group x_newgroup;
int x_ranges[ (int) *n][3];
int i, j;

	for (i=0;i<*n;i++) {
	  for (j=0;j<3;j++) {
		x_ranges[i][j] = ranges[i][j];
	  }
	}

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Group_range_incl\n");
#endif
	*ierr = (fortran_int) PMPI_Group_range_incl( (MPI_Group) *group, (int) *n, x_ranges, &x_newgroup) ;
        *newgroup = (fortran_int) x_newgroup ;

	for (i=0;i<*n;i++) {
	  for (j=0;j<3;j++) {
		ranges[i][j] = (fortran_int) x_ranges[i][j];
	  }
	}

}

//   
//   MPI_GROUP_RANGE_EXCL(group, n, ranges, newgroup)
//   [ IN group] group (handle)
//   [ IN n] number of elements in array ranks (integer)
//   [ IN ranges] a one-dimensional array of integer triplets of the form
//   (first rank, last rank, stride), indicating the ranks in group of
//   processes to be excluded from the output group newgroup.
//   [ OUT newgroup] new group derived from above, preserving the order in
//   group (handle)
//   
//   int MPI_Group_range_excl(MPI_Group group, int n, int ranges[][3],
//   MPI_Group *newgroup)
//   
//   MPI_GROUP_RANGE_EXCL(GROUP, N, RANGES, NEWGROUP, IERROR)
//   INTEGER GROUP, N, RANGES(3,*), NEWGROUP, IERROR
//   
void
mpi_group_range_excl_(
fortran_int  *group,
fortran_int  *n,
fortran_int  ranges[][3],
fortran_int  *newgroup,
fortran_int  *ierr)
{
MPI_Group x_newgroup;
int x_ranges[ (int) *n][3];
int i, j;

	for (i=0;i<*n;i++) {
	  for (j=0;j<3;j++) {
		x_ranges[i][j] = ranges[i][j];
	  }
	}

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Group_range_excl\n");
#endif
	*ierr = (fortran_int) PMPI_Group_range_excl( (MPI_Group) *group, (int) *n, x_ranges, &x_newgroup) ;
        *newgroup = (fortran_int) x_newgroup ;

	for (i=0;i<*n;i++) {
	  for (j=0;j<3;j++) {
		ranges[i][j] = (fortran_int) x_ranges[i][j];
	  }
	}

}

//   
//   MPI_GROUP_FREE(group)
//   [ INOUT group] group (handle)
//   
//   int MPI_Group_free(MPI_Group *group)
//   
//   MPI_GROUP_FREE(GROUP, IERROR)
//   INTEGER GROUP, IERROR
//
void
mpi_group_free_(
fortran_int  *group,
fortran_int  *ierr)
{
MPI_Group x_group = *group;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Group_free\n");
#endif
        *ierr = (fortran_int) PMPI_Group_free(&x_group);
        *group = (fortran_int) x_group ;
}

//   
//   MPI_COMM_SIZE(comm, size)
//   [ IN comm] communicator (handle)
//   [ OUT size] number of processes in the group of comm (integer)
//   
//   int MPI_Comm_size(MPI_Comm comm, int *size)
//   
//   MPI_COMM_SIZE(COMM, SIZE, IERROR)
//   INTEGER COMM, SIZE, IERROR
//   
void
mpi_comm_size_(
fortran_int  *comm,
fortran_int  *size,
fortran_int  *ierr)
{
int x_size ;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Comm_size\n");
#endif
        *ierr = (fortran_int) PMPI_Comm_size((MPI_Comm) *comm, &x_size);

*size = (fortran_int) x_size ;
#ifdef DEBUG
  printf("wrapper.c... output from PMPI_Comm_size: *size = %i...\n",(int) *size);
#endif
}

//   
//   MPI_COMM_RANK(comm, rank)
//   [ IN comm] communicator (handle)
//   [ OUT rank] rank of the calling process in group of comm (integer)
//   
//   int MPI_Comm_rank(MPI_Comm comm, int *rank)
//   
//   MPI_COMM_RANK(COMM, RANK, IERROR)
//   INTEGER COMM, RANK, IERROR
//   
void
mpi_comm_rank_(
fortran_int  *comm,
fortran_int  *rank,
fortran_int  *ierr)
{
int x_rank ;

#ifdef DEBUG
  printf("wrapper.c... input to	PMPI_Comm_rank: *comm = %i...\n",(int) *comm);
#endif
        *ierr = (fortran_int) PMPI_Comm_rank((MPI_Comm) *comm, &x_rank);
        *rank = (fortran_int) x_rank ;
#ifdef DEBUG
  printf("wrapper.c... output from PMPI_Comm_rank: *rank = %i...\n",(int) *rank);
#endif
}

//
//   MPI_COMM_COMPARE(comm1, comm2, result)
//   [ IN comm1] first communicator (handle)
//   [ IN comm2] second communicator (handle)
//   [ OUT result] result (integer)
//   
//   int MPI_Comm_compare(MPI_Comm comm1,MPI_Comm comm2, int *result)
//   
//   MPI_COMM_COMPARE(COMM1, COMM2, RESULT, IERROR)
//   INTEGER COMM1, COMM2, RESULT, IERROR
//   
void
mpi_comm_compare_(
fortran_int  *comm1,
fortran_int  *comm2,
fortran_int  *result,
fortran_int  *ierr)
{
int x_result;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Comm_compare\n");
#endif
        *ierr = (fortran_int) PMPI_Comm_compare((MPI_Comm) *comm1,(MPI_Comm) *comm2, &x_result);
        *result = (fortran_int) x_result ;
}

//   
//   MPI_COMM_DUP(comm, newcomm)
//   [ IN comm] communicator (handle)
//   [ OUT newcomm] copy of comm (handle)
//   
//   int MPI_Comm_dup(MPI_Comm comm, MPI_Comm *newcomm)
//   
//   MPI_COMM_DUP(COMM, NEWCOMM, IERROR)
//   INTEGER COMM, NEWCOMM, IERROR
//   
void
mpi_comm_dup_(
fortran_int  *comm,
fortran_int  *newcomm,
fortran_int  *ierr)
{
MPI_Comm x_newcomm;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Comm_dup\n");
#endif
        *ierr = (fortran_int) PMPI_Comm_dup((MPI_Comm) *comm, &x_newcomm);
        *newcomm = (fortran_int) x_newcomm ;
}

//   
//   MPI_COMM_CREATE(comm, group, newcomm)
//   [ IN comm] communicator (handle)
//   [ IN group] Group, which is a subset of the group of comm (handle)
//   [ OUT newcomm] new communicator (handle)
//   
//   int MPI_Comm_create(MPI_Comm comm, MPI_Group group, MPI_Comm *newcomm)
//   
//   MPI_COMM_CREATE(COMM, GROUP, NEWCOMM, IERROR)
//   INTEGER COMM, GROUP, NEWCOMM, IERROR
//   
void
mpi_comm_create_(
fortran_int  *comm,
fortran_int  *group,
fortran_int  *newcomm,
fortran_int  *ierr)
{
MPI_Comm x_newcomm;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Comm_create\n");
#endif
        *ierr = (fortran_int) PMPI_Comm_create((MPI_Comm) *comm, (MPI_Group) *group, &x_newcomm);
        *newcomm = (fortran_int) x_newcomm ;
}

//   
//   MPI_COMM_SPLIT(comm, color, key, newcomm)
//   [ IN comm] communicator (handle)
//   [ IN color] control of subset assignment (integer)
//   [ IN key] control of rank assigment (integer)
//   [ OUT newcomm] new communicator (handle)
//   
//   int MPI_Comm_split(MPI_Comm comm, int color, int key, MPI_Comm
//   *newcomm)
//   
//   MPI_COMM_SPLIT(COMM, COLOR, KEY, NEWCOMM, IERROR)
//   INTEGER COMM, COLOR, KEY, NEWCOMM, IERROR
//   
void
mpi_comm_split_(
fortran_int  *comm,
fortran_int  *color,
fortran_int  *key,
fortran_int  *newcomm,
fortran_int  *ierr)
{
MPI_Comm x_newcomm;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Comm_split\n");
#endif
        *ierr = (fortran_int) PMPI_Comm_split((MPI_Comm) *comm, (int) *color, (int) *key, &x_newcomm);
        *newcomm = (fortran_int) x_newcomm ;
}

//   
//   MPI_COMM_FREE(comm)
//   [ INOUT comm] communicator to be destroyed (handle)
//   
//   int MPI_Comm_free(MPI_Comm *comm)
//   
//   MPI_COMM_FREE(COMM, IERROR)
//   INTEGER COMM, IERROR
//   
void
mpi_comm_free_(
fortran_int  *comm,
fortran_int  *ierr)
{
MPI_Comm x_comm = *comm;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Comm_free\n");
#endif
        *ierr = (fortran_int) PMPI_Comm_free(&x_comm);
        *comm = (fortran_int) x_comm ;
}

//
//   MPI_COMM_TEST_INTER(comm, flag)
//   [ IN comm] communicator (handle)
//   [ OUT flag] (logical)
//   
//   int MPI_Comm_test_inter(MPI_Comm comm, int *flag)
//   
//   MPI_COMM_TEST_INTER(COMM, FLAG, IERROR)
//   INTEGER COMM, IERROR
//   LOGICAL FLAG
//   
void
mpi_comm_test_inter_(
fortran_int  *comm,
fortran_int  *flag,
fortran_int  *ierr)
{
int x_flag ;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Comm__test_inter\n");
#endif
        *ierr = (fortran_int) PMPI_Comm_test_inter((MPI_Comm) *comm, &x_flag);
        *flag = (fortran_int) x_flag ;
}

//   
//   MPI_COMM_REMOTE_SIZE(comm, size)
//   [ IN comm] inter-communicator (handle)
//   [ OUT size] number of processes in the remote group of comm (integer)
//   
//   int MPI_Comm_remote_size(MPI_Comm comm, int *size)
//   
//   MPI_COMM_REMOTE_SIZE(COMM, SIZE, IERROR)
//   INTEGER COMM, SIZE, IERROR
//   
void
mpi_comm_remote_size_(
fortran_int  *comm,
fortran_int  *size,
fortran_int  *ierr)
{
int x_size ;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Comm__remote_size\n");
#endif
        *ierr = (fortran_int) PMPI_Comm_remote_size((MPI_Comm) *comm, &x_size);
        *size = (fortran_int) x_size ;
}

//   
//   MPI_COMM_REMOTE_GROUP(comm, group)
//   [ IN comm] inter-communicator (handle)
//   [ OUT group] remote group corresponding to comm (handle)
//   
//   int MPI_Comm_remote_group(MPI_Comm comm, MPI_Group *group)
//   
//   MPI_COMM_REMOTE_GROUP(COMM, GROUP, IERROR)
//   INTEGER COMM, GROUP, IERROR
//   
void
mpi_comm_remote_group_(
fortran_int  *comm,
fortran_int  *group,
fortran_int  *ierr)
{
// MPI_Group x_group = *group;
MPI_Group x_group;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Comm__remote_group\n");
#endif
        *ierr = (fortran_int) PMPI_Comm_remote_group((MPI_Comm) *comm, &x_group);
        *group = (fortran_int) x_group ;
}

//
// MPI_INIT()
//
// int MPI_Init(int *argc, char ***argv)
//
// MPI_INIT(IERROR)
// INTEGER IERROR
//
void
mpi_init_(
fortran_int *ierr)
{
#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Init...\n");
#endif
        *ierr = (fortran_int) PMPI_Init((void *)0, (void *)0);
}

//
// MPI_INITIALIZED( flag )
// [ OUT flag] Flag is true if MPI_INIT has been called and false
// otherwise.
//
// int MPI_Initialized(int *flag)
//
// MPI_INITIALIZED(FLAG, IERROR)
// LOGICAL FLAG
// INTEGER IERROR
//
void
mpi_initialized_(
fortran_int *flag,
fortran_int *ierr)
{
int x_flag;
#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Initialized...\n");
#endif
        *ierr = (fortran_int) PMPI_Initialized( &x_flag);

        *flag = (fortran_int) x_flag ;
}

//   
//   MPI_INTERCOMM_CREATE(local_comm, local_leader, peer_comm,
//   remote_leader, tag, newintercomm)
//   [ IN local_comm ] local intra-communicator (handle)
//   [ IN local_leader] rank of local group leader in local_comm (integer)
//   [ IN peer_comm] ``peer'' communicator; significant only at the
//   local_leader (handle)
//   [ IN remote_leader] rank of remote group leader in peer_comm;
//   significant only at the local_leader (integer)
//   [ IN tag] ``safe'' tag (integer)
//   [ OUT newintercomm ] new inter-communicator (handle)
//   
//   int MPI_Intercomm_create(MPI_Comm local_comm, int local_leader,
//   MPI_Comm peer_comm, int remote_leader, int tag, MPI_Comm
//   *newintercomm)
//   
//   MPI_INTERCOMM_CREATE(LOCAL_COMM, LOCAL_LEADER, PEER_COMM,
//   REMOTE_LEADER, TAG, NEWINTERCOMM, IERROR)
//   INTEGER LOCAL_COMM, LOCAL_LEADER, PEER_COMM, REMOTE_LEADER, TAG,
//   NEWINTERCOMM, IERROR
//   
void
mpi_intercomm_create_(
fortran_int  *local_comm,
fortran_int  *local_leader,
fortran_int  *peer_comm,
fortran_int  *remote_leader,
fortran_int  *tag,
fortran_int  *newintracomm,
fortran_int  *ierr)
{
MPI_Comm x_newintracomm;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Intercomm_create\n");
#endif
        *ierr = (fortran_int) PMPI_Intercomm_create( (MPI_Comm) *local_comm, (int) *local_leader, (MPI_Comm) *peer_comm, (int) *remote_leader, (int) *tag, &x_newintracomm );
        *newintracomm = (fortran_int) x_newintracomm ;
}

//
//   MPI_INTERCOMM_MERGE(intercomm, high, newintracomm)
//   [ IN intercomm] Inter-Communicator (handle)
//   [ IN high] (logical)
//   [ OUT newintracomm] new intra-communicator (handle)
//   
//   int MPI_Intercomm_merge(MPI_Comm intercomm, int high, MPI_Comm
//   *newintracomm)
//   
//   MPI_INTERCOMM_MERGE(INTERCOMM, HIGH, INTRACOMM, IERROR)
//   INTEGER INTERCOMM, INTRACOMM, IERROR
//   LOGICAL HIGH
//   
void
mpi_intercomm_merge_(
fortran_int  *intercomm,
fortran_int  *high,
fortran_int  *newintracomm,
fortran_int  *ierr)
{
MPI_Comm x_newintracomm;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Intercomm_merge\n");
#endif
        *ierr = (fortran_int) PMPI_Intercomm_merge( (MPI_Comm) *intercomm, (int) *high, &x_newintracomm );
        *newintracomm = (fortran_int) x_newintracomm ;
}

//   
//   MPI_KEYVAL_CREATE(copy_fn, delete_fn, keyval, extra_state)
//   [ IN copy_fn] Copy callback function for keyval
//   [ IN delete_fn] Delete callback function for keyval
//   [ OUT keyval] key value for future access (integer)
//   [ IN extra_state] Extra state for callback functions
//   
//   int MPI_Keyval_create(MPI_Copy_function *copy_fn, MPI_Delete_function
//   *delete_fn, int *keyval, void* extra_state)
//   
//   MPI_KEYVAL_CREATE(COPY_FN, DELETE_FN, KEYVAL, EXTRA_STATE, IERROR)
//   EXTERNAL COPY_FN, DELETE_FN
//   INTEGER KEYVAL, EXTRA_STATE, IERROR
//   
void
mpi_keyval_create_(
void *copy_fn,
void *delete_fn,
fortran_int  *keyval,
void *extra_state,
fortran_int  *ierr)
{
int x_keyval;

#ifdef DEBUG
  printf("wrapper.c... input to PMPI_Keyval_create: *keyval = %i\n",(int) *keyval);
#endif
        *ierr = (fortran_int) PMPI_Keyval_create( (MPI_Copy_function *) copy_fn, (MPI_Delete_function *) delete_fn, &x_keyval, (void *) extra_state);
        *keyval = (fortran_int) x_keyval ;
}

//   
//   MPI_KEYVAL_FREE(keyval)
//   [ INOUT keyval] Frees the integer key value (integer)
//   
//   int MPI_Keyval_free(int *keyval)
//   
//   MPI_KEYVAL_FREE(KEYVAL, IERROR)
//   INTEGER KEYVAL, IERROR
//   
void
mpi_keyval_free_(
fortran_int  *keyval,
fortran_int  *ierr)
{
int x_keyval = *keyval;
#ifdef DEBUG
  printf("wrapper.c... input to PMPI_Keyval_free: *keyval = %i\n",(int) *keyval);
#endif
        *ierr = (fortran_int) PMPI_Keyval_free( &x_keyval );
        *keyval = (fortran_int) x_keyval ;
}

//   
//   MPI_ATTR_PUT(comm, keyval, attribute_val)
//   [ IN comm] communicator to which attribute will be attached (handle)
//   [ IN keyval] key value, as returned by MPI_KEYVAL_CREATE (integer)
//   [ IN attribute_val] attribute value
//   
//   int MPI_Attr_put(MPI_Comm comm, int keyval, void* attribute_val)
//   
//   MPI_ATTR_PUT(COMM, KEYVAL, ATTRIBUTE_VAL, IERROR)
//   INTEGER COMM, KEYVAL, ATTRIBUTE_VAL, IERROR
//   
void
mpi_attr_put_(
fortran_int  *comm,
fortran_int  *keyval,
fortran_int  *attribute_val,
fortran_int  *ierr)
{
#ifdef DEBUG
  printf("wrapper.c... input to PMPI_Attr_put: *keyval = %i\n",(int) *keyval);
#endif
        *ierr = (fortran_int) PMPI_Attr_put( (MPI_Comm) *comm, (int) *keyval, (void *) attribute_val );
}

//   
//   MPI_ATTR_GET(comm, keyval, attribute_val, flag)
//   [ IN comm] communicator to which attribute is attached (handle)
//   [ IN keyval] key value (integer)
//   [ OUT attribute_val] attribute value, unless flag = false
//   [ OUT flag] true if an attribute value was extracted; false if no
//   attribute is associated with the key
//   
//   int MPI_Attr_get(MPI_Comm comm, int keyval, void *attribute_val, int
//   *flag)
//   
//   MPI_ATTR_GET(COMM, KEYVAL, ATTRIBUTE_VAL, FLAG, IERROR)
//   INTEGER COMM, KEYVAL, ATTRIBUTE_VAL, IERROR
//   LOGICAL FLAG
//   
void
mpi_attr_get_(
fortran_int  *comm,
fortran_int  *keyval,
fortran_int  *attribute_val,
fortran_int  *flag,
fortran_int  *ierr)
{
void *x_attribute_val;
int x_flag;

#ifdef DEBUG
  printf("wrapper.c... input to PMPI_Attr_get: *keyval = %i\n",(int) *keyval);
#endif
        *ierr = (fortran_int) PMPI_Attr_get( (MPI_Comm) *comm, (int) *keyval,  (void *) &x_attribute_val, &x_flag );
        *flag = (fortran_int) x_flag ;
        *attribute_val = (fortran_int) x_attribute_val ;
}

//   
//   MPI_ATTR_DELETE(comm, keyval)
//   [ IN comm] communicator to which attribute is attached (handle)
//   [ IN keyval] The key value of the deleted attribute (integer)
//   
//   int MPI_Attr_delete(MPI_Comm comm, int keyval)
//   
//   MPI_ATTR_DELETE(COMM, KEYVAL, IERROR)
//   INTEGER COMM, KEYVAL, IERROR
//   
void
mpi_attr_delete_(
fortran_int  *comm,
fortran_int  *keyval,
fortran_int  *ierr)
{
#ifdef DEBUG
  printf("wrapper.c... input to PMPI_Attr_delete: *keyval = %i\n",(int) *keyval);
#endif
        *ierr = (fortran_int) PMPI_Attr_delete( (MPI_Comm) *comm, (int) *keyval );
}

//   
//   MPI_CART_CREATE(comm_old, ndims, dims, periods, reorder, comm_cart)
//   [ IN comm_old] input communicator (handle)
//   [ IN ndims] number of dimensions of cartesian grid (integer)
//   [ IN dims] integer array of size ndims specifying the number of
//   processes in each dimension
//   [ IN periods] logical array of size ndims specifying whether the grid
//   is periodic ( true) or not ( false) in each dimension
//   [ IN reorder] ranking may be reordered ( true) or not ( false)
//   (logical)
//   [ OUT comm_cart] communicator with new cartesian topology (handle)
//   
//   int MPI_Cart_create(MPI_Comm comm_old, int ndims, int *dims, int
//   *periods, int reorder, MPI_Comm *comm_cart)
//   
//   MPI_CART_CREATE(COMM_OLD, NDIMS, DIMS, PERIODS, REORDER, COMM_CART,
//   IERROR)
//   INTEGER COMM_OLD, NDIMS, DIMS(*), COMM_CART, IERROR
//   LOGICAL PERIODS(*), REORDER
//   
void
mpi_cart_create_(
fortran_int  *comm_old,
fortran_int  *ndims,
fortran_int  dims[],
fortran_int  periods[],
fortran_int  *reorder,
fortran_int  *comm_cart,
fortran_int  *ierr)
{
MPI_Comm x_comm_cart;
int x_dims[ (int) *ndims];
int x_periods[ (int) *ndims];
int i;

	for (i=0;i<*ndims;i++) {
	  x_dims[i] = dims[i];
	}
	for (i=0;i<*ndims;i++) {
	  x_periods[i] = periods[i];
	}

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Cart_create\n");
#endif
	*ierr = (fortran_int) PMPI_Cart_create( (MPI_Comm) *comm_old, (int) *ndims, x_dims, x_periods, (int) *reorder, &x_comm_cart);

        *comm_cart = (fortran_int) x_comm_cart ;
}

//   
//   MPI_DIMS_CREATE(nnodes, ndims, dims)
//   [ IN nnodes] number of nodes in a grid (integer)
//   [ IN ndims] number of cartesian dimensions (integer)
//   [ INOUT dims] integer array of size ndims specifying the number of
//   nodes in each dimension
//   
//   int MPI_Dims_create(int nnodes, int ndims, int *dims)
//   
//   MPI_DIMS_CREATE(NNODES, NDIMS, DIMS, IERROR)
//   INTEGER NNODES, NDIMS, DIMS(*), IERROR
//   
void
mpi_dims_create_(
fortran_int  *nnodes,
fortran_int  *ndims,
fortran_int  dims[],
fortran_int  *ierr)
{
int x_dims[ (int) *ndims];
int i;

	for (i=0;i<*ndims;i++) {
	  x_dims[i] = dims[i];
	}

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Dims_create\n");
#endif
	*ierr = (fortran_int) PMPI_Dims_create( (int) *nnodes, (int) *ndims, x_dims);

	for (i=0;i<*ndims;i++) {
	  dims[i] = (fortran_int) x_dims[i];
	}

}

//   
//   MPI_GRAPH_CREATE(comm_old, nnodes, index, edges, reorder, comm_graph)
//   [ IN comm_old] input communicator (handle)
//   [ IN nnodes] number of nodes in graph (integer)
//   [ IN index] array of integers describing node degrees (see below)
//   [ IN edges] array of integers describing graph edges (see below)
//   [ IN reorder] ranking may be reordered ( true) or not ( false)
//   (logical)
//   [ OUT comm_graph] communicator with graph topology added (handle)
//   
//   int MPI_Graph_create(MPI_Comm comm_old, int nnodes, int *index, int
//   *edges, int reorder, MPI_Comm *comm_graph)
//   
//   MPI_GRAPH_CREATE(COMM_OLD, NNODES, INDEX, EDGES, REORDER, COMM_GRAPH,
//   IERROR)
//   INTEGER COMM_OLD, NNODES, INDEX(*), EDGES(*), COMM_GRAPH, IERROR
//   LOGICAL REORDER
//   
void
mpi_graph_create_(
fortran_int  *comm_old,
fortran_int  *nnodes,
fortran_int  index[],
fortran_int  edges[],
fortran_int  *reorder,
fortran_int  *comm_graph,
fortran_int  *ierr)
{
MPI_Comm x_comm_graph;
int x_index[ (int) *nnodes];
int x_edges[ (int) *nnodes];
int i;

	for (i=0;i<*nnodes;i++) {
	  x_index[i] = index[i];
	}
	for (i=0;i<*nnodes;i++) {
	  x_edges[i] = edges[i];
	}

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Graph_create\n");
#endif
	*ierr = (fortran_int) PMPI_Graph_create( (MPI_Comm) *comm_old, (int) *nnodes, x_index, x_edges, (int) *reorder, &x_comm_graph);

        *comm_graph = (fortran_int) x_comm_graph ;
}

//   
//   MPI_TOPO_TEST(comm, status)
//   [ IN comm] communicator (handle)
//   [ OUT status] topology type of communicator comm (choice)
//   
//   int MPI_Topo_test(MPI_Comm comm, int *status)
//   
//   MPI_TOPO_TEST(COMM, STATUS, IERROR)
//   INTEGER COMM, STATUS, IERROR
//   
void
mpi_topo_test_(
fortran_int  *comm,
fortran_int  *status,
fortran_int  *ierr)
{
int x_status;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Topo_test...\n");
#endif
        *ierr = (fortran_int) PMPI_Topo_test( (MPI_Comm) *comm, &x_status);

        *status = (fortran_int) x_status;
}

//   
//   MPI_GRAPHDIMS_GET(comm, nnodes, nedges)
//   [ IN comm] communicator for group with graph structure (handle)
//   [ OUT nnodes] number of nodes in graph (integer) (same as number of
//   processes in the group)
//   [ OUT nedges] number of edges in graph (integer)
//   
//   int MPI_Graphdims_get(MPI_Comm comm, int *nnodes, int *nedges)
//   
//   MPI_GRAPHDIMS_GET(COMM, NNODES, NEDGES, IERROR)
//   INTEGER COMM, NNODES, NEDGES, IERROR
//   
void
mpi_graphdims_get_(
fortran_int  *comm,
fortran_int  *nnodes,
fortran_int  *nedges,
fortran_int  *ierr)
{
int x_nnodes;
int x_nedges;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Graphdims_get...\n");
#endif
        *ierr = (fortran_int) PMPI_Graphdims_get( (MPI_Comm) *comm, &x_nnodes, &x_nedges);

        *nnodes = (fortran_int) x_nnodes ;
        *nedges = (fortran_int) x_nedges ;
}

//   
//   MPI_GRAPH_GET(comm, maxindex, maxedges, index, edges)
//   [ IN comm] communicator with graph structure (handle)
//   [ IN maxindex] length of vector index in the calling program
//   (integer)
//   [ IN maxedges] length of vector edges in the calling program
//   (integer)
//   [ OUT index] array of integers containing the graph structure (for
//   details see the definition of MPI_GRAPH_CREATE)
//   [ OUT edges] array of integers containing the graph structure
//   
//   int MPI_Graph_get(MPI_Comm comm, int maxindex, int maxedges, int
//   *index, int *edges)
//   
//   MPI_GRAPH_GET(COMM, MAXINDEX, MAXEDGES, INDEX, EDGES, IERROR)
//   INTEGER COMM, MAXINDEX, MAXEDGES, INDEX(*), EDGES(*), IERROR
//   
void
mpi_graph_get_(
fortran_int  *comm_old,
fortran_int  *maxindex,
fortran_int  *maxedges,
fortran_int  index[],
fortran_int  edges[],
fortran_int  *ierr)
{
int x_index[ (int) *maxindex];
int x_edges[ (int) *maxedges];
int i;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Graph_get\n");
#endif
	*ierr = (fortran_int) PMPI_Graph_get( (MPI_Comm) *comm_old, (int) *maxindex, (int) *maxedges, x_index, x_edges);

	for (i=0;i<*maxindex;i++) {
	  index[i] = (fortran_int) x_index[i];
	}
	for (i=0;i<*maxedges;i++) {
	  edges[i] = (fortran_int) x_edges[i];
	}

}

//   MPI_CARTDIM_GET(comm, ndims)
//   [ IN comm] communicator with cartesian structure (handle)
//   [ OUT ndims] number of dimensions of the cartesian structure (integer)
//   
//   int MPI_Cartdim_get(MPI_Comm comm, int *ndims)
//   
//   MPI_CARTDIM_GET(COMM, NDIMS, IERROR)
//   INTEGER COMM, NDIMS, IERROR
//   
void
mpi_cartdim_get_(
fortran_int  *comm,
fortran_int  *ndims,
fortran_int  *ierr)
{
int x_ndims;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Cartdim_get...\n");
#endif
        *ierr = (fortran_int) PMPI_Cartdim_get( (MPI_Comm) *comm, &x_ndims );

        *ndims = (fortran_int) x_ndims ;
}

//   
//   MPI_CART_GET(comm, maxdims, dims, periods, coords)
//   [ IN comm] communicator with cartesian structure (handle)
//   [ IN maxdims] length of vectors dims, periods, and coords in the
//   calling program (integer)
//   [ OUT dims] number of processes for each cartesian dimension (array of
//   integer)
//   [ OUT periods] periodicity ( true/ false) for each cartesian dimension
//   (array of logical)
//   [ OUT coords] coordinates of calling process in cartesian structure
//   (array of integer)
//   
//   int MPI_Cart_get(MPI_Comm comm, int maxdims, int *dims, int *periods,
//   int *coords)
//   
//   MPI_CART_GET(COMM, MAXDIMS, DIMS, PERIODS, COORDS, IERROR)
//   INTEGER COMM, MAXDIMS, DIMS(*), COORDS(*), IERROR
//   LOGICAL PERIODS(*)
//   
void
mpi_cart_get_(
fortran_int  *comm,
fortran_int  *maxdims,
fortran_int  *dims,
fortran_int  *periods,
fortran_int  *coords,
fortran_int  *ierr)
{
int x_dims;
int x_periods;
int x_coords;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Cart_get...\n");
#endif
        *ierr = (fortran_int) PMPI_Cart_get( (MPI_Comm) *comm, (int) *maxdims, &x_dims, &x_periods, &x_coords);

        *dims = (fortran_int) x_dims ;
        *periods = (fortran_int) x_periods ;
        *coords = (fortran_int) x_coords ;
}

//
//   MPI_CART_RANK(comm, coords, rank)
//   [ IN comm] communicator with cartesian structure (handle)
//   [ IN coords] integer array (of size ndims) specifying the cartesian
//   coordinates of a process
//   [ OUT rank] rank of specified process (integer)
//   
//   int MPI_Cart_rank(MPI_Comm comm, int *coords, int *rank)
//   
//   MPI_CART_RANK(COMM, COORDS, RANK, IERROR)
//   INTEGER COMM, COORDS(*), RANK, IERROR
//   
void
mpi_cart_rank_(
fortran_int  *comm,
fortran_int  coords[],
fortran_int  *rank,
fortran_int  *ierr)
{
int *x_coords;
int x_rank;
int i;
int ndims;
        *ierr = PMPI_Cartdim_get((MPI_Comm) *comm, &ndims);
	x_coords = malloc (ndims * sizeof(int) ) ;

	for (i=0;i<ndims;i++) {
	  x_coords[i] = coords[i];
	}

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Cart_rank\n");
#endif
	*ierr = (fortran_int) PMPI_Cart_rank( (MPI_Comm) *comm, x_coords, &x_rank);

	*rank = (fortran_int) x_rank;
}


//   
//   MPI_CART_COORDS(comm, rank, maxdims, coords)
//   [ IN comm] communicator with cartesian structure (handle)
//   [ IN rank] rank of a process within group of comm (integer)
//   [ IN maxdims] length of vector coord in the calling program (integer)
//   [ OUT coords] integer array (of size ndims) containing the cartesian
//   coordinates of specified process (integer)
//   
//   int MPI_Cart_coords(MPI_Comm comm, int rank, int maxdims, int *coords)
//   
//   MPI_CART_COORDS(COMM, RANK, MAXDIMS, COORDS, IERROR)
//   INTEGER COMM, RANK, MAXDIMS, COORDS(*), IERROR
//   
void
mpi_cart_coords_(
fortran_int  *comm,
fortran_int  *rank,
fortran_int  *maxdims,
fortran_int  coords[],
fortran_int  *ierr)
{
int i;
int *x_coords;
int ndims;
        *ierr = PMPI_Cartdim_get((MPI_Comm) *comm, &ndims);
	x_coords = malloc (ndims * sizeof(int) ) ;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Cart_coords...\n");
#endif
        *ierr = (fortran_int) PMPI_Cart_coords( (MPI_Comm) *comm, (int) *rank, (int) *maxdims, x_coords);

	for (i=0;i<ndims;i++) {
	  coords[i] = (fortran_int) x_coords[i];
	}
}

//   
//   MPI_GRAPH_NEIGHBORS_COUNT(comm, rank, nneighbors)
//   [ IN comm] communicator with graph topology (handle)
//   [ IN rank] rank of process in group of comm (integer)
//   [ OUT nneighbors] number of neighbors of specified process (integer)
//   
//   int MPI_Graph_neighbors_count(MPI_Comm comm, int rank, int
//   *nneighbors)
//   
//   MPI_GRAPH_NEIGHBORS_COUNT(COMM, RANK, NNEIGHBORS, IERROR)
//   INTEGER COMM, RANK, NNEIGHBORS, IERROR
//   
void
mpi_graph_neighbors_count_(
fortran_int  *comm,
fortran_int  *rank,
fortran_int  *nneighbors,
fortran_int  *ierr)
{
int x_nneighbors;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Graph_neighbors_count...\n");
#endif
        *ierr = (fortran_int) PMPI_Graph_neighbors_count( (MPI_Comm) *comm, (int) *rank, &x_nneighbors);

        *nneighbors = (fortran_int) x_nneighbors ;
}

//   
//   MPI_GRAPH_NEIGHBORS(comm, rank, maxneighbors, neighbors)
//   [ IN comm] communicator with graph topology (handle)
//   [ IN rank] rank of process in group of comm (integer)
//   [ IN maxneighbors] size of array neighbors (integer)
//   [ OUT neighbors] ranks of processes that are neighbors to specified
//   process (array of integer)
//   
//   int MPI_Graph_neighbors(MPI_Comm comm, int rank, int maxneighbors, int
//   *neighbors)
//   
//   MPI_GRAPH_NEIGHBORS(COMM, RANK, MAXNEIGHBORS, NEIGHBORS, IERROR)
//   INTEGER COMM, RANK, MAXNEIGHBORS, NEIGHBORS(*), IERROR
//   
void
mpi_graph_neighbors_(
fortran_int  *comm,
fortran_int  *rank,
fortran_int  *maxneighbors,
fortran_int  *neighbors,
fortran_int  *ierr)
{
int i;
int *x_neighbors;
int nneighbors;
        *ierr = PMPI_Graph_neighbors_count( (MPI_Comm) *comm, (int) *rank, &nneighbors);
	x_neighbors = malloc (nneighbors * sizeof(int) ) ;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Graph_neighbors...\n");
#endif
        *ierr = (fortran_int) PMPI_Graph_neighbors( (MPI_Comm) *comm, (int) *rank, (int) *maxneighbors, x_neighbors);

	for (i=0;i<nneighbors;i++) {
	  neighbors[i] = (fortran_int) x_neighbors[i];
	}
}

//   
//   MPI_CART_SHIFT(comm, direction, disp, rank_source, rank_dest)
//   [ IN comm] communicator with cartesian structure (handle)
//   [ IN direction] coordinate dimension of shift (integer)
//   [ IN disp] displacement (> 0: upwards shift, < 0: downwards shift)
//   (integer)
//   [ OUT rank_source] rank of source process (integer)
//   [ OUT rank_dest] rank of destination process (integer)
//   
//   int MPI_Cart_shift(MPI_Comm comm, int direction, int disp, int
//   *rank_source, int *rank_dest)
//   
//   MPI_CART_SHIFT(COMM, DIRECTION, DISP, RANK_SOURCE, RANK_DEST, IERROR)
//   INTEGER COMM, DIRECTION, DISP, RANK_SOURCE, RANK_DEST, IERROR
//
void
mpi_cart_shift_(
fortran_int  *comm,
fortran_int  *direction,
fortran_int  *disp,
fortran_int  *rank_source,
fortran_int  *rank_dest,
fortran_int  *ierr)
{
int x_rank_source;
int x_rank_dest;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Cart_shift...\n");
#endif
        *ierr = (fortran_int) PMPI_Cart_shift( (MPI_Comm) *comm, (int) *direction, (int) *disp, &x_rank_source, &x_rank_dest);

        *rank_source = (fortran_int) x_rank_source ;
        *rank_dest = (fortran_int) x_rank_dest ;
}

//   
//   MPI_CART_SUB(comm, remain_dims, newcomm)
//   [ IN comm] communicator with cartesian structure (handle)
//   [ IN remain_dims] the ith entry of remain_dims specifies whether the
//   ith dimension is kept in the subgrid ( true) or is dropped ( false)
//   (logical vector)
//   [ OUT newcomm] communicator containing the subgrid that includes the
//   calling process (handle)
//   
//   int MPI_Cart_sub(MPI_Comm comm, int *remain_dims, MPI_Comm *newcomm)
//   
//   MPI_CART_SUB(COMM, REMAIN_DIMS, NEWCOMM, IERROR)
//   INTEGER COMM, NEWCOMM, IERROR
//   LOGICAL REMAIN_DIMS(*)
//   
void
mpi_cart_sub_(
fortran_int  *comm,
fortran_int  remain_dims[],
fortran_int  *newcomm,
fortran_int  *ierr)
{
MPI_Comm x_newcomm;
int *x_remain_dims;
int i;
int ndims;
        *ierr = PMPI_Cartdim_get((MPI_Comm) *comm, &ndims);
	x_remain_dims = malloc (ndims * sizeof(int) ) ;

	for (i=0;i<ndims;i++) {
	  x_remain_dims[i] = remain_dims[i];
	}

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Cart_sub\n");
#endif
        *ierr = (fortran_int) PMPI_Cart_sub((MPI_Comm) *comm, x_remain_dims, &x_newcomm);
        *newcomm = (fortran_int) x_newcomm ;

	free(x_remain_dims);
}

//   
//   MPI_CART_MAP(comm, ndims, dims, periods, newrank)
//   [ IN comm] input communicator (handle)
//   [ IN ndims] number of dimensions of cartesian structure (integer)
//   [ IN dims] integer array of size ndims specifying the number of
//   processes in each coordinate direction
//   [ IN periods] logical array of size ndims specifying the periodicity
//   specification in each coordinate direction
//   [ OUT newrank] reordered rank of the calling process; MPI_UNDEFINED if
//   calling process does not belong to grid (integer)
//   
//   int MPI_Cart_map(MPI_Comm comm, int ndims, int *dims, int *periods,
//   int *newrank)
//   
//   MPI_CART_MAP(COMM, NDIMS, DIMS, PERIODS, NEWRANK, IERROR)
//   INTEGER COMM, NDIMS, DIMS(*), NEWRANK, IERROR
//   LOGICAL PERIODS(*)
//   
void
mpi_cart_map_(
fortran_int  *comm_old,
fortran_int  *ndims,
fortran_int  dims[],
fortran_int  periods[],
fortran_int  *newrank,
fortran_int  *ierr)
{
int x_newrank;
int x_dims[ (int) *ndims];
int x_periods[ (int) *ndims];
int i;

	for (i=0;i<*ndims;i++) {
	  x_dims[i] = dims[i];
	}
	for (i=0;i<*ndims;i++) {
	  x_periods[i] = periods[i];
	}

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Cart_map\n");
#endif
	*ierr = (fortran_int) PMPI_Cart_map( (MPI_Comm) *comm_old, (int) *ndims, x_dims, x_periods, &x_newrank);

        *newrank = (fortran_int) x_newrank ;
}

//   
//   MPI_GRAPH_MAP(comm, nnodes, index, edges, newrank)
//   [ IN comm] input communicator (handle)
//   [ IN nnodes] number of graph nodes (integer)
//   [ IN index] integer array specifying the graph structure, see
//   MPI_GRAPH_CREATE
//   [ IN edges] integer array specifying the graph structure
//   [ OUT newrank] reordered rank of the calling process; MPI_UNDEFINED if
//   the calling process does not belong to graph (integer)
//   
//   int MPI_Graph_map(MPI_Comm comm, int nnodes, int *index, int *edges,
//   int *newrank)
//   
//   MPI_GRAPH_MAP(COMM, NNODES, INDEX, EDGES, NEWRANK, IERROR)
//   INTEGER COMM, NNODES, INDEX(*), EDGES(*), NEWRANK, IERROR
//
void
mpi_graph_map_(
fortran_int  *comm,
fortran_int  *nnodes,
fortran_int  index[],
fortran_int  edges[],
fortran_int  *newrank,
fortran_int  *ierr)
{
int x_newrank;
int x_index[ (int) *nnodes];
int x_edges[ (int) *nnodes];
int i;

	for (i=0;i<*nnodes;i++) {
	  x_index[i] = index[i];
	}
	for (i=0;i<*nnodes;i++) {
	  x_edges[i] = edges[i];
	}

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Graph_map\n");
#endif
	*ierr = (fortran_int) PMPI_Graph_map( (MPI_Comm) *comm, (int) *nnodes, x_index, x_edges, &x_newrank);

        *newrank = (fortran_int) x_newrank ;
}

//   
//   MPI_GET_PROCESSOR_NAME( name, resultlen )
//   [ OUT name] A unique specifier for the actual (as opposed to virtual)
//   node.
//   [ OUT resultlen] Length (in printable characters) of the result
//   returned in name
//   
//   int MPI_Get_processor_name(char *name, int *resultlen)
//   
//   MPI_GET_PROCESSOR_NAME( NAME, RESULTLEN, IERROR)
//   CHARACTER*(*) NAME
//   INTEGER RESULTLEN,IERROR
//   
void
mpi_get_processor_name_(
char *name,
fortran_int  *resultlen,
fortran_int  *ierr,
int   namelen)
{
char x_name[namelen+1];
int x_resultlen;
        *ierr = (fortran_int) PMPI_Get_processor_name(x_name, &x_resultlen);
        *resultlen = (fortran_int) x_resultlen ;
	memcpy(name, x_name, namelen);
}

//   
//   MPI_ERRHANDLER_CREATE( function, errhandler )
//   [ IN function] user defined error handling procedure
//   [ OUT errhandler] MPI error handler (handle)
//   
//   int MPI_Errhandler_create(MPI_Handler_function *function,
//   MPI_Errhandler *errhandler)
//   
//   MPI_ERRHANDLER_CREATE(FUNCTION, HANDLER, IERROR)
//   EXTERNAL FUNCTION
//   INTEGER ERRHANDLER, IERROR
//
void
mpi_errhandler_create_(
void *function,
fortran_int  *errhandler,
fortran_int  *ierr)
{
MPI_Errhandler x_errhandler;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Errhandler_create...\n");
#endif
        *ierr = (fortran_int) PMPI_Errhandler_create( (MPI_Handler_function *) function, &x_errhandler );

	*errhandler = (fortran_int) x_errhandler;
}

//
//   MPI_ERRHANDLER_SET( comm, errhandler )
//   [ IN comm] communicator to set the error handler for (handle)
//   [ IN errhandler] new MPI error handler for communicator (handle)
//   
//   int MPI_Errhandler_set(MPI_Comm comm, MPI_Errhandler errhandler)
//   
//   MPI_ERRHANDLER_SET(COMM, ERRHANDLER, IERROR)
//   INTEGER COMM, ERRHANDLER, IERROR
//   
void
mpi_errhandler_set_(
fortran_int  *comm,
fortran_int  *errhandler,
fortran_int  *ierr)
{
#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Errhandler_set...\n");
#endif
        *ierr = (fortran_int) PMPI_Errhandler_set( (MPI_Comm) *comm, (MPI_Errhandler) *errhandler );
}

//   
//   MPI_ERRHANDLER_GET( comm, errhandler )
//   [ IN comm] communicator to get the error handler from (handle)
//   [ OUT errhandler] MPI error handler currently associated with
//   communicator (handle)
//   
//   int MPI_Errhandler_get(MPI_Comm comm, MPI_Errhandler *errhandler)
//   
//   MPI_ERRHANDLER_GET(COMM, ERRHANDLER, IERROR)
//   INTEGER COMM, ERRHANDLER, IERROR
//   
void
mpi_errhandler_get_(
fortran_int  *comm,
fortran_int  *errhandler,
fortran_int  *ierr)
{
MPI_Errhandler x_errhandler;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Errhandler_get...\n");
#endif
        *ierr = (fortran_int) PMPI_Errhandler_get( (MPI_Comm) *comm, &x_errhandler );

	*errhandler = (fortran_int) x_errhandler;
}

//   
//   MPI_ERRHANDLER_FREE( errhandler )
//   [ IN errhandler] MPI error handler (handle)
//   
//   int MPI_Errhandler_free(MPI_Errhandler *errhandler)
//   
//   MPI_ERRHANDLER_FREE(ERRHANDLER, IERROR)
//   INTEGER ERRHANDLER, IERROR
//
void
mpi_errhandler_free_(
fortran_int  *errhandler,
fortran_int  *ierr)
{

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Errhandler_free...\n");
#endif
        *ierr = (fortran_int) PMPI_Errhandler_free( (MPI_Errhandler *) errhandler );

}

//   
//   MPI_ERROR_STRING( errorcode, string, resultlen )
//   [ IN errorcode] Error code returned by an MPI routine
//   [ OUT string] Text that corresponds to the errorcode
//   [ OUT resultlen] Length (in printable characters) of the result
//   returned in string
//   
//   int MPI_Error_string(int errorcode, char *string, int *resultlen)
//   
//   MPI_ERROR_STRING(ERRORCODE, STRING, RESULTLEN, IERROR)
//   INTEGER ERRORCODE, RESULTLEN, IERROR
//   CHARACTER*(*) STRING
//   
void
mpi_error_string_(
fortran_int  *errorcode,
char *string,
fortran_int  *resultlen,
fortran_int  *ierr,
fortran_int str_len)
{
char x_string[str_len+1];
int x_resultlen ;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Error_string...\n");
#endif
        *ierr = (fortran_int) PMPI_Error_string( (int) *errorcode, string, &x_resultlen);

        *resultlen = (fortran_int) x_resultlen ;
	memcpy(string, x_string, *resultlen);
}

//                                       
//   MPI_ERROR_CLASS( errorcode, errorclass )
//   [ IN errorcode] Error code returned by an MPI routine
//   [ OUT errorclass] Error class associated with errorcode
//   
//   int MPI_Error_class(int errorcode, int *errorclass)
//   
//   MPI_ERROR_CLASS(ERRORCODE, ERRORCLASS, IERROR)
//   INTEGER ERRORCODE, ERRORCLASS, IERROR
//   
void
mpi_error_class_(
fortran_int  *errorcode,
fortran_int  *errorclass,
fortran_int  *ierr)
{
int x_errorclass;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Error_class...\n");
#endif
        *ierr = (fortran_int) PMPI_Error_class( (int) *errorcode, &x_errorclass);

        *errorclass = (fortran_int) x_errorclass ;
}

//
// Wrapper routines from the MPI-2.0 standard
//

// Support both extant linux fortran name mangling schemes
#ifdef __linux
#pragma weak mpi_file_close__ = pmpi_file_close_
#pragma weak mpi_file_delete__ = pmpi_file_delete_
#pragma weak mpi_file_get_amode__ = pmpi_file_get_amode_
#pragma weak mpi_file_get_atomicity__ = pmpi_file_get_atomicity_
#pragma weak mpi_file_get_byte_offset__ = pmpi_file_get_byte_offset_
#pragma weak mpi_file_get_errhandler__ = pmpi_file_get_errhandler_
#pragma weak mpi_file_get_group__ = pmpi_file_get_group_
#pragma weak mpi_file_get_info__ = pmpi_file_get_info_
#pragma weak mpi_file_get_position__ = pmpi_file_get_position_
#pragma weak mpi_file_get_position_shared__ = pmpi_file_get_position_shared_
#pragma weak mpi_file_get_size__ = pmpi_file_get_size_
#pragma weak mpi_file_get_type_extent__ = pmpi_file_get_type_extent_
#pragma weak mpi_file_get_view__ = pmpi_file_get_view_
#pragma weak mpi_info_create__ = pmpi_info_create_
#pragma weak mpi_info_delete__ = pmpi_info_delete_
#pragma weak mpi_info_dup__ = pmpi_info_dup_
#pragma weak mpi_info_free__ = pmpi_info_free_
#pragma weak mpi_info_get__ = pmpi_info_get_
#pragma weak mpi_info_get_nkeys__ = pmpi_info_get_nkeys_
#pragma weak mpi_info_get_nthkey__ = pmpi_info_get_nthkey_
#pragma weak mpi_info_get_valuelen__ = pmpi_info_get_valuelen_
#pragma weak mpi_info_set__ = pmpi_info_set_
#pragma weak mpi_file_iread__ = pmpi_file_iread_
#pragma weak mpi_file_iread_at__ = pmpi_file_iread_at_
#pragma weak mpi_file_iread_shared__ = pmpi_file_iread_shared_
#pragma weak mpi_file_iwrite__ = pmpi_file_iwrite_
#pragma weak mpi_file_iwrite_at__ = pmpi_file_iwrite_at_
#pragma weak mpi_file_iwrite_shared__ = pmpi_file_iwrite_shared_
#pragma weak mpi_file_open__ = pmpi_file_open_
#pragma weak mpi_file_preallocate__ = pmpi_file_preallocate_
#pragma weak mpi_file_read__ = pmpi_file_read_
#pragma weak mpi_file_read_all__ = pmpi_file_read_all_
#pragma weak mpi_file_read_all_begin__ = pmpi_file_read_all_begin_
#pragma weak mpi_file_read_all_end__ = pmpi_file_read_all_end_
#pragma weak mpi_file_read_at__ = pmpi_file_read_at_
#pragma weak mpi_file_read_at_all__ = pmpi_file_read_at_all_
#pragma weak mpi_file_read_at_all_begin__ = pmpi_file_read_at_all_begin_
#pragma weak mpi_file_read_at_all_end__ = pmpi_file_read_at_all_end_
#pragma weak mpi_file_read_ordered__ = pmpi_file_read_ordered_
#pragma weak mpi_file_read_ordered_begin__ = pmpi_file_read_ordered_begin_
#pragma weak mpi_file_read_ordered_end__ = pmpi_file_read_ordered_end_
#pragma weak mpi_file_read_shared__ = pmpi_file_read_shared_
#pragma weak mpi_file_seek__ = pmpi_file_seek_
#pragma weak mpi_file_seek_shared__ = pmpi_file_seek_shared_
#pragma weak mpi_file_set_atomicity__ = pmpi_file_set_atomicity_
#pragma weak mpi_file_set_errhandler__ = pmpi_file_set_errhandler_
#pragma weak mpi_file_set_info__ = pmpi_file_set_info_
#pragma weak mpi_file_set_size__ = pmpi_file_set_size_
#pragma weak mpi_file_set_view__ = pmpi_file_set_view_
#pragma weak mpi_file_sync__ = pmpi_file_sync_
#pragma weak mpi_file_write__ = pmpi_file_write_
#pragma weak mpi_file_write_all__ = pmpi_file_write_all_
#pragma weak mpi_file_write_all_begin__ = pmpi_file_write_all_begin_
#pragma weak mpi_file_write_all_end__ = pmpi_file_write_all_end_
#pragma weak mpi_file_write_at__ = pmpi_file_write_at_
#pragma weak mpi_file_write_at_all__ = pmpi_file_write_at_all_
#pragma weak mpi_file_write_at_all_begin__ = pmpi_file_write_at_all_begin_
#pragma weak mpi_file_write_at_all_end__ = pmpi_file_write_at_all_end_
#pragma weak mpi_file_write_ordered__ = pmpi_file_write_ordered_
#pragma weak mpi_file_write_ordered_begin__ = pmpi_file_write_ordered_begin_
#pragma weak mpi_file_write_ordered_end__ = pmpi_file_write_ordered_end_
#pragma weak mpi_file_write_shared__ = pmpi_file_write_shared_
#pragma weak mpi_get__ = pmpi_get_
#pragma weak mpi_get_version__ = pmpi_get_version_
#pragma weak mpi_init_thread__ = pmpi_init_thread_
#pragma weak mpi_is_thread_main__ = pmpi_is_thread_main_
#pragma weak mpi_put__ = pmpi_put_
#pragma weak mpi_query_thread__ = pmpi_query_thread_
#pragma weak mpi_type_create_darray__ = pmpi_type_create_darray_
#pragma weak mpi_type_create_subarray__ = pmpi_type_create_subarray_
#pragma weak mpi_win_create__ = pmpi_win_create_
#pragma weak mpi_win_fence__ = pmpi_win_fence_
#pragma weak mpi_win_free__ = pmpi_win_free_
#pragma weak mpio_test__ = pmpio_test_
#pragma weak mpio_wait__ = pmpio_wait_
#pragma weak mpi_type_create_hindexed__ = pmpi_type_create_hindexed_
#pragma weak mpi_type_create_hvector__ = pmpi_type_create_hvector_
#pragma weak mpi_type_create_struct__ = pmpi_type_create_struct_
#pragma weak mpi_comm_spawn__ = pmpi_comm_spawn_
#pragma weak mpi_comm_spawn_multiple__ = pmpi_comm_spawn_multiple_
#pragma weak mpi_comm_get_parent__ = pmpi_comm_get_parent_
#endif

//   
// _________________________________________________________________
// 
// MPI_GET_VERSION( version, subversion )
// [ OUT version] version number (integer)
// [ OUT subversion] subversion number (integer)
// 
// int MPI_Get_version(int *version, int *subversion)
// 
// MPI_GET_VERSION(VERSION, SUBVERSION, IERROR)
// INTEGER VERSION, SUBVERSION, IERROR
// 
void
mpi_get_version_(
fortran_int  *version,
fortran_int  *subversion,
fortran_int  *ierr)
{
int x_version ;
int x_subversion ;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Get_version...\n");
#endif
        *ierr = (fortran_int) PMPI_Get_version(&x_version, &x_subversion);

        *version = (fortran_int) x_version ;
        *subversion = (fortran_int) x_subversion ;
}

// _________________________________________________________________
// 
// MPI_TYPE_CREATE_INDEXED_BLOCK(count, blocklength,
// array_of_displacements, oldtype, newtype)
// [ IN count] length of array of displacements (integer)
// [ IN blocklength] size of block (integer)
// [ IN array_of_displacements] array of displacements (array of integer)
// [ IN oldtype] old datatype (handle)
// [ OUT newtype] new datatype (handle)
// 
// int MPI_Type_create_indexed_block(int count, int blocklength, int
// array_of_displacements[], MPI_Datatype oldtype, MPI_Datatype *newtype)
// 
// MPI_TYPE_CREATE_INDEXED_BLOCK(COUNT, BLOCKLENGTH,
// ARRAY_OF_DISPLACEMENTS, OLDTYPE, NEWTYPE, IERROR)
// INTEGER COUNT, BLOCKLENGTH, ARRAY_OF_DISPLACEMENTS(*), OLDTYPE,
// NEWTYPE, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

// _________________________________________________________________
// 
// MPI_REQUEST_GET_STATUS( request, flag, status )
// [ IN request] request (handle)
// [ OUT flag] boolean flag, same as from MPI_TEST (logical)
// [ OUT status] MPI_STATUS object if flag is true (Status)
// 
// int MPI_Request_get_status(MPI_Request request, int *flag, MPI_Status
// *status)
// 
// MPI_REQUEST_GET_STATUS( REQUEST, FLAG, STATUS, IERROR)
// INTEGER REQUEST, STATUS(MPI_STATUS_SIZE), IERROR
// LOGICAL FLAG
// 
// MPI-2 routine not yet implemented in SGI MPT

// _________________________________________________________________
// 
// MPI_FINALIZED(flag)
// [ OUT flag] true if MPI was finalized (logical)
// 
// int MPI_Finalized(int *flag)
// 
// MPI_FINALIZED(FLAG, IERROR)
// LOGICAL FLAG
// INTEGER IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

// _________________________________________________________________
// 
// MPI_INFO_CREATE(info)
// [ OUT info] info object created (handle)
// 
// int MPI_Info_create(MPI_Info *info)
// 
// MPI_INFO_CREATE(INFO, IERROR)
// INTEGER INFO, IERROR
// 
void
mpi_info_create_(
fortran_int  *info,
fortran_int  *ierr)
{
MPI_Info x_info;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Info_create\n");
#endif
        *ierr = (fortran_int) PMPI_Info_create(&x_info);
        *info = (fortran_int) x_info ;
}

// 
// MPI_INFO_SET(info, key, value)
// [ INOUT info] info object (handle)
// [ IN key] key (string)
// [ IN value] value (string)
//
// int MPI_Info_set(MPI_Info info, char *key, char *value)
//
// MPI_INFO_SET(INFO, KEY, VALUE, IERROR)
// INTEGER INFO, IERROR
// CHARACTER*(*) KEY, VALUE
//
void
mpi_info_set_(
fortran_int *info,
char *key,
char *value,
fortran_int *ier, 
fortran_int keylen,
fortran_int vallen)
{
MPI_Info x_info = *info;
char x_key[keylen+1];
char x_value[vallen+1];
	memcpy(x_key, key, keylen);
	x_key[keylen] = '\0';
	memcpy(x_value, value, vallen);
	x_value[vallen] = '\0';

	*ier = (fortran_int) PMPI_Info_set( x_info, x_key, x_value);

        *info = (fortran_int) x_info ;
}

// _________________________________________________________________
// 
// MPI_INFO_DELETE(info, key)
// [ INOUT info] info object (handle)
// [ IN key] key (string)
//
// int MPI_Info_delete(MPI_Info info, char *key)
//
// MPI_INFO_DELETE(INFO, KEY, IERROR)
// INTEGER INFO, IERROR
// CHARACTER*(*) KEY
// 
void
mpi_info_delete_(
fortran_int *info,
char *key,
fortran_int *ier,
fortran_int keylen)
{
MPI_Info x_info = *info;
char x_key[keylen+1];
	memcpy(x_key, key, keylen);
	x_key[keylen] = '\0';

	*ier = (fortran_int) PMPI_Info_delete( x_info, x_key);

        *info = (fortran_int) x_info ;
}

// _________________________________________________________________
//
// MPI_INFO_GET(info, key, valuelen, value, flag)
// [ IN info] info object (handle)
// [ IN key] key (string)
// [ IN valuelen] length of value arg (integer)
// [ OUT value] value (string)
// [ OUT flag] true if key defined, false if not (boolean)
//
// int MPI_Info_get(MPI_Info info, char *key, int valuelen, char *value,
// int *flag)
//
// MPI_INFO_GET(INFO, KEY, VALUELEN, VALUE, FLAG, IERROR)
// INTEGER INFO, VALUELEN, IERROR
// CHARACTER*(*) KEY, VALUE
// LOGICAL FLAG
//
void
mpi_info_get_(
fortran_int *info,
char *key,
fortran_int *valuelen,
char *value, 
fortran_int *flag,
fortran_int *ier,
fortran_int keylen,
fortran_int valspace)
{
int x_flag;
char x_value[*valuelen+1];
char x_key[keylen+1];
	memcpy(x_key, key, keylen);
	x_key[keylen] = '\0';


	*ier = (fortran_int) PMPI_Info_get( (MPI_Info) *info, x_key,  (int) *valuelen, x_value, &x_flag);

	memcpy(value, x_value, *valuelen);
        *flag = (fortran_int) x_flag ;
}


// _________________________________________________________________
// 
// MPI_INFO_GET_VALUELEN(info, key, valuelen, flag)
// [ IN info] info object (handle)
// [ IN key] key (string)
// [ OUT valuelen] length of value arg (integer)
// [ OUT flag] true if key defined, false if not (boolean)
//
// int MPI_Info_get_valuelen(MPI_Info info, char *key, int *valuelen, int
// *flag)
//
// MPI_INFO_GET_VALUELEN(INFO, KEY, VALUELEN, FLAG, IERROR)
// INTEGER INFO, VALUELEN, IERROR
// LOGICAL FLAG
// CHARACTER*(*) KEY
//
// POSSIBLE ERRORS IN mpi_info_get_valuelen_
void
mpi_info_get_valuelen_(
fortran_int *info,
char *key,
fortran_int *valuelen,
fortran_int *flag,
fortran_int *ier,
fortran_int keylen )
{
int x_valuelen;
int x_flag;
char x_key[keylen+1];
	memcpy(x_key, key, keylen);
	x_key[keylen] = '\0';

	*ier = (fortran_int) PMPI_Info_get_valuelen( (MPI_Info) *info, x_key, &x_valuelen, &x_flag);

	*flag = (fortran_int) x_flag;
	*valuelen = (fortran_int) x_valuelen;
}

// _________________________________________________________________
//
// MPI_INFO_GET_NKEYS(info, nkeys)
// [ IN info] info object (handle)
// [ OUT nkeys] number of defined keys (integer)
//
// int MPI_Info_get_nkeys(MPI_Info info, int *nkeys)
//
// MPI_INFO_GET_NKEYS(INFO, NKEYS, IERROR)
// INTEGER INFO, NKEYS, IERROR
//
void
mpi_info_get_nkeys_(
fortran_int *info,
fortran_int *nkeys,
fortran_int *ier)
{
int x_nkeys;
        *ier = (fortran_int) PMPI_Info_get_nkeys((MPI_Info) *info, &x_nkeys);
        *nkeys = (fortran_int) x_nkeys ;
}

// _________________________________________________________________
//
// MPI_INFO_GET_NTHKEY(info, n, key)
// [ IN info] info object (handle)
// [ IN n] key number (integer)
// [ OUT key] key (string)
//
// int MPI_Info_get_nthkey(MPI_Info info, int n, char *key)
//
// MPI_INFO_GET_NTHKEY(INFO, N, KEY, IERROR)
// INTEGER INFO, N, IERROR
// CHARACTER*(*) KEY
//
void
mpi_info_get_nthkey_(
fortran_int *info,
fortran_int *n,
char *key,
fortran_int *ier,
fortran_int keylen)
{
char x_key[keylen+1];
	memcpy(x_key, key, keylen);
	x_key[keylen] = '\0';

	*ier = (fortran_int) PMPI_Info_get_nthkey( (int) *info,  (int) *n, x_key);
	memcpy(key, x_key, keylen);
}


// _________________________________________________________________
// 
// MPI_INFO_DUP(info, newinfo)
// [ IN info] info object (handle)
// [ OUT newinfo] info object (handle)
//
// int MPI_Info_dup(MPI_Info info, MPI_Info *newinfo)
//
// MPI_INFO_DUP(INFO, NEWINFO, IERROR)
// INTEGER INFO, NEWINFO, IERROR
//
void
mpi_info_dup_(
fortran_int  *info,
fortran_int  *newinfo,
fortran_int  *ierr)
{
MPI_Info x_newinfo;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Info_dup\n");
#endif
        *ierr = (fortran_int) PMPI_Info_dup((MPI_Info) *info, &x_newinfo);
        *newinfo = (fortran_int) x_newinfo ;
}

// _________________________________________________________________
// 
// MPI_INFO_FREE(info)
// [ INOUT info] info object (handle)
//
// int MPI_Info_free(MPI_Info *info)
//
// MPI_INFO_FREE(INFO, IERROR)
// INTEGER INFO, IERROR
//
void
mpi_info_free_(
fortran_int  *info,
fortran_int  *ierr)
{
MPI_Info x_info = *info;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Info_free\n");
#endif
        *ierr = (fortran_int) PMPI_Info_free(&x_info);
        *info = (fortran_int) x_info ;
}

// _________________________________________________________________
// 
// MPI_ALLOC_MEM(size, info, baseptr)
// [ IN size] size of memory segment in bytes (nonnegative integer)
// [ IN info] info argument (handle)
// [ OUT baseptr] pointer to beginning of memory segment allocated
// 
// int MPI_Alloc_mem(MPI_Aint size, MPI_Info info, void *baseptr)
// 
// MPI_ALLOC_MEM(SIZE, INFO, BASEPTR, IERROR)
// INTEGER INFO, IERROR
// INTEGER(KIND=MPI_ADDRESS_KIND) SIZE, BASEPTR
//
// MPI-2 routine not yet implemented in SGI MPT

// _________________________________________________________________
// 
// MPI_FREE_MEM(base)
// [ IN base] initial address of memory segment allocated by
// MPI_ALLOC_MEM (choice)
// 
// int MPI_Free_mem(void *base)
// 
// MPI_FREE_MEM(BASE, IERROR)
// <type> BASE(*)
// INTEGER IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

// _________________________________________________________________
//
// MPI_COMM_CREATE_ERRHANDLER(function, errhandler)
// [ IN function] user defined error handling procedure (function)
// [ OUT errhandler] MPI error handler (handle)
// 
// int MPI_Comm_create_errhandler(MPI_Comm_errhandler_fn *function,
// MPI_Errhandler *errhandler)
// 
// MPI_COMM_CREATE_ERRHANDLER(FUNCTION, ERRHANDLER, IERROR)
// EXTERNAL FUNCTION
// INTEGER ERRHANDLER, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

// _________________________________________________________________
// 
// MPI_COMM_SET_ERRHANDLER(comm, errhandler)
// [ INOUT comm] communicator (handle)
// [ IN errhandler] new error handler for communicator (handle)
// 
// int MPI_Comm_set_errhandler(MPI_Comm comm, MPI_Errhandler errhandler)
// 
// MPI_COMM_SET_ERRHANDLER(COMM, ERRHANDLER, IERROR)
// INTEGER COMM, ERRHANDLER, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

// _________________________________________________________________
// 
// MPI_COMM_GET_ERRHANDLER(comm, errhandler)
// [ IN comm] communicator (handle)
// [ OUT errhandler] error handler currently associated with communicator
// (handle)
// 
// int MPI_Comm_get_errhandler(MPI_Comm comm, MPI_Errhandler *errhandler)
// 
// MPI_COMM_GET_ERRHANDLER(COMM, ERRHANDLER, IERROR)
// INTEGER COMM, ERRHANDLER, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

// _________________________________________________________________
// 
// MPI_WIN_CREATE_ERRHANDLER(function, errhandler)
// [ IN function] user defined error handling procedure (function)
// [ OUT errhandler] MPI error handler (handle)
// 
// int MPI_Win_create_errhandler(MPI_Win_errhandler_fn *function,
// MPI_Errhandler *errhandler)
// 
// MPI_WIN_CREATE_ERRHANDLER(FUNCTION, ERRHANDLER, IERROR)
// EXTERNAL FUNCTION
// INTEGER ERRHANDLER, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

// _________________________________________________________________
// 
// MPI_WIN_SET_ERRHANDLER(win, errhandler)
// [ INOUT win] window (handle)
// [ IN errhandler] new error handler for window (handle)
// 
// int MPI_Win_set_errhandler(MPI_Win win, MPI_Errhandler errhandler)
// 
// MPI_WIN_SET_ERRHANDLER(WIN, ERRHANDLER, IERROR)
// INTEGER WIN, ERRHANDLER, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

// _________________________________________________________________
// 
// MPI_WIN_GET_ERRHANDLER(win, errhandler)
// [ IN win] window (handle)
// [ OUT errhandler] error handler currently associated with window
// (handle)
// 
// int MPI_Win_get_errhandler(MPI_Win win, MPI_Errhandler *errhandler)
// 
// MPI_WIN_GET_ERRHANDLER(WIN, ERRHANDLER, IERROR)
// INTEGER WIN, ERRHANDLER, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

// _________________________________________________________________
// 
// MPI_FILE_CREATE_ERRHANDLER(function, errhandler)
// [ IN function] user defined error handling procedure (function)
// [ OUT errhandler] MPI error handler (handle)
// 
// int MPI_File_create_errhandler(MPI_File_errhandler_fn *function,
// MPI_Errhandler *errhandler)
// 
// MPI_FILE_CREATE_ERRHANDLER(FUNCTION, ERRHANDLER, IERROR)
// EXTERNAL FUNCTION
// INTEGER ERRHANDLER, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

// _________________________________________________________________
// 
// MPI_FILE_SET_ERRHANDLER(file, errhandler)
// [ INOUT file] file (handle)
// [ IN errhandler] new error handler for file (handle)
// 
// int MPI_File_set_errhandler(MPI_File file, MPI_Errhandler errhandler)
// 
// MPI_FILE_SET_ERRHANDLER(FILE, ERRHANDLER, IERROR)
// INTEGER FILE, ERRHANDLER, IERROR
// 
void
mpi_file_set_errhandler_(
fortran_int *fh,
fortran_int *err_handler,
fortran_int *ierr)
{
    MPI_File fh_c;
    MPI_Errhandler err_handler_c;
 
    fh_c = PMPI_File_f2c( (int) *fh);
    err_handler_c = (MPI_Errhandler)(*err_handler);

    *ierr = (fortran_int) PMPI_File_set_errhandler(fh_c,err_handler_c);
}

// _________________________________________________________________
// 
// MPI_FILE_GET_ERRHANDLER(file, errhandler)
// [ IN file] file (handle)
// [ OUT errhandler] error handler currently associated with file
// (handle)
// 
// int MPI_File_get_errhandler(MPI_File file, MPI_Errhandler *errhandler)
// 
// MPI_FILE_GET_ERRHANDLER(FILE, ERRHANDLER, IERROR)
// INTEGER FILE, ERRHANDLER, IERROR
// 
void
mpi_file_get_errhandler_(
fortran_int *fh,
fortran_int *err_handler,
fortran_int *ierr)
{
    MPI_File fh_c;
    MPI_Errhandler err_handler_c;

    fh_c = PMPI_File_f2c(*fh);
    *ierr = PMPI_File_get_errhandler(fh_c, &err_handler_c);
    /* KLUDGE *err_handler = MPI_Errhandler_c2f(err_handler_c); */
    *err_handler = (fortran_int) (err_handler_c);
}

// _________________________________________________________________
// 
// MPI_TYPE_CREATE_HVECTOR( count, blocklength, stride, oldtype, newtype)
// [ IN count] number of blocks (nonnegative integer)
// [ IN blocklength] number of elements in each block (nonnegative
// integer)
// [ IN stride] number of bytes between start of each block (integer)
// [ IN oldtype] old datatype (handle)
// [ OUT newtype] new datatype (handle)
// 
// int MPI_Type_create_hvector(int count, int blocklength, MPI_Aint
// stride, MPI_Datatype oldtype, MPI_Datatype *newtype)
// 
// MPI_TYPE_CREATE_HVECTOR(COUNT, BLOCKLENGTH, STIDE, OLDTYPE, NEWTYPE,
// IERROR)
// INTEGER COUNT, BLOCKLENGTH, OLDTYPE, NEWTYPE, IERROR
// INTEGER(KIND=MPI_ADDRESS_KIND) STRIDE
// 
void
mpi_type_create_hvector_(
fortran_int  *count,
fortran_int  *blocklength,
fortran_int  *stride,
fortran_int  *oldtype,
fortran_int  *newtype,
fortran_int  *ierr)
{
MPI_Datatype x_newtype;
MPI_Datatype x_oldtype = convert_type(*oldtype);

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Type_create_hvector with *oldtype = %i, x_oldtype = %i\n",(int) *oldtype, x_oldtype);
#endif
        *ierr = (fortran_int) PMPI_Type_create_hvector( (int) *count, (int) *blocklength, (int) *stride, (MPI_Datatype) x_oldtype, &x_newtype);

        *newtype = (fortran_int) x_newtype ;
}

// 
// MPI_TYPE_CREATE_HINDEXED( count, array_of_blocklengths,
// array_of_displacements, oldtype, newtype)
// [ IN count] number of blocks --- also number of entries in
// array_of_displacements and array_of_blocklengths (integer)
// [ IN array_of_blocklengths] number of elements in each block (array of
// nonnegative integers)
// [ IN array_of_displacements] byte displacement of each block (array of
// integer)
// [ IN oldtype] old datatype (handle)
// [ OUT newtype] new datatype (handle)
// 
// int MPI_Type_create_hindexed(int count, int array_of_blocklengths[],
// MPI_Aint array_of_displacements[], MPI_Datatype oldtype, MPI_Datatype
// *newtype)
// 
// MPI_TYPE_CREATE_HINDEXED(COUNT, ARRAY_OF_BLOCKLENGTHS,
// ARRAY_OF_DISPLACEMENTS, OLDTYPE, NEWTYPE, IERROR)
// INTEGER COUNT, ARRAY_OF_BLOCKLENGTHS(*), OLDTYPE, NEWTYPE, IERROR
// INTEGER(KIND=MPI_ADDRESS_KIND) ARRAY_OF_DISPLACEMENTS(*)
// 
//
void
mpi_type_create_hindexed_(
fortran_int  *count,
fortran_int  array_of_blocklengths[],
fortran_int  array_of_displacements[],
fortran_int  *oldtype,
fortran_int  *newtype,
fortran_int  *ierr)
{
MPI_Aint x_array_of_displacements[ (int) *count];
int x_array_of_blocklengths[ (int) *count];
MPI_Datatype x_newtype;
MPI_Datatype x_oldtype = convert_type(*oldtype);
int i;

	for (i=0;i< (int) *count;i++) {
	  x_array_of_blocklengths[i] = array_of_blocklengths[i];
	}
	for (i=0;i< (int) *count;i++) {
	  x_array_of_displacements[i] = array_of_displacements[i];
	}

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Type_create_hindexed...\n");
#endif
        *ierr = (fortran_int) PMPI_Type_create_hindexed( (int) *count, x_array_of_blocklengths, x_array_of_displacements, (MPI_Datatype) x_oldtype, &x_newtype);

        *newtype = (fortran_int) x_newtype ;
	free(x_array_of_displacements);
	free(x_array_of_blocklengths);
}

// 
// MPI_TYPE_CREATE_STRUCT(count, array_of_blocklengths,
// array_of_displacements, array_of_types, newtype)
// [ IN count] number of blocks (integer) --- also number of entries in
// arrays array_of_types, array_of_displacements and
// array_of_blocklengths
// [ IN array_of_blocklength] number of elements in each block (array of
// integer)
// [ IN array_of_displacements] byte displacement of each block (array of
// integer)
// [ IN array_of_types] type of elements in each block (array of handles
// to datatype objects)
// [ OUT newtype] new datatype (handle)
// 
// int MPI_Type_create_struct(int count, int array_of_blocklengths[],
// MPI_Aint array_of_displacements[], MPI_Datatype array_of_types[],
// MPI_Datatype *newtype)
// 
// MPI_TYPE_CREATE_STRUCT(COUNT, ARRAY_OF_BLOCKLENGTHS,
// ARRAY_OF_DISPLACEMENTS, ARRAY_OF_TYPES, NEWTYPE, IERROR)
// INTEGER COUNT, ARRAY_OF_BLOCKLENGTHS(*), ARRAY_OF_TYPES(*), NEWTYPE,
// IERROR
// INTEGER(KIND=MPI_ADDRESS_KIND) ARRAY_OF_DISPLACEMENTS(*)
// 
void
mpi_type_create_struct_(
fortran_int  *count,
fortran_int  array_of_blocklengths[],
fortran_int  array_of_displacements[],
fortran_int  *array_of_types[],
fortran_int  *newtype,
fortran_int  *ierr)
{
MPI_Aint x_array_of_displacements[ (int) *count];
int x_array_of_blocklengths[ (int) *count];
MPI_Datatype x_newtype;
MPI_Datatype x_array_of_types[ (int) *count ];
int i;

	for (i=0;i< (int) *count;i++) {
	  x_array_of_blocklengths[i] = array_of_blocklengths[i];
	}
	for (i=0;i< (int) *count;i++) {
	  x_array_of_displacements[i] = array_of_displacements[i];
	}
	for (i=0;i< (int) *count;i++) {
	  x_array_of_types[i] = convert_type( x_array_of_types[i] );
	}

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Type_create_struct...\n");
#endif
        *ierr = (fortran_int) PMPI_Type_create_struct( (int) *count, x_array_of_blocklengths, x_array_of_displacements, x_array_of_types, &x_newtype);

        *newtype = (fortran_int) x_newtype ;
	free(x_array_of_displacements);
	free(x_array_of_blocklengths);
}

// 
// MPI_GET_ADDRESS(location, address)
// [ IN location] location in caller memory (choice)
// [ OUT address] address of location (integer)
// 
// int MPI_Get_address(void *location, MPI_Aint *address)
// 
// MPI_GET_ADDRESS(LOCATION, ADDRESS, IERROR)
// <type> LOCATION(*)
// INTEGER IERROR
// INTEGER(KIND=MPI_ADDRESS_KIND) ADDRESS
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_TYPE_GET_EXTENT(datatype, lb, extent)
// [ IN datatype] datatype to get information on (handle)
// [ OUT lb] lower bound of datatype (integer)
// [ OUT extent] extent of datatype (integer)
// 
// int MPI_Type_get_extent(MPI_Datatype datatype, MPI_Aint *lb, MPI_Aint
// *extent)
// 
// MPI_TYPE_GET_EXTENT(DATATYPE, LB, EXTENT, IERROR)
// INTEGER DATATYPE, IERROR
// INTEGER(KIND = MPI_ADDRESS_KIND) LB, EXTENT
// 
// MPI-2 routine not yet implemented in SGI MPT
// 

// 
// MPI_TYPE_CREATE_RESIZED(oldtype, lb, extent, newtype)
// [ IN oldtype] input datatype (handle)
// [ IN lb] new lower bound of datatype (integer)
// [ IN extent] new extent of datatype (integer)
// [ OUT newtype] output datatype (handle)
// 
// int MPI_Type_create_resized(MPI_Datatype oldtype, MPI_Aint lb,
// MPI_Aint extent, MPI_Datatype *newtype)
// 
// MPI_TYPE_CREATE_RESIZED(OLDTYPE, LB, EXTENT, NEWTYPE, IERROR)
// INTEGER OLDTYPE, NEWTYPE, IERROR
// INTEGER(KIND=MPI_ADDRESS_KIND) LB, EXTENT
// 
// MPI-2 routine not yet implemented in SGI MPT
// 

// 
// MPI_TYPE_GET_TRUE_EXTENT(datatype, true_lb, true_extent)
// [ IN datatype] datatype to get information on (handle)
// [ OUT true_lb] true lower bound of datatype (integer)
// [ OUT true_extent] true size of datatype (integer)
// 
// int MPI_Type_get_true_extent(MPI_Datatype datatype, MPI_Aint *true_lb,
// MPI_Aint *true_extent)
// 
// MPI_TYPE_GET_TRUE_EXTENT(DATATYPE, TRUE_LB, TRUE_EXTENT, IERROR)
// INTEGER DATATYPE, IERROR
// INTEGER(KIND = MPI_ADDRESS_KIND) TRUE_LB, TRUE_EXTENT
// 
// MPI-2 routine not yet implemented in SGI MPT
// 

// 
// MPI_TYPE_CREATE_SUBARRAY(ndims, array_of_sizes, array_of_subsizes,
// array_of_starts, order, oldtype, newtype)
// [ IN ndims] number of array dimensions (positive integer)
// [ IN array_of_sizes] number of elements of type oldtype in each
// dimension of the full array (array of positive integers)
// [ IN array_of_subsizes] number of elements of type oldtype in each
// dimension of the subarray (array of positive integers)
// [ IN array_of_starts] starting coordinates of the subarray in each
// dimension (array of nonnegative integers)
// [ IN order] array storage order flag (state)
// [ IN oldtype] array element datatype (handle)
// [ OUT newtype] new datatype (handle)
// 
// int MPI_Type_create_subarray(int ndims, int array_of_sizes[], int
// array_of_subsizes[], int array_of_starts[], int order, MPI_Datatype
// oldtype, MPI_Datatype *newtype)
// 
// MPI_TYPE_CREATE_SUBARRAY(NDIMS, ARRAY_OF_SIZES, ARRAY_OF_SUBSIZES,
// ARRAY_OF_STARTS, ORDER, OLDTYPE, NEWTYPE, IERROR)
// INTEGER NDIMS, ARRAY_OF_SIZES(*), ARRAY_OF_SUBSIZES(*),
// ARRAY_OF_STARTS(*), ORDER, OLDTYPE, NEWTYPE, IERROR
//
void
mpi_type_create_subarray_(
fortran_int *ndims,
fortran_int array_of_sizes[],
fortran_int array_of_subsizes[],
fortran_int array_of_starts[],
fortran_int *order,
fortran_int *oldtype,
fortran_int *newtype,
fortran_int *ierr )
{
int x_array_of_sizes[ (int) *ndims];
int x_array_of_subsizes[ (int) *ndims];
int x_array_of_starts[ (int) *ndims];
int i;

MPI_Datatype x_newtype;
MPI_Datatype x_oldtype = convert_type(*oldtype);

	for (i=0;i< (int) *ndims;i++) {
	  x_array_of_sizes[i] = array_of_sizes[i];
	}
	for (i=0;i< (int) *ndims;i++) {
	  x_array_of_subsizes[i] = array_of_subsizes[i];
	}
	for (i=0;i< (int) *ndims;i++) {
	  x_array_of_starts[i] = array_of_starts[i];
	}

	*ierr = (fortran_int) PMPI_Type_create_subarray( (int) *ndims, x_array_of_sizes, x_array_of_subsizes, x_array_of_starts, (int) *order, (MPI_Datatype) x_oldtype, &x_newtype);

        *newtype = (fortran_int) x_newtype ;

}

//
// MPI_TYPE_CREATE_DARRAY(size, rank, ndims, array_of_gsizes,
// array_of_distribs, array_of_dargs, array_of_psizes, order, oldtype,
// newtype)
// [ IN size] size of process group (positive integer)
// [ IN rank] rank in process group (nonnegative integer)
// [ IN ndims] number of array dimensions as well as process grid
// dimensions (positive integer)
// [ IN array_of_gsizes] number of elements of type oldtype in each
// dimension of global array (array of positive integers)
// [ IN array_of_distribs] distribution of array in each dimension (array
// of state)
// [ IN array_of_dargs] distribution argument in each dimension (array of
// positive integers)
// [ IN array_of_psizes] size of process grid in each dimension (array of
// positive integers)
// [ IN order] array storage order flag (state)
// [ IN oldtype] old datatype (handle)
// [ OUT newtype] new datatype (handle)
// 
// int MPI_Type_create_darray(int size, int rank, int ndims, int
// array_of_gsizes[], int array_of_distribs[], int array_of_dargs[], int
// array_of_psizes[], int order, MPI_Datatype oldtype, MPI_Datatype
// *newtype)
//
// MPI_TYPE_CREATE_DARRAY(SIZE, RANK, NDIMS, ARRAY_OF_GSIZES,
// ARRAY_OF_DISTRIBS, ARRAY_OF_DARGS, ARRAY_OF_PSIZES, ORDER, OLDTYPE,
// NEWTYPE, IERROR)
// INTEGER SIZE, RANK, NDIMS, ARRAY_OF_GSIZES(*), ARRAY_OF_DISTRIBS(*),
// ARRAY_OF_DARGS(*), ARRAY_OF_PSIZES(*), ORDER, OLDTYPE, NEWTYPE, IERROR
//
void
mpi_type_create_darray_(
fortran_int *size,
fortran_int *rank,
fortran_int *ndims,
fortran_int array_of_gsizes[],
fortran_int array_of_distribs[],
fortran_int array_of_dargs[],
fortran_int array_of_psizes[],
fortran_int *order,
fortran_int *oldtype,
fortran_int *newtype,
fortran_int *ierr )
{
MPI_Datatype x_newtype;
MPI_Datatype x_oldtype = convert_type(*oldtype);
int x_array_of_gsizes[ (int) *ndims];
int x_array_of_distribs[ (int) *ndims];
int x_array_of_dargs[ (int) *ndims];
int x_array_of_psizes[ (int) *size];
int i;
// Are these loop indices correct?
	for (i=0;i< (int) *ndims;i++) {
	  x_array_of_gsizes[i] = array_of_gsizes[i];
	}
	for (i=0;i< (int) *ndims;i++) {
	  x_array_of_distribs[i] = array_of_distribs[i];
	}
	for (i=0;i< (int) *ndims;i++) {
	  x_array_of_dargs[i] = array_of_dargs[i];
	}
	for (i=0;i< (int) *size;i++) {
	  x_array_of_psizes[i] = array_of_psizes[i];
	}

    *ierr = (fortran_int) PMPI_Type_create_darray( (int) *size, (int) *rank, (int) *ndims, x_array_of_gsizes, x_array_of_distribs, x_array_of_dargs, x_array_of_psizes, (int) *order, x_oldtype, &x_newtype);

        *newtype = (fortran_int) x_newtype ;
}

// 
// MPI_PACK_EXTERNAL(datarep, inbuf, incount, datatype, outbuf, outsize,
// position )
// [ IN datarep] data representation (string)
// [ IN inbuf] input buffer start (choice)
// [ IN incount] number of input data items (integer)
// [ IN datatype] datatype of each input data item (handle)
// [ OUT outbuf] output buffer start (choice)
// [ IN outsize] output buffer size, in bytes (integer)
// [ INOUT position] current position in buffer, in bytes (integer)
// 
// int MPI_Pack_external(char *datarep, void *inbuf, int incount,
// MPI_Datatype datatype, void *outbuf, MPI_Aint outsize, MPI_Aint
// *position)
// 
// MPI_PACK_EXTERNAL(DATAREP, INBUF, INCOUNT, DATATYPE, OUTBUF, OUTSIZE,
// POSITION, IERROR)
// INTEGER INCOUNT, DATATYPE, IERROR
// INTEGER(KIND=MPI_ADDRESS_KIND) OUTSIZE, POSITION
// CHARACTER*(*) DATAREP
// <type> INBUF(*), OUTBUF(*)
// 
// MPI-2 routine not yet implemented in SGI MPT
// 

// 
// MPI_UNPACK_EXTERNAL(datarep, inbuf, incount, datatype, outbuf,
// outsize, position )
// [ IN datarep] data representation (string)
// [ IN inbuf] input buffer start (choice)
// [ IN insize] input buffer size, in bytes (integer)
// [ INOUT position] current position in buffer, in bytes (integer)
// [ OUT outbuf] output buffer start (choice)
// [ IN outcount] number of output data items (integer)
// [ IN datatype] datatype of output data item (handle)
// 
// int MPI_Unpack_external(char *datarep, void *inbuf, MPI_Aint insize,
// MPI_Aint *position, void *outbuf, int outcount, MPI_Datatype datatype)
// 
// MPI_UNPACK_EXTERNAL(DATAREP, INBUF, INSIZE, POSITION, OUTBUF,
// OUTCOUNT, DATATYPE, IERROR)
// INTEGER OUTCOUNT, DATATYPE, IERROR
// INTEGER(KIND=MPI_ADDRESS_KIND) INSIZE, POSITION
// CHARACTER*(*) DATAREP
// <type> INBUF(*), OUTBUF(*)
// 
// MPI-2 routine not yet implemented in SGI MPT
// 

// 
// MPI_PACK_EXTERNAL_SIZE( datarep, incount, datatype, size )
// [ IN datarep] data representation (string)
// [ IN incount] number of input data items (integer)
// [ IN datatype] datatype of each input data item (handle)
// [ OUT size] output buffer size, in bytes (integer)
// 
// int MPI_Pack_external_size(char *datarep, int incount, MPI_Datatype
// datatype, MPI_Aint *size)
// 
// MPI_PACK_EXTERNAL_SIZE(DATAREP, INCOUNT, DATATYPE, SIZE, IERROR)
// INTEGER INCOUNT, DATATYPE, IERROR
// INTEGER(KIND=MPI_ADDRESS_KIND) SIZE
// CHARACTER*(*) DATAREP
// 
// MPI-2 routine not yet implemented in SGI MPT
//

//
// MPI_COMM_SPAWN(command, argv, maxprocs, info, root, comm, intercomm,
// array_of_errcodes)
// [ IN command] name of program to be spawned (string, significant only
// at root)
// [ IN argv] arguments to command (array of strings, significant only at
// root)
// [ IN maxprocs] maximum number of processes to start (integer,
// significant only at root)
// [ IN info] a set of key-value pairs telling the runtime system where
// and how to start the processes (handle, significant only at root)
// [ IN root] rank of process in which previous arguments are examined
// (integer)
// [ IN comm] intracommunicator containing group of spawning processes
// (handle)
// [ OUT intercomm] intercommunicator between original group and the
// newly spawned group (handle)
// [ OUT array_of_errcodes] one code per process (array of integer)
//
// int MPI_Comm_spawn(char *command, char *argv[], int maxprocs, MPI_Info
// info, int root, MPI_Comm comm, MPI_Comm *intercomm, int
// array_of_errcodes[])
//
// MPI_COMM_SPAWN(COMMAND, ARGV, MAXPROCS, INFO, ROOT, COMM, INTERCOMM,
// ARRAY_OF_ERRCODES, IERROR)
// CHARACTER*(*) COMMAND, ARGV(*)
// INTEGER INFO, MAXPROCS, ROOT, COMM, INTERCOMM, ARRAY_OF_ERRCODES(*),
// IERROR
// 
// INCOMPLETE.  Not finished implementing Fortran-to-C conversion
// on array of character data (parameter #2, argv[]).
#if ABI64
void
mpi_comm_spawn_(
char *command,
char *argv[],
fortran_int *maxprocs,
fortran_int *info,
fortran_int *root,
fortran_int *comm,
fortran_int *intercomm,
fortran_int array_of_errcodes[],
fortran_int  *ierr,
int command_len)
{
int i;
int x_array_of_errcodes[ (int) *maxprocs ] ;
MPI_Comm x_intercomm = *intercomm;
char x_command[command_len+1];
	memcpy(x_command, command, command_len);
	x_command[command_len] = '\0';

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Comm_spawn...\n");
#endif

	for (i=0;i<*maxprocs;i++) {
	  x_array_of_errcodes[i] = array_of_errcodes[i];
	}

        *ierr = (fortran_int) PMPI_Comm_spawn(x_command,
	argv, (int) *maxprocs, (MPI_Info) *info, (int) *root,
	(MPI_Comm) *comm, &x_intercomm, x_array_of_errcodes) ;

	for (i=0;i<*maxprocs;i++) {
	  array_of_errcodes[i] = (fortran_int) x_array_of_errcodes[i];
	}
        *intercomm = (fortran_int) x_intercomm ;
}
#endif

// 
// MPI_COMM_GET_PARENT(parent)
// [ OUT parent] the parent communicator (handle)
//
// int MPI_Comm_get_parent(MPI_Comm *parent)
//
// MPI_COMM_GET_PARENT(PARENT, IERROR)
// INTEGER PARENT, IERROR
// 
#if ABI64
void
mpi_comm_get_parent_(
fortran_int *parent,
fortran_int  *ierr)
{
MPI_Comm x_parent = *parent;
#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Comm_get_parent...\n");
#endif

        *ierr = (fortran_int) PMPI_Comm_get_parent( &x_parent ) ;

        *parent = (fortran_int) x_parent ;
}
#endif

// 
// MPI_COMM_SPAWN_MULTIPLE(count, array_of_commands, array_of_argv,
// array_of_maxprocs, array_of_info, root, comm, intercomm,
// array_of_errcodes)
// [ IN count] number of commands (positive integer, significant to MPI
// only at root --- see advice to users)
// [ IN array_of_commands] programs to be executed (array of strings,
// significant only at root)
// [ IN array_of_argv] arguments for commands (array of array of strings,
// significant only at root)
// [ IN array_of_maxprocs] maximum number of processes to start for each
// command (array of integer, significant only at root)
// [ IN array_of_info] info objects telling the runtime system where and
// how to start processes (array of handles, significant only at root)
// [ IN root] rank of process in which previous arguments are examined
// (integer)
// [ IN comm] intracommunicator containing group of spawning processes
// (handle)
// [ OUT intercomm] intercommunicator between original group and newly
// spawned group (handle)
// [ OUT array_of_errcodes] one error code per process (array of integer)
//
// int MPI_Comm_spawn_multiple(int count, char *array_of_commands[], char
// **array_of_argv[], int array_of_maxprocs[], MPI_Info array_of_info[],
// int root, MPI_Comm comm, MPI_Comm *intercomm, int array_of_errcodes[])
//
// MPI_COMM_SPAWN_MULTIPLE(COUNT, ARRAY_OF_COMMANDS, ARRAY_OF_ARGV,
// ARRAY_OF_MAXPROCS, ARRAY_OF_INFO, ROOT, COMM, INTERCOMM,
// ARRAY_OF_ERRCODES, IERROR)
// INTEGER COUNT, ARRAY_OF_INFO(*), ARRAY_OF_MAXPROCS(*), ROOT, COMM,
// INTERCOMM, ARRAY_OF_ERRCODES(*), IERROR
// CHARACTER*(*) ARRAY_OF_COMMANDS(*), ARRAY_OF_ARGV(COUNT, *)
//
// INCOMPLETE.  Not finished implementing Fortran-to-C conversion
// on arrays of character data (parameters #2 dn #3, array_of_commands
// and array_of_argv[], respectively.
#if ABI64
void
mpi_comm_spawn_multiple_(
fortran_int *count,
char *array_of_commands[],
char *array_of_argv[],
fortran_int array_of_maxprocs[],
fortran_int array_of_info[],
fortran_int *root,
fortran_int *comm,
fortran_int *intercomm,
fortran_int array_of_errcodes[],
fortran_int  *ierr,
int command_len[*count])
{
int i;
int x_array_of_maxprocs[ (int) *count ] ;
MPI_Info x_array_of_info[ (int) *count ] ;
int x_array_of_errcodes[ (int) *count ] ;
MPI_Comm x_intercomm = *intercomm;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Comm_spawn_multiple...\n");
#endif

	for (i=0;i< (int) *count;i++) {
	  x_array_of_maxprocs[i] = array_of_maxprocs[i];
	}
	for (i=0;i< (int) *count;i++) {
	  x_array_of_info[i] = array_of_info[i];
	}
	for (i=0;i< (int) *count;i++) {
	  x_array_of_errcodes[i] = array_of_errcodes[i];
	}

        *ierr = (fortran_int) PMPI_Comm_spawn_multiple( (int) *count,
	array_of_commands,
	&array_of_argv, x_array_of_maxprocs, x_array_of_info, (int) *root,
	(MPI_Comm) *comm, &x_intercomm, x_array_of_errcodes) ;

	for (i=0;i<*count;i++) {
	  array_of_errcodes[i] = (fortran_int) x_array_of_errcodes[i];
	}
        *intercomm = (fortran_int) x_intercomm ;
}
#endif

//
// MPI_OPEN_PORT(info, port_name)
// [ IN info] implementation-specific information on how to establish an
// address (handle)
// [ OUT port_name] newly established port (string)
//
// int MPI_Open_port(MPI_Info info, char *port_name)
//
// MPI_OPEN_PORT(INFO, PORT_NAME, IERROR)
// CHARACTER*(*) PORT_NAME
// INTEGER INFO, IERROR
//
// MPI-2 routine not yet implemented in SGI MPT
//

// 
// MPI_CLOSE_PORT(port_name)
// [ IN port_name] a port (string)
//
// int MPI_Close_port(char *port_name)
//
// MPI_CLOSE_PORT(PORT_NAME, IERROR)
// CHARACTER*(*) PORT_NAME
// INTEGER IERROR
//
// MPI-2 routine not yet implemented in SGI MPT
// 

// 
// MPI_COMM_ACCEPT(port_name, info, root, comm, newcomm)
// [ IN port_name] port name (string, used only on root)
// [ IN info] implementation-dependent information (handle, used only on
// root)
// [ IN root] rank in comm of root node (integer)
// [ IN comm] intracommunicator over which call is collective (handle)
// [ OUT newcomm] intercommunicator with client as remote group (handle)
//
// int MPI_Comm_accept(char *port_name, MPI_Info info, int root, MPI_Comm
// comm, MPI_Comm *newcomm)
//
// MPI_COMM_ACCEPT(PORT_NAME, INFO, ROOT, COMM, NEWCOMM, IERROR)
// CHARACTER*(*) PORT_NAME
// INTEGER INFO, ROOT, COMM, NEWCOMM, IERROR
//
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_COMM_CONNECT(port_name, info, root, comm, newcomm)
// [ IN port_name] network address (string, used only on root)
// [ IN info] implementation-dependent information (handle, used only on
// root)
// [ IN root] rank in comm of root node (integer)
// [ IN comm] intracommunicator over which call is collective (handle)
// [ OUT newcomm] intercommunicator with server as remote group (handle)
//
// int MPI_Comm_connect(char *port_name, MPI_Info info, int root,
// MPI_Comm comm, MPI_Comm *newcomm)
//
// MPI_COMM_CONNECT(PORT_NAME, INFO, ROOT, COMM, NEWCOMM, IERROR)
// CHARACTER*(*) PORT_NAME
// INTEGER INFO, ROOT, COMM, NEWCOMM, IERROR
//
// MPI-2 routine not yet implemented in SGI MPT

//
// MPI_PUBLISH_NAME(service_name, info, port_name)
// [ IN service_name] a service name to associate with the port (string)
// [ IN info] implementation-specific information (handle)
// [ IN port_name] a port name (string)
//
// int MPI_Publish_name(char *service_name, MPI_Info info, char
// *port_name)
//
// MPI_PUBLISH_NAME(SERVICE_NAME, INFO, PORT_NAME, IERROR)
// INTEGER INFO, IERROR
// CHARACTER*(*) SERVICE_NAME, PORT_NAME
// 
// MPI-2 routine not yet implemented in SGI MPT

//
// MPI_UNPUBLISH_NAME(service_name, info, port_name)
// [ IN service_name] a service name (string)
// [ IN info] implementation-specific information (handle)
// [ IN port_name] a port name (string)
//
// int MPI_Unpublish_name(char *service_name, MPI_Info info, char
// *port_name)
//
// MPI_UNPUBLISH_NAME(SERVICE_NAME, INFO, PORT_NAME, IERROR)
// INTEGER INFO, IERROR
// CHARACTER*(*) SERVICE_NAME, PORT_NAME
//
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_LOOKUP_NAME(service_name, info, port_name)
// [ IN service_name] a service name (string)
// [ IN info] implementation-specific information (handle)
// [ OUT port_name] a port name (string)
//
// int MPI_Lookup_name(char *service_name, MPI_Info info, char
// *port_name)
//
// MPI_LOOKUP_NAME(SERVICE_NAME, INFO, PORT_NAME, IERROR)
// CHARACTER*(*) SERVICE_NAME, PORT_NAME
// INTEGER INFO, IERROR
//
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_COMM_DISCONNECT(comm)
// [ INOUT comm] communicator (handle)
//
// int MPI_Comm_disconnect(MPI_Comm *comm)
//
// MPI_COMM_DISCONNECT(COMM, IERROR)
// INTEGER COMM, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_COMM_JOIN(fd, intercomm)
// [ IN fd] socket file descriptor
// [ OUT intercomm] new intercommunicator (handle)
//
// int MPI_Comm_join(int fd, MPI_Comm *intercomm)
//
// MPI_COMM_JOIN(FD, INTERCOMM, IERROR)
// INTEGER FD, INTERCOMM, IERROR
//
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_WIN_CREATE(base, size, disp_unit, info, comm, win)
// [ IN base] initial address of window (choice)
// [ IN size] size of window in bytes (nonnegative integer)
// [ IN disp_unit] local unit size for displacements, in bytes (positive
// integer)
// [ IN info] info argument (handle)
// [ IN comm] communicator (handle)
// [ OUT win] window object returned by the call (handle)
// 
// int MPI_Win_create(void *base, MPI_Aint size, int disp_unit, MPI_Info
// info, MPI_Comm comm, MPI_Win *win)
// 
// MPI_WIN_CREATE(BASE, SIZE, DISP_UNIT, INFO, COMM, WIN, IERROR)
// <type> BASE(*)
// INTEGER(KIND=MPI_ADDRESS_KIND) SIZE
// INTEGER DISP_UNIT, INFO, COMM, WIN, IERROR
// 
#if ABI64
void
mpi_win_create_(
void *base,
fortran_int *size,
fortran_int *disp_unit,
fortran_int *info,
fortran_int *comm,
fortran_int *win,
fortran_int *ierr)
{
MPI_Win x_win;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Win_create\n");
#endif
        *ierr = (fortran_int) PMPI_Win_create( (void *) base,
(MPI_Aint) *size, (int) *disp_unit, (MPI_Info) *info, (MPI_Comm) *comm, &x_win);

        *win = (fortran_int) x_win;
}
#endif

//
// MPI_WIN_FREE(win)
// [ INOUT win] window object (handle)
// 
// int MPI_Win_free(MPI_Win *win)
// 
// MPI_WIN_FREE(WIN, IERROR)
// INTEGER WIN, IERROR
// 
#if ABI64
void
mpi_win_free_(
fortran_int *win,
fortran_int  *ierr)
{
MPI_Win x_win = *win;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Win_free with window = %i\n",(int) *win);
#endif
        *ierr = (fortran_int) PMPI_Win_free( (MPI_Win *) &x_win );

        *win = (fortran_int) x_win;
}
#endif

// 
// MPI_WIN_GET_GROUP(win, group)
// [ IN win] window object (handle)
// [ OUT group] group of processes which share access to the window
// (handle)
// 
// int MPI_Win_get_group(MPI_Win win, MPI_Group *group)
//
// MPI_WIN_GET_GROUP(WIN, GROUP, IERROR)
// INTEGER WIN, GROUP, IERROR
//
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_PUT(origin_addr, origin_count, origin_datatype, target_rank,
// target_disp, target_count, target_datatype, win)
// [ IN origin_addr] initial address of origin buffer (choice)
// [ IN origin_count] number of entries in origin buffer (nonnegative
// integer)
// [ IN origin_datatype] datatype of each entry in origin buffer (handle)
// [ IN target_rank] rank of target (nonnegative integer)
// 
// [ IN target_disp] displacement from start of window to target buffer
// (nonnegative integer)
// 
// [ IN target_count] number of entries in target buffer (nonnegative
// integer)
// [ IN target_datatype] datatype of each entry in target buffer (handle)
// 
// [ IN win] window object used for communication (handle)
// 
// int MPI_Put(void *origin_addr, int origin_count, MPI_Datatype
// origin_datatype, int target_rank, MPI_Aint target_disp, int
// target_count, MPI_Datatype target_datatype, MPI_Win win)
// 
// MPI_PUT(ORIGIN_ADDR, ORIGIN_COUNT, ORIGIN_DATATYPE, TARGET_RANK,
// TARGET_DISP, TARGET_COUNT, TARGET_DATATYPE, WIN, IERROR)
// <type> ORIGIN_ADDR(*)
// INTEGER(KIND=MPI_ADDRESS_KIND) TARGET_DISP
// INTEGER ORIGIN_COUNT, ORIGIN_DATATYPE, TARGET_RANK, TARGET_COUNT,
// TARGET_DATATYPE, WIN, IERROR
// 
#if ABI64
void
mpi_put_(
void *origin_addr,
fortran_int *origin_count,
fortran_int *origin_datatype,
fortran_int *target_rank,
fortran_int *target_disp,
fortran_int *target_count,
fortran_int *target_datatype,
fortran_int *win,
fortran_int *ierr)
{
MPI_Datatype x_origin_datatype = convert_type(*origin_datatype);
MPI_Datatype x_target_datatype = convert_type(*target_datatype);

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Put...\n");
#endif
 	*ierr = (fortran_int) PMPI_Put( (void *) origin_addr,
        (int) *origin_count, (MPI_Datatype) x_origin_datatype,
        (int) *target_rank, (MPI_Aint) target_disp, (int) *target_count,
        x_target_datatype, (MPI_Win) *win);

}
#endif

// 
// MPI_GET(origin_addr, origin_count, origin_datatype, target_rank,
// target_disp, target_count, target_datatype, win)
// [ OUT origin_addr] initial address of origin buffer (choice)
// [ IN origin_count] number of entries in origin buffer (nonnegative
// integer)
// [ IN origin_datatype] datatype of each entry in origin buffer (handle)
// [ IN target_rank] rank of target (nonnegative integer)
// [ IN target_disp] displacement from window start to the beginning of
// the target buffer (nonnegative integer)
// [ IN target_count] number of entries in target buffer (nonnegative
// integer)
// [ IN target_datatype] datatype of each entry in target buffer (handle)
// [ IN win] window object used for communication (handle)
// 
// int MPI_Get(void *origin_addr, int origin_count, MPI_Datatype
// origin_datatype, int target_rank, MPI_Aint target_disp, int
// target_count, MPI_Datatype target_datatype, MPI_Win win)
// 
// MPI_GET(ORIGIN_ADDR, ORIGIN_COUNT, ORIGIN_DATATYPE, TARGET_RANK,
// TARGET_DISP, TARGET_COUNT, TARGET_DATATYPE, WIN, IERROR)
// <type> ORIGIN_ADDR(*)
// INTEGER(KIND=MPI_ADDRESS_KIND) TARGET_DISP
// INTEGER ORIGIN_COUNT, ORIGIN_DATATYPE, TARGET_RANK, TARGET_COUNT,
// TARGET_DATATYPE, WIN, IERROR
// 
#if ABI64
void
mpi_get_(
void *origin_addr,
fortran_int *origin_count,
fortran_int *origin_datatype,
fortran_int *target_rank,
fortran_int *target_disp,
fortran_int *target_count,
fortran_int *target_datatype,
fortran_int *win,
fortran_int *ierr)
{
MPI_Datatype x_origin_datatype = convert_type(*origin_datatype);
MPI_Datatype x_target_datatype = convert_type(*target_datatype);

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Get...\n");
#endif
 	*ierr = (fortran_int) PMPI_Get( (void *) origin_addr,
        (int) *origin_count, (MPI_Datatype) x_origin_datatype,
        (int) *target_rank, (MPI_Aint) target_disp, (int) *target_count,
        x_target_datatype, (MPI_Win) *win);

}
#endif

// 
// MPI_ACCUMULATE(origin_addr, origin_count, origin_datatype,
// target_rank, target_disp, target_count, target_datatype, op, win)
// [ IN origin_addr] initial address of buffer (choice)
// [ IN origin_count] number of entries in buffer (nonnegative integer)
// [ IN origin_datatype] datatype of each buffer entry (handle)
// [ IN target_rank] rank of target (nonnegative integer)
// [ IN target_disp] displacement from start of window to beginning of
// target buffer (nonnegative integer)
// [ IN target_count] number of entries in target buffer (nonnegative
// integer)
// [ IN target_datatype] datatype of each entry in target buffer (handle)
// [ IN op] reduce operation (handle)
// [ IN win] window object (handle)
// 
// int MPI_Accumulate(void *origin_addr, int origin_count, MPI_Datatype
// origin_datatype, int target_rank, MPI_Aint target_disp, int
// target_count, MPI_Datatype target_datatype, MPI_Op op, MPI_Win win)
// 
// MPI_ACCUMULATE(ORIGIN_ADDR, ORIGIN_COUNT, ORIGIN_DATATYPE,
// TARGET_RANK, TARGET_DISP, TARGET_COUNT, TARGET_DATATYPE, OP, WIN,
// IERROR)
// <type> ORIGIN_ADDR(*)
// INTEGER(KIND=MPI_ADDRESS_KIND) TARGET_DISP
// INTEGER ORIGIN_COUNT, ORIGIN_DATATYPE,TARGET_RANK, TARGET_COUNT,
// TARGET_DATATYPE, OP, WIN, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_WIN_FENCE(assert, win)
// [ IN assert] program assertion (integer)
// 
// [ IN win] window object (handle)
// 
// int MPI_Win_fence(int assert, MPI_Win win)
// 
// MPI_WIN_FENCE(ASSERT, WIN, IERROR)
// INTEGER ASSERT, WIN, IERROR
// 
#if ABI64
void
mpi_win_fence_(
fortran_int  *assert,
fortran_int  *win,
fortran_int  *ierr)
{
#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Win_fence\n");
#endif
        *ierr = (fortran_int) PMPI_Win_fence( (int) *assert, (MPI_Win) *win );
}
#endif

// 
// MPI_WIN_START(group, assert, win)
// [ IN group] group of target processes (handle)
// [ IN assert] program assertion (integer)
// [ IN win] window object (handle)
// 
// int MPI_Win_start(MPI_Group group, int assert, MPI_Win win)
// 
// MPI_WIN_START(GROUP, ASSERT, WIN, IERROR)
// INTEGER GROUP, ASSERT, WIN, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_WIN_COMPLETE(win)
// [ IN win] window object (handle)
// 
// int MPI_Win_complete(MPI_Win win)
// 
// MPI_WIN_COMPLETE(WIN, IERROR)
// INTEGER WIN, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_WIN_POST(group, assert, win)
// [ IN group] group of origin processes (handle)
// [ IN assert] program assertion (integer)
// [ IN win] window object (handle)
// 
// int MPI_Win_post(MPI_Group group, int assert, MPI_Win win)
// 
// MPI_WIN_POST(GROUP, ASSERT, WIN, IERROR)
// INTEGER GROUP, ASSERT, WIN, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_WIN_WAIT(win)
// [ IN win] window object (handle)
// 
// int MPI_Win_wait(MPI_Win win)
// 
// MPI_WIN_WAIT(WIN, IERROR)
// INTEGER WIN, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_WIN_TEST(win, flag)
// [ IN win] window object (handle)
// [ OUT flag] success flag (logical)
// 
// int MPI_Win_test(MPI_Win win, int *flag)
// 
// MPI_WIN_TEST(WIN, FLAG, IERROR)
// INTEGER WIN, IERROR
// LOGICAL FLAG
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_WIN_LOCK(lock_type, rank, assert, win)
// [ IN lock_type] either MPI_LOCK_EXCLUSIVE or MPI_LOCK_SHARED (state)
// [ IN rank] rank of locked window (nonnegative integer)
// [ IN assert] program assertion (integer)
// [ IN win] window object (handle)
// 
// int MPI_Win_lock(int lock_type, int rank, int assert, MPI_Win win)
// 
// MPI_WIN_LOCK(LOCK_TYPE, RANK, ASSERT, WIN, IERROR)
// INTEGER LOCK_TYPE, RANK, ASSERT, WIN, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_WIN_UNLOCK(rank, win)
// [ IN rank] rank of window (nonnegative integer)
// [ IN win] window object (handle)
// 
// int MPI_Win_unlock(int rank, MPI_Win win)
// 
// MPI_WIN_UNLOCK(RANK, WIN, IERROR)
// INTEGER RANK, WIN, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_ALLTOALLW(sendbuf, sendcounts, sdispls, sendtypes, recvbuf,
// recvcounts, rdispls, recvtypes, comm)
// [ IN sendbuf] starting address of send buffer (choice)
// [ IN sendcounts] integer array equal to the group size specifying the
// number of elements to send to each processor (integer)
// [ IN sdispls] integer array (of length group size). Entry j specifies
// the displacement in bytes (relative to sendbuf) from which to take the
// outgoing data destined for process j
// [ IN sendtypes] array of datatypes (of length group size). Entry j
// specifies the type of data to send to process j (handle)
// [ OUT recvbuf] address of receive buffer (choice)
// [ IN recvcounts] integer array equal to the group size specifying the
// number of elements that can be received from each processor (integer)
// [ IN rdispls] integer array (of length group size). Entry i specifies
// the displacement in bytes (relative to recvbuf) at which to place the
// incoming data from process i
// [ IN recvtypes] array of datatypes (of length group size). Entry i
// specifies the type of data received from process i (handle)
// [ IN comm] communicator (handle)
// 
// int MPI_Alltoallw(void *sendbuf, int sendcounts[], int sdispls[],
// MPI_Datatype sendtypes[], void *recvbuf, int recvcounts[], int
// rdispls[], MPI_Datatype recvtypes[], MPI_Comm comm)
// 
// MPI_ALLTOALLW(SENDBUF, SENDCOUNTS, SDISPLS, SENDTYPES, RECVBUF,
// RECVCOUNTS, RDISPLS, RECVTYPES, COMM, IERROR)
// <type> SENDBUF(*), RECVBUF(*)
// INTEGER SENDCOUNTS(*), SDISPLS(*), SENDTYPES(*), RECVCOUNTS(*),
// RDISPLS(*), RECVTYPES(*), COMM, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_EXSCAN(sendbuf, recvbuf, count, datatype, op, comm)
// [ IN sendbuf] starting address of send buffer (choice)
// [ OUT recvbuf] starting address of receive buffer (choice)
// [ IN count] number of elements in input buffer (integer)
// [ IN datatype] data type of elements of input buffer (handle)
// [ IN op] operation (handle)
// [ IN comm] intracommunicator (handle)
// 
// int MPI_Exscan(void *sendbuf, void *recvbuf, int count, MPI_Datatype
// datatype, MPI_Op op, MPI_Comm comm)
// 
// MPI_EXSCAN(SENDBUF, RECVBUF, COUNT, DATATYPE, OP, COMM, IERROR)
// <type> SENDBUF(*), RECVBUF(*)
// INTEGER COUNT, DATATYPE, OP, COMM, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_GREQUEST_START(query_fn, free_fn, cancel_fn, extra_state, request)
// [ IN query_fn] callback function invoked when request status is
// queried (function)
// [ IN free_fn] callback function invoked when request is freed
// (function)
// [ IN cancel_fn] callback function invoked when request is cancelled
// (function)
// [ IN extra_state] extra state
// [ OUT request] generalized request (handle)
// 
// int MPI_Grequest_start(MPI_Grequest_query_function *query_fn,
// MPI_Grequest_free_function *free_fn, MPI_Grequest_cancel_function
// *cancel_fn, void *extra_state, MPI_Request *request)
// 
// MPI_GREQUEST_START(QUERY_FN, FREE_FN, CANCEL_FN, EXTRA_STATE, REQUEST,
// IERROR)
// INTEGER REQUEST, IERROR
// EXTERNAL QUERY_FN, FREE_FN, CANCEL_FN
// INTEGER (KIND=MPI_ADDRESS_KIND) EXTRA_STATE
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_GREQUEST_COMPLETE(request)
// [ INOUT request] generalized request (handle)
// 
// int MPI_Grequest_complete(MPI_Request request)
// 
// MPI_GREQUEST_COMPLETE(REQUEST, IERROR)
// INTEGER REQUEST, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_STATUS_SET_ELEMENTS(status, datatype, count)
// [ INOUT status] status to associate count with (Status)
// [ IN datatype] datatype associated with count (handle)
// [ IN count] number of elements to associate with status (integer)
// 
// int MPI_Status_set_elements(MPI_Status *status, MPI_Datatype datatype,
// int count)
// 
// MPI_STATUS_SET_ELEMENTS(STATUS, DATATYPE, COUNT, IERROR)
// INTEGER STATUS(MPI_STATUS_SIZE), DATATYPE, COUNT, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_STATUS_SET_CANCELLED(status, flag)
// [ INOUT status] status to associate cancel flag with (Status)
// [ IN flag] if true indicates request was cancelled (logical)
// 
// int MPI_Status_set_cancelled(MPI_Status *status, int flag)
// 
// MPI_STATUS_SET_CANCELLED(STATUS, FLAG, IERROR)
// INTEGER STATUS(MPI_STATUS_SIZE), IERROR
// LOGICAL FLAG
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_COMM_SET_NAME (comm, comm_name)
// [ INOUT comm] communicator whose identifier is to be set (handle)
// [ IN comm_name] the character string which is remembered as the name
// (string)
// 
// int MPI_Comm_set_name(MPI_Comm comm, char *comm_name)
// 
// MPI_COMM_SET_NAME(COMM, COMM_NAME, IERROR)
// INTEGER COMM, IERROR
// CHARACTER*(*) COMM_NAME
// 
// MPI-2 routine not yet implemented in SGI MPT

//
// MPI_COMM_GET_NAME (comm, comm_name, resultlen)
// [ IN comm] communicator whose name is to be returned (handle)
// [ OUT comm_name] the name previously stored on the communicator, or an
// empty string if no such name exists (string)
// [ OUT resultlen] length of returned name (integer)
// 
// int MPI_Comm_get_name(MPI_Comm comm, char *comm_name, int *resultlen)
// 
// MPI_COMM_GET_NAME(COMM, COMM_NAME, RESULTLEN, IERROR)
// INTEGER COMM, RESULTLEN, IERROR
// CHARACTER*(*) COMM_NAME
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_TYPE_SET_NAME (type, type_name)
// [ INOUT type] datatype whose identifier is to be set (handle)
// [ IN type_name] the character string which is remembered as the name
// (string)
// 
// int MPI_Type_set_name(MPI_Datatype type, char *type_name)
// 
// MPI_TYPE_SET_NAME(TYPE, TYPE_NAME, IERROR)
// INTEGER TYPE, IERROR
// CHARACTER*(*) TYPE_NAME
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_TYPE_GET_NAME (type, type_name, resultlen)
// [ IN type] datatype whose name is to be returned (handle)
// [ OUT type_name] the name previously stored on the datatype, or a
// empty string if no such name exists (string)
// [ OUT resultlen] length of returned name (integer)
// 
// int MPI_Type_get_name(MPI_Datatype type, char *type_name, int
// *resultlen)
// 
// MPI_TYPE_GET_NAME(TYPE, TYPE_NAME, RESULTLEN, IERROR)
// INTEGER TYPE, RESULTLEN, IERROR
// CHARACTER*(*) TYPE_NAME
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_WIN_SET_NAME (win, win_name)
// [ INOUT win] window whose identifier is to be set (handle)
// [ IN win_name] the character string which is remembered as the name
// (string)
// 
// int MPI_Win_set_name(MPI_Win win, char *win_name)
// 
// MPI_WIN_SET_NAME(WIN, WIN_NAME, IERROR)
// INTEGER WIN, IERROR
// CHARACTER*(*) WIN_NAME
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_WIN_GET_NAME (win, win_name, resultlen)
// [ IN win] window whose name is to be returned (handle)
// [ OUT win_name] the name previously stored on the window, or a empty
// string if no such name exists (string)
// [ OUT resultlen] length of returned name (integer)
// 
// int MPI_Win_get_name(MPI_Win win, char *win_name, int *resultlen)
// 
// MPI_WIN_GET_NAME(WIN, WIN_NAME, RESULTLEN, IERROR)
// INTEGER WIN, RESULTLEN, IERROR
// CHARACTER*(*) WIN_NAME
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_ADD_ERROR_CLASS(errorclass)
// [ OUT errorclass] value for the new error class (integer)
// 
// int MPI_Add_error_class(int *errorclass)
// 
// MPI_ADD_ERROR_CLASS(ERRORCLASS, IERROR)
// INTEGER ERRORCLASS, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

//
// MPI_ADD_ERROR_CODE(errorclass, errorcode)
// [ IN errorclass] error class (integer)
// [ OUT errorcode] new error code to associated with errorclass
// (integer)
// 
// int MPI_Add_error_code(int errorclass, int *errorcode)
// 
// MPI_ADD_ERROR_CODE(ERRORCLASS, ERRORCODE, IERROR)
// INTEGER ERRORCLASS, ERRORCODE, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

//
// MPI_ADD_ERROR_STRING(errorcode, string)
// [ IN errorcode] error code or class (integer)
// [ IN string] text corresponding to errorcode (string)
// 
// int MPI_Add_error_string(int errorcode, char *string)
// 
// MPI_ADD_ERROR_STRING(ERRORCODE, STRING, IERROR)
// INTEGER ERRORCODE, IERROR
// CHARACTER*(*) STRING
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_COMM_CALL_ERRHANDLER (comm, errorcode)
// [ IN comm] communicator with error handler (handle)
// [ IN errorcode] error code (integer)
// 
// int MPI_Comm_call_errhandler(MPI_Comm comm, int errorcode)
// 
// MPI_COMM_CALL_ERRHANDLER(COMM, ERRORCODE, IERROR)
// INTEGER COMM, ERRORCODE, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

//
// MPI_WIN_CALL_ERRHANDLER (win, errorcode)
// [ IN win] window with error handler (handle)
// [ IN errorcode] error code (integer)
// 
// int MPI_Win_call_errhandler(MPI_Win win, int errorcode)
// 
// MPI_WIN_CALL_ERRHANDLER(WIN, ERRORCODE, IERROR)
// INTEGER WIN, ERRORCODE, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_FILE_CALL_ERRHANDLER (fh, errorcode)
// [ IN fh] file with error handler (handle)
// [ IN errorcode] error code (integer)
// 
// int MPI_File_call_errhandler(MPI_File fh, int errorcode)
// 
// MPI_FILE_CALL_ERRHANDLER(FH, ERRORCODE, IERROR)
// INTEGER FH, ERRORCODE, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_TYPE_GET_ENVELOPE(datatype, num_integers, num_addresses,
// num_datatypes, combiner)
// [ IN datatype] datatype to access (handle)
// [ OUT num_integers] number of input integers used in the call
// constructing combiner (nonnegative integer)
// [ OUT num_addresses] number of input addresses used in the call
// constructing combiner (nonnegative integer)
// [ OUT num_datatypes] number of input datatypes used in the call
// constructing combiner (nonnegative integer)
// [ OUT combiner] combiner (state)
// 
// int MPI_Type_get_envelope(MPI_Datatype datatype, int *num_integers,
// int *num_addresses, int *num_datatypes, int *combiner)
// 
// MPI_TYPE_GET_ENVELOPE(DATATYPE, NUM_INTEGERS, NUM_ADDRESSES,
// NUM_DATATYPES, COMBINER, IERROR)
// INTEGER DATATYPE, NUM_INTEGERS, NUM_ADDRESSES, NUM_DATATYPES,
// COMBINER, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_TYPE_GET_CONTENTS(datatype, max_integers, max_addresses,
// max_datatypes, array_of_integers, array_of_addresses,
// array_of_datatypes)
// [ IN datatype] datatype to access (handle)
// [ IN max_integers] number of elements in array_of_integers
// (non-negative integer)
// [ IN max_addresses] number of elements in array_of_addresses
// (non-negative integer)
// [ IN max_datatypes] number of elements in array_of_datatypes
// (non-negative integer)
// [ OUT array_of_integers] contains integer arguments used in
// constructing datatype (array of integers)
// [ OUT array_of_addresses] contains address arguments used in
// constructing datatype (array of integers)
// [ OUT array_of_datatypes] contains datatype arguments used in
// constructing datatype (array of handles)
// 
// int MPI_Type_get_contents(MPI_Datatype datatype, int max_integers, int
// max_addresses, int max_datatypes, int array_of_integers[], MPI_Aint
// array_of_addresses[], MPI_Datatype array_of_datatypes[])
// 
// MPI_TYPE_GET_CONTENTS(DATATYPE, MAX_INTEGERS, MAX_ADDRESSES,
// MAX_DATATYPES, ARRAY_OF_INTEGERS, ARRAY_OF_ADDRESSES,
// ARRAY_OF_DATATYPES, IERROR)
// INTEGER DATATYPE, MAX_INTEGERS, MAX_ADDRESSES, MAX_DATATYPES,
// ARRAY_OF_INTEGERS(*), ARRAY_OF_DATATYPES(*), IERROR
// INTEGER(KIND=MPI_ADDRESS_KIND) ARRAY_OF_ADDRESSES(*)
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_INIT_THREAD(required, provided)
// [ IN required] desired level of thread support (integer)
// [ OUT provided] provided level of thread support (integer)
// 
// int MPI_Init_thread(int *argc, char *((*argv)[]), int required, int
// *provided)
// 
// MPI_INIT_THREAD(REQUIRED, PROVIDED, IERROR)
// INTEGER REQUIRED, PROVIDED, IERROR
// 
void
mpi_init_thread_(
fortran_int  *required,
fortran_int  *provided,
fortran_int  *ierr)
{
int x_provided ;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Init_thread...\n");
#endif
        *ierr = (fortran_int) PMPI_Init_thread( (void *)0, (void *)0, (int) *required, &x_provided);

        *provided = (fortran_int) x_provided ;
#ifdef DEBUG
  printf("wrapper.c... output from PMPI_Is_thread_main: *provided = %i...\n",(int) *provided);
#endif
}

// 
// MPI_QUERY_THREAD(provided)
// [ OUT provided] provided level of thread support (integer)
// 
// int MPI_Query_thread(int *provided)
// 
// MPI_QUERY_THREAD(PROVIDED, IERROR)
// INTEGER PROVIDED, IERROR
// 
void
mpi_query_thread_(
fortran_int  *provided,
fortran_int  *ierr)
{
int x_provided ;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Query_thread...\n");
#endif
        *ierr = (fortran_int) PMPI_Query_thread( &x_provided);

        *provided = (fortran_int) x_provided ;
#ifdef DEBUG
  printf("wrapper.c... output from PMPI_Is_thread_main: *provided = %i...\n",(int) *provided);
#endif
}

// 
// MPI_IS_THREAD_MAIN(flag)
// [ OUT flag] true if calling thread is main thread, false otherwise
// (logical)
// 
// int MPI_Is_thread_main(int *flag)
// 
// MPI_IS_THREAD_MAIN(FLAG, IERROR)
// LOGICAL FLAG
// INTEGER IERROR
// 
void
mpi_is_thread_main_(
fortran_int  *flag,
fortran_int  *ierr)
{
int x_flag ;

#ifdef DEBUG
  printf("wrapper.c... calling PMPI_Is_thread_main...\n");
#endif
        *ierr = (fortran_int) PMPI_Is_thread_main( &x_flag);

        *flag = (fortran_int) x_flag ;
#ifdef DEBUG
  printf("wrapper.c... output from PMPI_Is_thread_main: *flag = %i...\n",(int) *flag);
#endif
}

// 
// MPI_COMM_CREATE_KEYVAL(comm_copy_attr_fn, comm_delete_attr_fn,
// comm_keyval, extra_state)
// [ IN comm_copy_attr_fn] copy callback function for comm_keyval
// (function)
// [ IN comm_delete_attr_fn] delete callback function for comm_keyval
// (function)
// [ OUT comm_keyval] key value for future access (integer)
// [ IN extra_state] extra state for callback functions
// 
// int MPI_Comm_create_keyval(MPI_Comm_copy_attr_function
// *comm_copy_attr_fn, MPI_Comm_delete_attr_function
// *comm_delete_attr_fn, int *comm_keyval, void *extra_state)
// 
// MPI_COMM_CREATE_KEYVAL(COMM_COPY_ATTR_FN, COMM_DELETE_ATTR_FN,
// COMM_KEYVAL, EXTRA_STATE, IERROR)
// EXTERNAL COMM_COPY_ATTR_FN, COMM_DELETE_ATTR_FN
// INTEGER COMM_KEYVAL, IERROR
// INTEGER(KIND=MPI_ADDRESS_KIND) EXTRA_STATE
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_COMM_FREE_KEYVAL(comm_keyval)
// [ INOUT comm_keyval] key value (integer)
// 
// int MPI_Comm_free_keyval(int *comm_keyval)
// 
// MPI_COMM_FREE_KEYVAL(COMM_KEYVAL, IERROR)
// INTEGER COMM_KEYVAL, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_COMM_SET_ATTR(comm, comm_keyval, attribute_val)
// [ INOUT comm] communicator from which attribute will be attached
// (handle)
// [ IN comm_keyval] key value (integer)
// [ IN attribute_val] attribute value
// 
// int MPI_Comm_set_attr(MPI_Comm comm, int comm_keyval, void
// *attribute_val)
// 
// MPI_COMM_SET_ATTR(COMM, COMM_KEYVAL, ATTRIBUTE_VAL, IERROR)
// INTEGER COMM, COMM_KEYVAL, IERROR
// INTEGER(KIND=MPI_ADDRESS_KIND) ATTRIBUTE_VAL
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_COMM_GET_ATTR(comm, comm_keyval, attribute_val, flag)
// [ IN comm] communicator to which the attribute is attached (handle)
// [ IN comm_keyval] key value (integer)
// [ OUT attribute_val] attribute value, unless flag = false
// [ OUT flag] false if no attribute is associated with the key (logical)
// 
// int MPI_Comm_get_attr(MPI_Comm comm, int comm_keyval, void
// *attribute_val, int *flag)
// 
// MPI_COMM_GET_ATTR(COMM, COMM_KEYVAL, ATTRIBUTE_VAL, FLAG, IERROR)
// INTEGER COMM, COMM_KEYVAL, IERROR
// INTEGER(KIND=MPI_ADDRESS_KIND) ATTRIBUTE_VAL
// LOGICAL FLAG
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_COMM_DELETE_ATTR(comm, comm_keyval)
// [ INOUT comm] communicator from which the attribute is deleted
// (handle)
// [ IN comm_keyval] key value (integer)
// 
// int MPI_Comm_delete_attr(MPI_Comm comm, int comm_keyval)
// 
// MPI_COMM_DELETE_ATTR(COMM, COMM_KEYVAL, IERROR)
// INTEGER COMM, COMM_KEYVAL, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

//
// MPI_WIN_CREATE_KEYVAL(win_copy_attr_fn, win_delete_attr_fn,
// win_keyval, extra_state)
// [ IN win_copy_attr_fn] copy callback function for win_keyval
// (function)
// [ IN win_delete_attr_fn] delete callback function for win_keyval
// (function)
// [ OUT win_keyval] key value for future access (integer)
// [ IN extra_state] extra state for callback functions
// 
// int MPI_Win_create_keyval(MPI_Win_copy_attr_function
// *win_copy_attr_fn, MPI_Win_delete_attr_function *win_delete_attr_fn,
// int *win_keyval, void *extra_state)
// 
// MPI_WIN_CREATE_KEYVAL(WIN_COPY_ATTR_FN, WIN_DELETE_ATTR_FN,
// WIN_KEYVAL, EXTRA_STATE, IERROR)
// EXTERNAL WIN_COPY_ATTR_FN, WIN_DELETE_ATTR_FN
// INTEGER WIN_KEYVAL, IERROR
// INTEGER(KIND=MPI_ADDRESS_KIND) EXTRA_STATE
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_WIN_FREE_KEYVAL(win_keyval)
// [ INOUT win_keyval] key value (integer)
// 
// int MPI_Win_free_keyval(int *win_keyval)
// 
// MPI_WIN_FREE_KEYVAL(WIN_KEYVAL, IERROR)
// INTEGER WIN_KEYVAL, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_WIN_SET_ATTR(win, win_keyval, attribute_val)
// [ INOUT win] window to which attribute will be attached (handle)
// [ IN win_keyval] key value (integer)
// [ IN attribute_val] attribute value
// 
// int MPI_Win_set_attr(MPI_Win win, int win_keyval, void *attribute_val)
// 
// MPI_WIN_SET_ATTR(WIN, WIN_KEYVAL, ATTRIBUTE_VAL, IERROR)
// INTEGER WIN, WIN_KEYVAL, IERROR
// INTEGER(KIND=MPI_ADDRESS_KIND) ATTRIBUTE_VAL
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_WIN_GET_ATTR(win, win_keyval, attribute_val, flag)
// [ IN win] window to which the attribute is attached (handle)
// [ IN win_keyval] key value (integer)
// [ OUT attribute_val] attribute value, unless flag = false
// [ OUT flag] false if no attribute is associated with the key (logical)
// 
// int MPI_Win_get_attr(MPI_Win win, int win_keyval, void *attribute_val,
// int *flag)
// 
// MPI_WIN_GET_ATTR(WIN, WIN_KEYVAL, ATTRIBUTE_VAL, FLAG, IERROR)
// INTEGER WIN, WIN_KEYVAL, IERROR
// INTEGER(KIND=MPI_ADDRESS_KIND) ATTRIBUTE_VAL
// LOGICAL FLAG
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_WIN_DELETE_ATTR(win, win_keyval)
// [ INOUT win] window from which the attribute is deleted (handle)
// [ IN win_keyval] key value (integer)
// 
// int MPI_Win_delete_attr(MPI_Win win, int win_keyval)
// 
// MPI_WIN_DELETE_ATTR(WIN, WIN_KEYVAL, IERROR)
// INTEGER WIN, WIN_KEYVAL, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_TYPE_CREATE_KEYVAL(type_copy_attr_fn, type_delete_attr_fn,
// type_keyval, extra_state)
// [ IN type_copy_attr_fn] copy callback function for type_keyval
// (function)
// [ IN type_delete_attr_fn] delete callback function for type_keyval
// (function)
// [ OUT type_keyval] key value for future access (integer)
// [ IN extra_state] extra state for callback functions
// 
// int MPI_Type_create_keyval(MPI_Type_copy_attr_function
// *type_copy_attr_fn, MPI_Type_delete_attr_function
// *type_delete_attr_fn, int *type_keyval, void *extra_state)
// 
// MPI_TYPE_CREATE_KEYVAL(TYPE_COPY_ATTR_FN, TYPE_DELETE_ATTR_FN,
// TYPE_KEYVAL, EXTRA_STATE, IERROR)
// EXTERNAL TYPE_COPY_ATTR_FN, TYPE_DELETE_ATTR_FN
// INTEGER TYPE_KEYVAL, IERROR
// INTEGER(KIND=MPI_ADDRESS_KIND) EXTRA_STATE
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_TYPE_FREE_KEYVAL(type_keyval)
// [ INOUT type_keyval] key value (integer)
// 
// int MPI_Type_free_keyval(int *type_keyval)
// 
// MPI_TYPE_FREE_KEYVAL(TYPE_KEYVAL, IERROR)
// INTEGER TYPE_KEYVAL, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_TYPE_SET_ATTR(type, type_keyval, attribute_val)
// [ INOUT type] datatype to which attribute will be attached (handle)
// [ IN type_keyval] key value (integer)
// [ IN attribute_val] attribute value
// 
// int MPI_Type_set_attr(MPI_Datatype type, int type_keyval, void
// *attribute_val)
// 
// MPI_TYPE_SET_ATTR(TYPE, TYPE_KEYVAL, ATTRIBUTE_VAL, IERROR)
// INTEGER TYPE, TYPE_KEYVAL, IERROR
// INTEGER(KIND=MPI_ADDRESS_KIND) ATTRIBUTE_VAL
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_TYPE_GET_ATTR(type, type_keyval, attribute_val, flag)
// [ IN type] datatype to which the attribute is attached (handle)
// [ IN type_keyval] key value (integer)
// [ OUT attribute_val] attribute value, unless flag = false
// [ OUT flag] false if no attribute is associated with the key (logical)
// 
// int MPI_Type_get_attr(MPI_Datatype type, int type_keyval, void
// *attribute_val, int *flag)
// 
// MPI_TYPE_GET_ATTR(TYPE, TYPE_KEYVAL, ATTRIBUTE_VAL, FLAG, IERROR)
// INTEGER TYPE, TYPE_KEYVAL, IERROR
// INTEGER(KIND=MPI_ADDRESS_KIND) ATTRIBUTE_VAL
// LOGICAL FLAG
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_TYPE_DELETE_ATTR(type, type_keyval)
// [ INOUT type] datatype from which the attribute is deleted (handle)
// [ IN type_keyval] key value (integer)
// 
// int MPI_Type_delete_attr(MPI_Datatype type, int type_keyval)
// 
// MPI_TYPE_DELETE_ATTR(TYPE, TYPE_KEYVAL, IERROR)
// INTEGER TYPE, TYPE_KEYVAL, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_TYPE_DUP(type, newtype)
// [ IN type] datatype (handle)
// [ OUT newtype] copy of type (handle)
// 
// int MPI_Type_dup(MPI_Datatype type, MPI_Datatype *newtype)
// 
// MPI_TYPE_DUP(TYPE, NEWTYPE, IERROR)
// INTEGER TYPE, NEWTYPE, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_FILE_OPEN(comm, filename, amode, info, fh)
// [ IN comm] communicator (handle)
// [ IN filename] name of file to open (string)
// [ IN amode] file access mode (integer)
// [ IN info] info object (handle)
// [ OUT fh] new file handle (handle)
// 
// int MPI_File_open(MPI_Comm comm, char *filename, int amode, MPI_Info
// info, MPI_File *fh)
//
// MPI_FILE_OPEN(COMM, FILENAME, AMODE, INFO, FH, IERROR)
// CHARACTER*(*) FILENAME
// INTEGER COMM, AMODE, INFO, FH, IERROR
//
void
mpi_file_open_(
fortran_int *comm,
char *filename,
fortran_int *amode,
fortran_int *info,
fortran_int *fh,
fortran_int *ierr,
fortran_int str_len )
{
MPI_File fh_c;
MPI_Info info_c;
char x_filename[str_len+1];
	memcpy(x_filename, filename, str_len);
	x_filename[str_len] = '\0';

    info_c = PMPI_Info_f2c(*info);

    *ierr = (fortran_int) PMPI_File_open( (MPI_Comm) *comm, x_filename, (int) *amode, info_c, &fh_c);

    *fh = (fortran_int) PMPI_File_c2f(fh_c);
}

// 
// MPI_FILE_CLOSE(fh)
// [ INOUT fh] file handle (handle)
// 
// int MPI_File_close(MPI_File *fh)
//
// MPI_FILE_CLOSE(FH, IERROR)
// INTEGER FH, IERROR
//
void
mpi_file_close_(
fortran_int  *fh,
fortran_int  *ierr)
{
    MPI_File fh_c;

    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_close(&fh_c);
    *fh = (fortran_int) PMPI_File_c2f(fh_c);
}

// 
// MPI_FILE_DELETE(filename, info)
// [ IN filename] name of file to delete (string)
// [ IN info] info object (handle)
// 
// int MPI_File_delete(char *filename, MPI_Info info)
//
// MPI_FILE_DELETE(FILENAME, INFO, IERROR)
// CHARACTER*(*) FILENAME
// INTEGER INFO, IERROR
//
void
mpi_file_delete_(
char *filename,
fortran_int *info,
fortran_int *ierr,
fortran_int str_len)
{
MPI_Info info_c;
char x_filename[str_len+1];
	memcpy(x_filename, filename, str_len);
	x_filename[str_len] = '\0';

    info_c = PMPI_Info_f2c( (MPI_Info) *info);

    *ierr = (fortran_int) PMPI_File_delete(x_filename, info_c);

}

// 
// MPI_FILE_SET_SIZE(fh, size)
// [ INOUT fh] file handle (handle)
// [ IN size] size to truncate or expand file (integer)
// 
// int MPI_File_set_size(MPI_File fh, MPI_Offset size)
//
// MPI_FILE_SET_SIZE(FH, SIZE, IERROR)
// INTEGER FH, IERROR
// INTEGER(KIND=MPI_OFFSET_KIND) SIZE
//
void
mpi_file_set_size_(
fortran_int *fh,
fortran_int *size,
fortran_int *ierr )
{
    MPI_File fh_c;
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_set_size(fh_c, (MPI_Offset) *size);
}

// 
// MPI_FILE_PREALLOCATE(fh, size)
// [ INOUT fh] file handle (handle)
// [ IN size] size to preallocate file (integer)
// 
// int MPI_File_preallocate(MPI_File fh, MPI_Offset size)
//
// MPI_FILE_PREALLOCATE(FH, SIZE, IERROR)
// INTEGER FH, IERROR
// INTEGER(KIND=MPI_OFFSET_KIND) SIZE
//
void
mpi_file_preallocate_(
fortran_int *fh,
fortran_int *size,
fortran_int *ierr )
{
    MPI_File fh_c;
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_preallocate(fh_c, (MPI_Offset) *size);
}

// 
// MPI_FILE_GET_SIZE(fh, size)
// [ IN fh] file handle (handle)
// [ OUT size] size of the file in bytes (integer)
// 
// int MPI_File_get_size(MPI_File fh, MPI_Offset *size)
//
// MPI_FILE_GET_SIZE(FH, SIZE, IERROR)
// INTEGER FH, IERROR
// INTEGER(KIND=MPI_OFFSET_KIND) SIZE
//
void
mpi_file_get_size_(
fortran_int *fh,
fortran_int *size,
fortran_int *ierr )
{
MPI_Offset x_size = *size ;

    MPI_File fh_c;
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_get_size(fh_c, &x_size);

    *size = (fortran_int) x_size ;
}

// 
// MPI_FILE_GET_GROUP(fh, group)
// [ IN fh] file handle (handle)
// [ OUT group] group which opened the file (handle)
// 
// int MPI_File_get_group(MPI_File fh, MPI_Group *group)
//
// MPI_FILE_GET_GROUP(FH, GROUP, IERROR)
// INTEGER FH, GROUP, IERROR
//
void mpi_file_get_group_(
fortran_int *fh,
fortran_int *group,
fortran_int *ierr )
{
MPI_Group x_group;

    MPI_File fh_c;
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_get_group(fh_c, &x_group);

    *group = (fortran_int) x_group ;
}

// 
// MPI_FILE_GET_AMODE(fh, amode)
// [ IN fh] file handle (handle)
// [ OUT amode] file access mode used to open the file (integer)
// 
// int MPI_File_get_amode(MPI_File fh, int *amode)
//
// MPI_FILE_GET_AMODE(FH, AMODE, IERROR)
// INTEGER FH, AMODE, IERROR
//
void mpi_file_get_amode_(
fortran_int *fh,
fortran_int *amode,
fortran_int *ierr )
{
int x_amode;

    MPI_File fh_c;
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_get_amode(fh_c, &x_amode);

    *amode = (fortran_int) x_amode ;
}

//
// MPI_FILE_SET_INFO(fh, info)
// [ INOUT fh] file handle (handle)
// [ IN info] info object (handle)
// 
// int MPI_File_set_info(MPI_File fh, MPI_Info info)
//
// MPI_FILE_SET_INFO(FH, INFO, IERROR)
// INTEGER FH, INFO, IERROR
//
void mpi_file_set_info_(
fortran_int *fh,
fortran_int *info,
fortran_int *ierr )
{
    MPI_File fh_c;
    MPI_Info info_c;
    
    fh_c = PMPI_File_f2c( (int) *fh);
    info_c = PMPI_Info_f2c(*info);

    *ierr = (fortran_int) PMPI_File_set_info(fh_c, info_c);
}

//
// MPI_FILE_GET_INFO(fh, info_used)
// [ IN fh] file handle (handle)
// [ OUT info_used] new info object (handle)
// 
// int MPI_File_get_info(MPI_File fh, MPI_Info *info_used)
//
// MPI_FILE_GET_INFO(FH, INFO_USED, IERROR)
// INTEGER FH, INFO_USED, IERROR
//
void mpi_file_get_info_(
fortran_int *fh,
fortran_int *info_used,
fortran_int *ierr )
{
    MPI_File fh_c;
    MPI_Info info_used_c;
    
    fh_c = PMPI_File_f2c( (int) *fh);

    *ierr = (fortran_int) PMPI_File_get_info(fh_c, &info_used_c);

    *info_used = (fortran_int) PMPI_Info_c2f(info_used_c);
}

// 
// MPI_FILE_SET_VIEW(fh, disp, etype, filetype, datarep, info)
// [ INOUT fh] file handle (handle)
// [ IN disp] displacement (integer)
// [ IN etype] elementary datatype (handle)
// [ IN filetype] filetype (handle)
// [ IN datarep] data representation (string)
// [ IN info] info object (handle)
// 
// int MPI_File_set_view(MPI_File fh, MPI_Offset disp, MPI_Datatype
// etype, MPI_Datatype filetype, char *datarep, MPI_Info info)
//
// MPI_FILE_SET_VIEW(FH, DISP, ETYPE, FILETYPE, DATAREP, INFO, IERROR)
// INTEGER FH, ETYPE, FILETYPE, INFO, IERROR
// CHARACTER*(*) DATAREP
// INTEGER(KIND=MPI_OFFSET_KIND) DISP
//
void
mpi_file_set_view_(
fortran_int *fh,
fortran_int *disp,
fortran_int *etype,
fortran_int *filetype,
char *datarep,
fortran_int *info,
fortran_int *ierr,
fortran_int str_len)
{
MPI_Datatype x_etype = convert_type(*etype);
MPI_Datatype x_filetype = convert_type(*filetype);
    MPI_File fh_c;
    MPI_Info info_c;
char x_datarep[str_len+1];
    memcpy(x_datarep, datarep, str_len);
    x_datarep[str_len] = '\0';
    
    info_c = PMPI_Info_f2c(*info);

    fh_c = PMPI_File_f2c( (int) *fh);
 
    *ierr = (fortran_int) PMPI_File_set_view(fh_c, (MPI_Offset) *disp, (MPI_Datatype) x_etype, (MPI_Datatype) x_filetype, x_datarep, info_c);

    *fh = (fortran_int) PMPI_File_c2f(fh_c);
}

// 
// MPI_FILE_GET_VIEW(fh, disp, etype, filetype, datarep)
// [ IN fh] file handle (handle)
// [ OUT disp] displacement (integer)
// [ OUT etype] elementary datatype (handle)
// [ OUT filetype] filetype (handle)
// [ OUT datarep] data representation (string)
// 
// int MPI_File_get_view(MPI_File fh, MPI_Offset *disp, MPI_Datatype
// *etype, MPI_Datatype *filetype, char *datarep)
//
// MPI_FILE_GET_VIEW(FH, DISP, ETYPE, FILETYPE, DATAREP, IERROR)
// INTEGER FH, ETYPE, FILETYPE, IERROR
// CHARACTER*(*) DATAREP, INTEGER(KIND=MPI_OFFSET_KIND) DISP
//
void
mpi_file_get_view_(
fortran_int *fh,
fortran_int *disp,
fortran_int *etype,
fortran_int *filetype,
char *datarep,
fortran_int *ierr,
fortran_int str_len )
{
MPI_Offset x_disp;

    MPI_File fh_c;
MPI_Datatype x_filetype ;
MPI_Datatype x_etype ;
char x_datarep[str_len+1];

    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_get_view(fh_c, &x_disp, &x_etype, &x_filetype, x_datarep);

    *disp = (fortran_int) x_disp ;
    *etype = (fortran_int) x_etype ;
    *filetype = (fortran_int) x_filetype ;
    memcpy(datarep, x_datarep, str_len);
}

// 
// MPI_FILE_READ_AT(fh, offset, buf, count, datatype, status)
// [ IN fh] file handle (handle)
// [ IN offset] file offset (integer)
// [ OUT buf] initial address of buffer (choice)
// [ IN count] number of elements in buffer (integer)
// [ IN datatype] datatype of each buffer element (handle)
// [ OUT status] status object (Status)
// 
// int MPI_File_read_at(MPI_File fh, MPI_Offset offset, void *buf, int
// count, MPI_Datatype datatype, MPI_Status *status)
//
// MPI_FILE_READ_AT(FH, OFFSET, BUF, COUNT, DATATYPE, STATUS, IERROR)
// <type> BUF(*)
// INTEGER FH, COUNT, DATATYPE, STATUS(MPI_STATUS_SIZE), IERROR
// INTEGER(KIND=MPI_OFFSET_KIND) OFFSET
//
void
mpi_file_read_at_(
fortran_int *fh,
fortran_int *offset,
void *buf,
fortran_int *count,
fortran_int *datatype,
fortran_MPI_Status *status,
fortran_int *ierr )
{
MPI_Status x_status;
    MPI_File fh_c;
MPI_Datatype x_datatype = convert_type(*datatype);
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_read_at(fh_c, (MPI_Offset) *offset,buf, (int) *count, (MPI_Datatype) x_datatype,&x_status);

    copy_status_struct(status, &x_status, MPI_WRAPPER_TO_FORTRAN);
}

// 
// MPI_FILE_READ_AT_ALL(fh, offset, buf, count, datatype, status)
// [ IN fh] file handle (handle)
// [ IN offset] file offset (integer)
// [ OUT buf] initial address of buffer (choice)
// [ IN count] number of elements in buffer (integer)
// [ IN datatype] datatype of each buffer element (handle)
// [ OUT status] status object (Status)
// 
// int MPI_File_read_at_all(MPI_File fh, MPI_Offset offset, void *buf,
// int count, MPI_Datatype datatype, MPI_Status *status)
//
// MPI_FILE_READ_AT_ALL(FH, OFFSET, BUF, COUNT, DATATYPE, STATUS, IERROR)
// <type> BUF(*)
// INTEGER FH, COUNT, DATATYPE, STATUS(MPI_STATUS_SIZE), IERROR
// INTEGER(KIND=MPI_OFFSET_KIND) OFFSET
//
void
mpi_file_read_at_all_(
fortran_int *fh,
fortran_int *offset,
void *buf,
fortran_int *count,
fortran_int *datatype,
fortran_MPI_Status *status,
fortran_int *ierr )
{
MPI_Status x_status;
    MPI_File fh_c;
MPI_Datatype x_datatype = convert_type(*datatype);
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_read_at_all(fh_c, (MPI_Offset) *offset,buf, (int) *count, (MPI_Datatype) x_datatype,&x_status);

    copy_status_struct(status, &x_status, MPI_WRAPPER_TO_FORTRAN);
}

// 
// MPI_FILE_WRITE_AT(fh, offset, buf, count, datatype, status)
// [ INOUT fh] file handle (handle)
// [ IN offset] file offset (integer)
// [ IN buf] initial address of buffer (choice)
// [ IN count] number of elements in buffer (integer)
// [ IN datatype] datatype of each buffer element (handle)
// [ OUT status] status object (Status)
// 
// int MPI_File_write_at(MPI_File fh, MPI_Offset offset, void *buf, int
// count, MPI_Datatype datatype, MPI_Status *status)
//
// MPI_FILE_WRITE_AT(FH, OFFSET, BUF, COUNT, DATATYPE, STATUS, IERROR)
// <type> BUF(*)
// INTEGER FH, COUNT, DATATYPE, STATUS(MPI_STATUS_SIZE), IERROR
// INTEGER(KIND=MPI_OFFSET_KIND) OFFSET
//
void
mpi_file_write_at_(
fortran_int *fh,
fortran_int *offset,
void *buf,
fortran_int *count,
fortran_int *datatype,
fortran_MPI_Status *status,
fortran_int *ierr )
{
MPI_Status x_status;
    MPI_File fh_c;
MPI_Datatype x_datatype = convert_type(*datatype);
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_write_at(fh_c, (MPI_Offset) *offset,buf, (int) *count, (MPI_Datatype) x_datatype,&x_status);

    copy_status_struct(status, &x_status, MPI_WRAPPER_TO_FORTRAN);
}

// 
// MPI_FILE_WRITE_AT_ALL(fh, offset, buf, count, datatype, status)
// [ INOUT fh] file handle (handle)
// [ IN offset] file offset (integer)
// [ IN buf] initial address of buffer (choice)
// [ IN count] number of elements in buffer (integer)
// [ IN datatype] datatype of each buffer element (handle)
// [ OUT status] status object (Status)
// 
// int MPI_File_write_at_all(MPI_File fh, MPI_Offset offset, void *buf,
// int count, MPI_Datatype datatype, MPI_Status *status)
//
// MPI_FILE_WRITE_AT_ALL(FH, OFFSET, BUF, COUNT, DATATYPE, STATUS,
// IERROR)
// <type> BUF(*)
// INTEGER FH, COUNT, DATATYPE, STATUS(MPI_STATUS_SIZE), IERROR
// INTEGER(KIND=MPI_OFFSET_KIND) OFFSET
//
void
mpi_file_write_at_all_(
fortran_int *fh,
fortran_int *offset,
void *buf,
fortran_int *count,
fortran_int *datatype,
fortran_MPI_Status *status,
fortran_int *ierr )
{
MPI_Status x_status;
    MPI_File fh_c;
MPI_Datatype x_datatype = convert_type(*datatype);
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_write_at_all(fh_c, (MPI_Offset) *offset,buf, (int) *count, (MPI_Datatype) x_datatype,&x_status);

    copy_status_struct(status, &x_status, MPI_WRAPPER_TO_FORTRAN);
}

// 
// MPI_FILE_IREAD_AT(fh, offset, buf, count, datatype, request)
// [ IN fh] file handle (handle)
// [ IN offset] file offset (integer)
// [ OUT buf] initial address of buffer (choice)
// [ IN count] number of elements in buffer (integer)
// [ IN datatype] datatype of each buffer element (handle)
// [ OUT request] request object (handle)
// 
// int MPI_File_iread_at(MPI_File fh, MPI_Offset offset, void *buf, int
// count, MPI_Datatype datatype, MPI_Request *request)
//
// MPI_FILE_IREAD_AT(FH, OFFSET, BUF, COUNT, DATATYPE, REQUEST, IERROR)
// <type> BUF(*)
// INTEGER FH, COUNT, DATATYPE, REQUEST, IERROR
// INTEGER(KIND=MPI_OFFSET_KIND) OFFSET
//
void
mpi_file_iread_at_(
fortran_int *fh,
fortran_int *offset,
void *buf,
fortran_int *count,
fortran_int *datatype,
fortran_int *request,
fortran_int *ierr )
{
    MPI_File fh_c;
    MPIO_Request req_c;
MPI_Datatype x_datatype = convert_type(*datatype);
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_iread_at(fh_c, (MPI_Offset) *offset,buf, (int) *count, (MPI_Datatype) x_datatype,&req_c);
    *request = (fortran_int) PMPIO_Request_c2f(req_c);
}

// 
// MPI_FILE_IWRITE_AT(fh, offset, buf, count, datatype, request)
// [ INOUT fh] file handle (handle)
// [ IN offset] file offset (integer)
// [ IN buf] initial address of buffer (choice)
// [ IN count] number of elements in buffer (integer)
// [ IN datatype] datatype of each buffer element (handle)
// [ OUT request] request object (handle)
// 
// int MPI_File_iwrite_at(MPI_File fh, MPI_Offset offset, void *buf, int
// count, MPI_Datatype datatype, MPI_Request *request)
//
// MPI_FILE_IWRITE_AT(FH, OFFSET, BUF, COUNT, DATATYPE, REQUEST, IERROR)
// <type> BUF(*)
// INTEGER FH, COUNT, DATATYPE, REQUEST, IERROR
// INTEGER(KIND=MPI_OFFSET_KIND) OFFSET
//
void
mpi_file_iwrite_at_(
fortran_int *fh,
fortran_int *offset,
void *buf,
fortran_int *count,
fortran_int *datatype,
fortran_int *request,
fortran_int *ierr )
{
    MPI_File fh_c;
    MPIO_Request req_c;
MPI_Datatype x_datatype = convert_type(*datatype);
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_iwrite_at(fh_c, (MPI_Offset) *offset,buf, (int) *count, (MPI_Datatype) x_datatype,&req_c);

    *request = (fortran_int) PMPIO_Request_c2f(req_c);
}

// 
// MPI_FILE_READ(fh, buf, count, datatype, status)
// [ INOUT fh] file handle (handle)
// [ OUT buf] initial address of buffer (choice)
// [ IN count] number of elements in buffer (integer)
// [ IN datatype] datatype of each buffer element (handle)
// [ OUT status] status object (Status)
// 
// int MPI_File_read(MPI_File fh, void *buf, int count, MPI_Datatype
// datatype, MPI_Status *status)
//
// MPI_FILE_READ(FH, BUF, COUNT, DATATYPE, STATUS, IERROR)
// <type> BUF(*)
// INTEGER FH, COUNT, DATATYPE, STATUS(MPI_STATUS_SIZE), IERROR
//
void
mpi_file_read_(
fortran_int *fh,
void *buf,
fortran_int *count,
fortran_int *datatype,
fortran_MPI_Status *status,
fortran_int *ierr )
{
MPI_Status x_status;
    MPI_File fh_c;
MPI_Datatype x_datatype = convert_type(*datatype);
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_read(fh_c,buf, (int) *count, (MPI_Datatype) x_datatype,&x_status);

    copy_status_struct(status, &x_status, MPI_WRAPPER_TO_FORTRAN);
}

//
// MPI_FILE_READ_ALL(fh, buf, count, datatype, status)
// [ INOUT fh] file handle (handle)
// [ OUT buf] initial address of buffer (choice)
// [ IN count] number of elements in buffer (integer)
// [ IN datatype] datatype of each buffer element (handle)
// [ OUT status] status object (Status)
// 
// int MPI_File_read_all(MPI_File fh, void *buf, int count, MPI_Datatype
// datatype, MPI_Status *status)
//
// MPI_FILE_READ_ALL(FH, BUF, COUNT, DATATYPE, STATUS, IERROR)
// <type> BUF(*)
// INTEGER FH, COUNT, DATATYPE, STATUS(MPI_STATUS_SIZE), IERROR
//
void
mpi_file_read_all_(
fortran_int *fh,
void *buf,
fortran_int *count,
fortran_int *datatype,
fortran_MPI_Status *status,
fortran_int *ierr )
{
MPI_Status x_status;
    MPI_File fh_c;
MPI_Datatype x_datatype = convert_type(*datatype);
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_read_all(fh_c,buf, (int) *count, (MPI_Datatype) x_datatype,&x_status);

    copy_status_struct(status, &x_status, MPI_WRAPPER_TO_FORTRAN);
}

// 
// MPI_FILE_WRITE(fh, buf, count, datatype, status)
// [ INOUT fh] file handle (handle)
// [ IN buf] initial address of buffer (choice)
// [ IN count] number of elements in buffer (integer)
// [ IN datatype] datatype of each buffer element (handle)
// [ OUT status] status object (Status)
// 
// int MPI_File_write(MPI_File fh, void *buf, int count, MPI_Datatype
// datatype, MPI_Status *status)
//
// MPI_FILE_WRITE(FH, BUF, COUNT, DATATYPE, STATUS, IERROR)
// <type> BUF(*)
// INTEGER FH, COUNT, DATATYPE, STATUS(MPI_STATUS_SIZE), IERROR
//
void
mpi_file_write_(
fortran_int *fh,
void *buf,
fortran_int *count,
fortran_int *datatype,
fortran_MPI_Status *status,
fortran_int *ierr )
{
MPI_Status x_status;
    MPI_File fh_c;
MPI_Datatype x_datatype = convert_type(*datatype);
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_write(fh_c, buf, (int) *count, (MPI_Datatype) x_datatype,&x_status);

    copy_status_struct(status, &x_status, MPI_WRAPPER_TO_FORTRAN);
}

// 
// MPI_FILE_WRITE_ALL(fh, buf, count, datatype, status)
// [ INOUT fh] file handle (handle)
// [ IN buf] initial address of buffer (choice)
// [ IN count] number of elements in buffer (integer)
// [ IN datatype] datatype of each buffer element (handle)
// [ OUT status] status object (Status)
// 
// int MPI_File_write_all(MPI_File fh, void *buf, int count, MPI_Datatype
// datatype, MPI_Status *status)
//
// MPI_FILE_WRITE_ALL(FH, BUF, COUNT, DATATYPE, STATUS, IERROR)
// <type> BUF(*)
// INTEGER FH, COUNT, DATATYPE, STATUS(MPI_STATUS_SIZE), IERROR
//
void
mpi_file_write_all_(
fortran_int *fh,
void *buf,
fortran_int *count,
fortran_int *datatype,
fortran_MPI_Status *status,
fortran_int *ierr )
{
MPI_Status x_status;
    MPI_File fh_c;
MPI_Datatype x_datatype = convert_type(*datatype);
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_write_all(fh_c,buf, (int) *count, (MPI_Datatype) x_datatype,&x_status);

    copy_status_struct(status, &x_status, MPI_WRAPPER_TO_FORTRAN);
}

// 
// MPI_FILE_IREAD(fh, buf, count, datatype, request)
// [ INOUT fh] file handle (handle)
// [ OUT buf] initial address of buffer (choice)
// [ IN count] number of elements in buffer (integer)
// [ IN datatype] datatype of each buffer element (handle)
// [ OUT request] request object (handle)
// 
// int MPI_File_iread(MPI_File fh, void *buf, int count, MPI_Datatype
// datatype, MPI_Request *request)
//
// MPI_FILE_IREAD(FH, BUF, COUNT, DATATYPE, REQUEST, IERROR)
// <type> BUF(*)
// INTEGER FH, COUNT, DATATYPE, REQUEST, IERROR
//
void
mpi_file_iread_(
fortran_int *fh,
void *buf,
fortran_int *count,
fortran_int *datatype,
fortran_int *request,
fortran_int *ierr )
{
    MPI_File fh_c;
    MPIO_Request req_c;
MPI_Datatype x_datatype = convert_type(*datatype);
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_iread(fh_c,buf, (int) *count, (MPI_Datatype) x_datatype,&req_c);
    *request = (fortran_int) PMPIO_Request_c2f(req_c);
}

//
// MPI_FILE_IWRITE(fh, buf, count, datatype, request)
// [ INOUT fh] file handle (handle)
// [ IN buf] initial address of buffer (choice)
// [ IN count] number of elements in buffer (integer)
// [ IN datatype] datatype of each buffer element (handle)
// [ OUT request] request object (handle)
// 
// int MPI_File_iwrite(MPI_File fh, void *buf, int count, MPI_Datatype
// datatype, MPI_Request *request)
//
// MPI_FILE_IWRITE(FH, BUF, COUNT, DATATYPE, REQUEST, IERROR)
// <type> BUF(*)
// INTEGER FH, COUNT, DATATYPE, REQUEST, IERROR
//
void
mpi_file_iwrite_(
fortran_int *fh,
void *buf,
fortran_int *count,
fortran_int *datatype,
fortran_int *request,
fortran_int *ierr )
{
    MPI_File fh_c;
    MPIO_Request req_c;
MPI_Datatype x_datatype = convert_type(*datatype);
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_iwrite(fh_c,buf, (int) *count, (MPI_Datatype) x_datatype,&req_c);

    *request = (fortran_int) PMPIO_Request_c2f(req_c);
}

// 
// MPI_FILE_SEEK(fh, offset, whence)
// [ INOUT fh] file handle (handle)
// [ IN offset] file offset (integer)
// [ IN whence] update mode (state)
// 
// int MPI_File_seek(MPI_File fh, MPI_Offset offset, int whence)
//
// MPI_FILE_SEEK(FH, OFFSET, WHENCE, IERROR)
// INTEGER FH, WHENCE, IERROR
// INTEGER(KIND=MPI_OFFSET_KIND) OFFSET
//
void
mpi_file_seek_(
fortran_int *fh,
fortran_int *offset,
fortran_int *whence,
fortran_int *ierr )
{
    MPI_File fh_c;
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_seek(fh_c, (MPI_Offset) *offset, (int) *whence);
}

// 
// MPI_FILE_GET_POSITION(fh, offset)
// [ IN fh] file handle (handle)
// [ OUT offset] offset of individual pointer (integer)
// 
// int MPI_File_get_position(MPI_File fh, MPI_Offset *offset)
//
// MPI_FILE_GET_POSITION(FH, OFFSET, IERROR)
// INTEGER FH, IERROR
// INTEGER(KIND=MPI_OFFSET_KIND) OFFSET
//
void
mpi_file_get_position_(
fortran_int *fh,
fortran_int *offset,
fortran_int *ierr )
{
MPI_Offset x_offset;
    MPI_File fh_c;
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_get_position(fh_c, &x_offset);

    *offset = (fortran_int) x_offset ;
}

//
// MPI_FILE_GET_BYTE_OFFSET(fh, offset, disp)
// [ IN fh] file handle (handle)
// [ IN offset] offset (integer)
// [ OUT disp] absolute byte position of offset (integer)
// 
// int MPI_File_get_byte_offset(MPI_File fh, MPI_Offset offset,
// MPI_Offset *disp)
//
// MPI_FILE_GET_BYTE_OFFSET(FH, OFFSET, DISP, IERROR)
// INTEGER FH, IERROR
// INTEGER(KIND=MPI_OFFSET_KIND) OFFSET, DISP
//
void
mpi_file_get_byte_offset_(
fortran_int *fh,
fortran_int *offset,
fortran_int *disp,
fortran_int *ierr )
{
MPI_Offset x_disp;
    MPI_File fh_c;
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_get_byte_offset(fh_c, (MPI_Offset) *offset, &x_disp);

    *disp = (fortran_int) x_disp ;
}

// 
// MPI_FILE_READ_SHARED(fh, buf, count, datatype, status)
// [ INOUT fh] file handle (handle)
// [ OUT buf] initial address of buffer (choice)
// [ IN count] number of elements in buffer (integer)
// [ IN datatype] datatype of each buffer element (handle)
// [ OUT status] status object (Status)
// 
// int MPI_File_read_shared(MPI_File fh, void *buf, int count,
// MPI_Datatype datatype, MPI_Status *status)
//
// MPI_FILE_READ_SHARED(FH, BUF, COUNT, DATATYPE, STATUS, IERROR)
// <type> BUF(*)
// INTEGER FH, COUNT, DATATYPE, STATUS(MPI_STATUS_SIZE), IERROR
//
void
mpi_file_read_shared_(
fortran_int *fh,
void *buf,
fortran_int *count,
fortran_int *datatype,
fortran_MPI_Status *status,
fortran_int *ierr )
{
MPI_Status x_status;
    MPI_File fh_c;
MPI_Datatype x_datatype = convert_type(*datatype);
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_read_shared(fh_c,buf, (int) *count, (MPI_Datatype) x_datatype,&x_status);

    copy_status_struct(status, &x_status, MPI_WRAPPER_TO_FORTRAN);
}

// 
// MPI_FILE_WRITE_SHARED(fh, buf, count, datatype, status)
// [ INOUT fh] file handle (handle)
// [ IN buf] initial address of buffer (choice)
// [ IN count] number of elements in buffer (integer)
// [ IN datatype] datatype of each buffer element (handle)
// [ OUT status] status object (Status)
// 
// int MPI_File_write_shared(MPI_File fh, void *buf, int count,
// MPI_Datatype datatype, MPI_Status *status)
//
// MPI_FILE_WRITE_SHARED(FH, BUF, COUNT, DATATYPE, STATUS, IERROR)
// <type> BUF(*)
// INTEGER FH, COUNT, DATATYPE, STATUS(MPI_STATUS_SIZE), IERROR
//
void
mpi_file_write_shared_(
fortran_int *fh,
void *buf,
fortran_int *count,
fortran_int *datatype,
fortran_MPI_Status *status,
fortran_int *ierr )
{
MPI_Status x_status;
    MPI_File fh_c;
MPI_Datatype x_datatype = convert_type(*datatype);
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_write_shared(fh_c, buf, (int) *count, (MPI_Datatype) x_datatype,&x_status);

    copy_status_struct(status, &x_status, MPI_WRAPPER_TO_FORTRAN);
}

// 
// MPI_FILE_IREAD_SHARED(fh, buf, count, datatype, request)
// [ INOUT fh] file handle (handle)
// [ OUT buf] initial address of buffer (choice)
// [ IN count] number of elements in buffer (integer)
// [ IN datatype] datatype of each buffer element (handle)
// [ OUT request] request object (handle)
// 
// int MPI_File_iread_shared(MPI_File fh, void *buf, int count,
// MPI_Datatype datatype, MPI_Request *request)
//
// MPI_FILE_IREAD_SHARED(FH, BUF, COUNT, DATATYPE, REQUEST, IERROR)
// <type> BUF(*)
// INTEGER FH, COUNT, DATATYPE, REQUEST, IERROR
//
void
mpi_file_iread_shared_(
fortran_int *fh,
void *buf,
fortran_int *count,
fortran_int *datatype,
fortran_int *request,
fortran_int *ierr )
{
    MPI_File fh_c;
    MPIO_Request req_c;
MPI_Datatype x_datatype = convert_type(*datatype);
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_iread_shared(fh_c,buf, (int) *count, (MPI_Datatype) x_datatype,&req_c);
    *request = (fortran_int) PMPIO_Request_c2f(req_c);
}

// 
// MPI_FILE_IWRITE_SHARED(fh, buf, count, datatype, request)
// [ INOUT fh] file handle (handle)
// [ IN buf] initial address of buffer (choice)
// [ IN count] number of elements in buffer (integer)
// [ IN datatype] datatype of each buffer element (handle)
// [ OUT request] request object (handle)
// 
// int MPI_File_iwrite_shared(MPI_File fh, void *buf, int count,
// MPI_Datatype datatype, MPI_Request *request)
//
// MPI_FILE_IWRITE_SHARED(FH, BUF, COUNT, DATATYPE, REQUEST, IERROR)
// <type> BUF(*)
// INTEGER FH, COUNT, DATATYPE, REQUEST, IERROR
//
void
mpi_file_iwrite_shared_(
fortran_int *fh,
void *buf,
fortran_int *count,
fortran_int *datatype,
fortran_int *request,
fortran_int *ierr )
{
    MPI_File fh_c;
    MPIO_Request req_c;
MPI_Datatype x_datatype = convert_type(*datatype);
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_iwrite_shared(fh_c,buf, (int) *count, (MPI_Datatype) x_datatype,&req_c);
    *request = (fortran_int) PMPIO_Request_c2f(req_c);
}

//
// MPI_FILE_READ_ORDERED(fh, buf, count, datatype, status)
// [ INOUT fh] file handle (handle)
// [ OUT buf] initial address of buffer (choice)
// [ IN count] number of elements in buffer (integer)
// [ IN datatype] datatype of each buffer element (handle)
// [ OUT status] status object (Status)
// 
// int MPI_File_read_ordered(MPI_File fh, void *buf, int count,
// MPI_Datatype datatype, MPI_Status *status)
//
// MPI_FILE_READ_ORDERED(FH, BUF, COUNT, DATATYPE, STATUS, IERROR)
// <type> BUF(*)
// INTEGER FH, COUNT, DATATYPE, STATUS(MPI_STATUS_SIZE), IERROR
//
void
mpi_file_read_ordered_(
fortran_int *fh,
void *buf,
fortran_int *count,
fortran_int *datatype,
fortran_MPI_Status *status,
fortran_int *ierr )
{
MPI_Status x_status;
    MPI_File fh_c;
MPI_Datatype x_datatype = convert_type(*datatype);
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_read_ordered(fh_c,buf, (int) *count, (MPI_Datatype) x_datatype,&x_status);

    copy_status_struct(status, &x_status, MPI_WRAPPER_TO_FORTRAN);
}

// 
// MPI_FILE_WRITE_ORDERED(fh, buf, count, datatype, status)
// [ INOUT fh] file handle (handle)
// [ IN buf] initial address of buffer (choice)
// [ IN count] number of elements in buffer (integer)
// [ IN datatype] datatype of each buffer element (handle)
// [ OUT status] status object (Status)
// 
// int MPI_File_write_ordered(MPI_File fh, void *buf, int count,
// MPI_Datatype datatype, MPI_Status *status)
//
// MPI_FILE_WRITE_ORDERED(FH, BUF, COUNT, DATATYPE, STATUS, IERROR)
// <type> BUF(*)
// INTEGER FH, COUNT, DATATYPE, STATUS(MPI_STATUS_SIZE), IERROR
//
void
mpi_file_write_ordered_(
fortran_int *fh,
void *buf,
fortran_int *count,
fortran_int *datatype,
fortran_MPI_Status *status,
fortran_int *ierr )
{
MPI_Status x_status;
    MPI_File fh_c;
MPI_Datatype x_datatype = convert_type(*datatype);
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_write_ordered(fh_c,buf, (int) *count, (MPI_Datatype) x_datatype,&x_status);

    copy_status_struct(status, &x_status, MPI_WRAPPER_TO_FORTRAN);
}

// 
// MPI_FILE_SEEK_SHARED(fh, offset, whence)
// [ INOUT fh] file handle (handle)
// [ IN offset] file offset (integer)
// [ IN whence] update mode (state)
// 
// int MPI_File_seek_shared(MPI_File fh, MPI_Offset offset, int whence)
//
// MPI_FILE_SEEK_SHARED(FH, OFFSET, WHENCE, IERROR)
// INTEGER FH, WHENCE, IERROR
// INTEGER(KIND=MPI_OFFSET_KIND) OFFSET
//
void
mpi_file_seek_shared_(
fortran_int *fh,
fortran_int *offset,
fortran_int *whence,
fortran_int *ierr )
{
    MPI_File fh_c;
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_seek_shared(fh_c, (MPI_Offset) *offset, (int) *whence);
}

// 
// MPI_FILE_GET_POSITION_SHARED(fh, offset)
// [ IN fh] file handle (handle)
// [ OUT offset] offset of shared pointer (integer)
// 
// int MPI_File_get_position_shared(MPI_File fh, MPI_Offset *offset)
//
// MPI_FILE_GET_POSITION_SHARED(FH, OFFSET, IERROR)
// INTEGER FH, IERROR
// INTEGER(KIND=MPI_OFFSET_KIND) OFFSET
//
void
mpi_file_get_position_shared_(
fortran_int *fh,
fortran_int *offset,
fortran_int *ierr )
{
MPI_Offset x_offset;
    MPI_File fh_c;
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_get_position_shared(fh_c, &x_offset);

    *offset = (fortran_int) x_offset ;
}

// 
// MPI_FILE_READ_AT_ALL_BEGIN(fh, offset, buf, count, datatype)
// [ IN fh] file handle (handle)
// [ IN offset] file offset (integer)
// [ OUT buf] initial address of buffer (choice)
// [ IN count] number of elements in buffer (integer)
// [ IN datatype] datatype of each buffer element (handle)
// 
// int MPI_File_read_at_all_begin(MPI_File fh, MPI_Offset offset, void
// *buf, int count, MPI_Datatype datatype)
//
// MPI_FILE_READ_AT_ALL_BEGIN(FH, OFFSET, BUF, COUNT, DATATYPE, IERROR)
// <type> BUF(*)
// INTEGER FH, COUNT, DATATYPE, IERROR
// INTEGER(KIND=MPI_OFFSET_KIND) OFFSET
//
void
mpi_file_read_at_all_begin_(
fortran_int *fh,
fortran_int *offset,
void *buf,
fortran_int *count,
fortran_int *datatype,
fortran_int *ierr )
{
    MPI_File fh_c;
MPI_Datatype x_datatype = convert_type(*datatype);
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_read_at_all_begin(fh_c, (MPI_Offset) *offset,buf, (int) *count, (MPI_Datatype) x_datatype);
}

// 
// MPI_FILE_READ_AT_ALL_END(fh, buf, status)
// [ IN fh] file handle (handle)
// [ OUT buf] initial address of buffer (choice)
// [ OUT status] status object (Status)
// 
// int MPI_File_read_at_all_end(MPI_File fh, void *buf, MPI_Status
// *status)
//
// MPI_FILE_READ_AT_ALL_END(FH, BUF, STATUS, IERROR)
// <type> BUF(*)
// INTEGER FH, STATUS(MPI_STATUS_SIZE), IERROR
//
void
mpi_file_read_at_all_end_(
fortran_int *fh,
void *buf,
fortran_MPI_Status *status,
fortran_int *ierr )
{
MPI_Status x_status;
    MPI_File fh_c;
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_read_at_all_end(fh_c,buf,&x_status);

    copy_status_struct(status, &x_status, MPI_WRAPPER_TO_FORTRAN);
}

// 
// MPI_FILE_WRITE_AT_ALL_BEGIN(fh, offset, buf, count, datatype)
// [ INOUT fh] file handle (handle)
// [ IN offset] file offset (integer)
// [ IN buf] initial address of buffer (choice)
// [ IN count] number of elements in buffer (integer)
// [ IN datatype] datatype of each buffer element (handle)
// 
// int MPI_File_write_at_all_begin(MPI_File fh, MPI_Offset offset, void
// *buf, int count, MPI_Datatype datatype)
//
// MPI_FILE_WRITE_AT_ALL_BEGIN(FH, OFFSET, BUF, COUNT, DATATYPE, IERROR)
// <type> BUF(*)
// INTEGER FH, COUNT, DATATYPE, IERROR
// INTEGER(KIND=MPI_OFFSET_KIND) OFFSET
//
void
mpi_file_write_at_all_begin_(
fortran_int *fh,
fortran_int *offset,
void *buf,
fortran_int *count,
fortran_int *datatype,
fortran_int *ierr )
{
    MPI_File fh_c;
MPI_Datatype x_datatype = convert_type(*datatype);
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_write_at_all_begin(fh_c, (MPI_Offset) *offset,buf, (int) *count, (MPI_Datatype) x_datatype);
}

// 
// MPI_FILE_WRITE_AT_ALL_END(fh, buf, status)
// [ INOUT fh] file handle (handle)
// [ IN buf] initial address of buffer (choice)
// [ OUT status] status object (Status)
// 
// int MPI_File_write_at_all_end(MPI_File fh, void *buf, MPI_Status
// *status)
//
// MPI_FILE_WRITE_AT_ALL_END(FH, BUF, STATUS, IERROR)
// <type> BUF(*)
// INTEGER FH, STATUS(MPI_STATUS_SIZE), IERROR
//
void
mpi_file_write_at_all_end_(
fortran_int *fh,
void *buf,
fortran_MPI_Status *status,
fortran_int *ierr )
{
MPI_Status x_status;
    MPI_File fh_c;
    
    fh_c = PMPI_File_f2c( (int) *fh);

    *ierr = (fortran_int) PMPI_File_write_at_all_end(fh_c,buf,&x_status);

    copy_status_struct(status, &x_status, MPI_WRAPPER_TO_FORTRAN);
}

// 
// MPI_FILE_READ_ALL_BEGIN(fh, buf, count, datatype)
// [ INOUT fh] file handle (handle)
// [ OUT buf] initial address of buffer (choice)
// [ IN count] number of elements in buffer (integer)
// [ IN datatype] datatype of each buffer element (handle)
// 
// int MPI_File_read_all_begin(MPI_File fh, void *buf, int count,
// MPI_Datatype datatype)
//
// MPI_FILE_READ_ALL_BEGIN(FH, BUF, COUNT, DATATYPE, IERROR)
// <type> BUF(*)
// INTEGER FH, COUNT, DATATYPE, IERROR
//
void
mpi_file_read_all_begin_(
fortran_int *fh,
void *buf,
fortran_int *count,
fortran_int *datatype,
fortran_int *ierr )
{
    MPI_File fh_c;
MPI_Datatype x_datatype = convert_type(*datatype);
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_read_all_begin(fh_c,buf, (int) *count, (MPI_Datatype) x_datatype);
}

// 
// MPI_FILE_READ_ALL_END(fh, buf, status)
// [ INOUT fh] file handle (handle)
// [ OUT buf] initial address of buffer (choice)
// [ OUT status] status object (Status)
// 
// int MPI_File_read_all_end(MPI_File fh, void *buf, MPI_Status *status)
//
// MPI_FILE_READ_ALL_END(FH, BUF, STATUS, IERROR)
// <type> BUF(*)
// INTEGER FH, STATUS(MPI_STATUS_SIZE), IERROR
//
void
mpi_file_read_all_end_(
fortran_int *fh,
void *buf,
fortran_MPI_Status *status,
fortran_int *ierr )
{
MPI_Status x_status;
    MPI_File fh_c;
    
    fh_c = PMPI_File_f2c( (int) *fh);

    *ierr = (fortran_int) PMPI_File_read_all_end(fh_c,buf,&x_status);

    copy_status_struct(status, &x_status, MPI_WRAPPER_TO_FORTRAN);
}

// 
// MPI_FILE_WRITE_ALL_BEGIN(fh, buf, count, datatype)
// [ INOUT fh] file handle (handle)
// [ IN buf] initial address of buffer (choice)
// [ IN count] number of elements in buffer (integer)
// [ IN datatype] datatype of each buffer element (handle)
// 
// int MPI_File_write_all_begin(MPI_File fh, void *buf, int count,
// MPI_Datatype datatype)
//
// MPI_FILE_WRITE_ALL_BEGIN(FH, BUF, COUNT, DATATYPE, IERROR)
// <type> BUF(*)
// INTEGER FH, COUNT, DATATYPE, IERROR
//
void
mpi_file_write_all_begin_(
fortran_int *fh,
void *buf,
fortran_int *count,
fortran_int *datatype,
fortran_int *ierr )
{
    MPI_File fh_c;
MPI_Datatype x_datatype = convert_type(*datatype);
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_write_all_begin(fh_c,buf, (int) *count,(MPI_Datatype) x_datatype);
}

// 
// MPI_FILE_WRITE_ALL_END(fh, buf, status)
// [ INOUT fh] file handle (handle)
// [ IN buf] initial address of buffer (choice)
// [ OUT status] status object (Status)
// 
// int MPI_File_write_all_end(MPI_File fh, void *buf, MPI_Status *status)
//
// MPI_FILE_WRITE_ALL_END(FH, BUF, STATUS, IERROR)
// <type> BUF(*)
// INTEGER FH, STATUS(MPI_STATUS_SIZE), IERROR
//
void
mpi_file_write_all_end_(
fortran_int *fh,
void *buf,
fortran_MPI_Status *status,
fortran_int *ierr )
{
MPI_Status x_status;
    MPI_File fh_c;
    
    fh_c = PMPI_File_f2c( (int) *fh);

    *ierr = (fortran_int) PMPI_File_write_all_end(fh_c,buf,&x_status);

    copy_status_struct(status, &x_status, MPI_WRAPPER_TO_FORTRAN);
}

// 
// MPI_FILE_READ_ORDERED_BEGIN(fh, buf, count, datatype)
// [ INOUT fh] file handle (handle)
// [ OUT buf] initial address of buffer (choice)
// [ IN count] number of elements in buffer (integer)
// [ IN datatype] datatype of each buffer element (handle)
// 
// int MPI_File_read_ordered_begin(MPI_File fh, void *buf, int count,
// MPI_Datatype datatype)
//
// MPI_FILE_READ_ORDERED_BEGIN(FH, BUF, COUNT, DATATYPE, IERROR)
// <type> BUF(*)
// INTEGER FH, COUNT, DATATYPE, IERROR
//
void
mpi_file_read_ordered_begin_(
fortran_int *fh,
void *buf,
fortran_int *count,
fortran_int *datatype,
fortran_int *ierr )
{
    MPI_File fh_c;
MPI_Datatype x_datatype = convert_type(*datatype);
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_read_ordered_begin(fh_c,buf, (int) *count, (MPI_Datatype) x_datatype);
}

// 
// MPI_FILE_READ_ORDERED_END(fh, buf, status)
// [ INOUT fh] file handle (handle)
// [ OUT buf] initial address of buffer (choice)
// [ OUT status] status object (Status)
// 
// int MPI_File_read_ordered_end(MPI_File fh, void *buf, MPI_Status
// *status)
//
// MPI_FILE_READ_ORDERED_END(FH, BUF, STATUS, IERROR)
// <type> BUF(*)
// INTEGER FH, STATUS(MPI_STATUS_SIZE), IERROR
//
void
mpi_file_read_ordered_end_(
fortran_int *fh,
void *buf,
fortran_MPI_Status *status,
fortran_int *ierr )
{
MPI_Status x_status;
    MPI_File fh_c;
    
    fh_c = PMPI_File_f2c( (int) *fh);

    *ierr = (fortran_int) PMPI_File_read_ordered_end(fh_c,buf,&x_status);

    copy_status_struct(status, &x_status, MPI_WRAPPER_TO_FORTRAN);
}

// 
// MPI_FILE_WRITE_ORDERED_BEGIN(fh, buf, count, datatype)
// [ INOUT fh] file handle (handle)
// [ IN buf] initial address of buffer (choice)
// [ IN count] number of elements in buffer (integer)
// [ IN datatype] datatype of each buffer element (handle)
// 
// int MPI_File_write_ordered_begin(MPI_File fh, void *buf, int count,
// MPI_Datatype datatype)
// MPI_FILE_WRITE_ORDERED_BEGIN(FH, BUF, COUNT, DATATYPE, IERROR)
//
// <type> BUF(*)
// INTEGER FH, COUNT, DATATYPE, IERROR
//
void
mpi_file_write_ordered_begin_(
fortran_int *fh,
void *buf,
fortran_int *count,
fortran_int *datatype,
fortran_int *ierr )
{
    MPI_File fh_c;
MPI_Datatype x_datatype = convert_type(*datatype);
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_write_ordered_begin(fh_c,buf, (int) *count, (MPI_Datatype) x_datatype);
}

// 
// MPI_FILE_WRITE_ORDERED_END(fh, buf, status)
// [ INOUT fh] file handle (handle)
// [ IN buf] initial address of buffer (choice)
// [ OUT status] status object (Status)
// 
// int MPI_File_write_ordered_end(MPI_File fh, void *buf, MPI_Status
// *status)
//
// MPI_FILE_WRITE_ORDERED_END(FH, BUF, STATUS, IERROR)
// <type> BUF(*)
// INTEGER FH, STATUS(MPI_STATUS_SIZE), IERROR
//
void
mpi_file_write_ordered_end_(
fortran_int *fh,
void *buf,
fortran_MPI_Status *status,
fortran_int *ierr )
{
MPI_Status x_status;
    MPI_File fh_c;
    
    fh_c = PMPI_File_f2c( (int) *fh);

    *ierr = (fortran_int) PMPI_File_write_ordered_end(fh_c,buf,&x_status);

    copy_status_struct(status, &x_status, MPI_WRAPPER_TO_FORTRAN);
}

// 
// MPI_FILE_GET_TYPE_EXTENT(fh, datatype, extent)
// [ IN fh] file handle (handle)
// [ IN datatype] datatype (handle)
// [ OUT extent] datatype extent (integer)
// 
// int MPI_File_get_type_extent(MPI_File fh, MPI_Datatype datatype,
// MPI_Aint *extent)
//
// MPI_FILE_GET_TYPE_EXTENT(FH, DATATYPE, EXTENT, IERROR)
// INTEGER FH, DATATYPE, IERROR
// INTEGER(KIND=MPI_ADDRESS_KIND) EXTENT
//
void
mpi_file_get_type_extent_(
fortran_int *fh,
fortran_int *datatype,
fortran_int *extent,
fortran_int *ierr )
{
    MPI_File fh_c;
    MPI_Aint extent_c;
MPI_Datatype x_datatype = convert_type(*datatype);
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_get_type_extent(fh_c, (MPI_Datatype) x_datatype, &extent_c);
    *extent = (fortran_int) extent_c;
}

// 
// MPI_REGISTER_DATAREP(datarep, read_conversion_fn, write_conversion_fn,
// dtype_file_extent_fn, extra_state)
// [ IN datarep] data representation identifier (string)
// [ IN read_conversion_fn] function invoked to convert from file
// representation to native representation (function)
// [ IN write_conversion_fn] function invoked to convert from native
// representation to file representation (function)
// [ IN dtype_file_extent_fn] function invoked to get the extent of a
// datatype as represented in the file (function)
// [ IN extra_state] extra state
// 
// int MPI_Register_datarep(char *datarep,
// MPI_Datarep_conversion_function *read_conversion_fn,
// MPI_Datarep_conversion_function *write_conversion_fn,
// MPI_Datarep_extent_function *dtype_file_extent_fn, void *extra_state)
//
// MPI_REGISTER_DATAREP(DATAREP, READ_CONVERSION_FN, WRITE_CONVERSION_FN,
// DTYPE_FILE_EXTENT_FN, EXTRA_STATE, IERROR)
// CHARACTER*(*) DATAREP
// EXTERNAL READ_CONVERSION_FN, WRITE_CONVERSION_FN, DTYPE_FILE_EXTENT_FN
// INTEGER(KIND=MPI_ADDRESS_KIND) EXTRA_STATE
// INTEGER IERROR
//
// MPI-2 routine not yet implemented in SGI MPT

// 
// typedef int MPI_Datarep_extent_function(MPI_Datatype datatype,
// MPI_Aint *file_extent, void *extra_state);
//
// SUBROUTINE DATAREP_EXTENT_FUNCTION(DATATYPE, EXTENT, EXTRA_STATE,
// IERROR)
// INTEGER DATATYPE, IERROR
// INTEGER(KIND=MPI_ADDRESS_KIND) EXTENT, EXTRA_STATE
//
// MPI-2 routine not yet implemented in SGI MPT

// 
// typedef int MPI_Datarep_conversion_function(void *userbuf,
// MPI_Datatype datatype, int count, void *filebuf, MPI_Offset position,
// void *extra_state);
//
// SUBROUTINE DATAREP_CONVERSION_FUNCTION(USERBUF, DATATYPE, COUNT,
// FILEBUF, POSITION, EXTRA_STATE, IERROR)
// <TYPE> USERBUF(*), FILEBUF(*)
// INTEGER COUNT, DATATYPE, IERROR
// INTEGER(KIND=MPI_OFFSET_KIND) POSITION
// INTEGER(KIND=MPI_ADDRESS_KIND) EXTRA_STATE
//
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_FILE_SET_ATOMICITY(fh, flag)
// [ INOUT fh] file handle (handle)
// [ IN flag] true to set atomic mode, false to set nonatomic mode
// (logical)
// 
// int MPI_File_set_atomicity(MPI_File fh, int flag)
//
// MPI_FILE_SET_ATOMICITY(FH, FLAG, IERROR)
// INTEGER FH, IERROR
// LOGICAL FLAG
//
void
mpi_file_set_atomicity_(
fortran_int *fh,
fortran_int *flag,
fortran_int *ierr )
{
    MPI_File fh_c;
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_set_atomicity(fh_c, (int) *flag);
}

//
// MPI_FILE_GET_ATOMICITY(fh, flag)
// [ IN fh] file handle (handle)
// [ OUT flag] true if atomic mode, false if nonatomic mode (logical)
// 
// int MPI_File_get_atomicity(MPI_File fh, int *flag)
//
// MPI_FILE_GET_ATOMICITY(FH, FLAG, IERROR)
// INTEGER FH, IERROR
// LOGICAL FLAG
//
void
mpi_file_get_atomicity_(
fortran_int *fh,
int *flag,
fortran_int *ierr )
{
    MPI_File fh_c;
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_get_atomicity(fh_c, flag);
}

// 
// MPI_FILE_SYNC(fh)
// [ INOUT fh] file handle (handle)
// 
// int MPI_File_sync(MPI_File fh)
//
// MPI_FILE_SYNC(FH, IERROR)
// INTEGER FH, IERROR
//
void
mpi_file_sync_(
fortran_int *fh,
fortran_int *ierr )
{
    MPI_File fh_c;
    
    fh_c = PMPI_File_f2c( (int) *fh);
    *ierr = (fortran_int) PMPI_File_sync(fh_c);
}

//
// MPI_TYPE_CREATE_F90_REAL(p, r, newtype)
// [ IN p] precision, in decimal digits (integer)
// [ IN r] decimal exponent range (integer)
// [ OUT newtype] the requested MPI datatype (handle)
//
// int MPI_Type_create_f90_real(int p, int r, MPI_Datatype *newtype)
//
// MPI_TYPE_CREATE_F90_REAL(P, R, NEWTYPE, IERROR)
// INTEGER P, R, NEWTYPE, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_TYPE_CREATE_F90_COMPLEX(p, r, newtype)
// [ IN p] precision, in decimal digits (integer)
// [ IN r] decimal exponent range (integer)
// [ OUT newtype] the requested MPI datatype (handle)
// 
// int MPI_Type_create_f90_complex(int p, int r, MPI_Datatype *newtype)
//
// MPI_TYPE_CREATE_F90_COMPLEX(P, R, NEWTYPE, IERROR)
// INTEGER P, R, NEWTYPE, IERROR
//
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_TYPE_CREATE_F90_INTEGER(r, newtype)
// [ IN r] decimal exponent range, i.e., number of decimal digits
// (integer)
// [ OUT newtype] the requested MPI datatype (handle)
//
// int MPI_Type_create_f90_integer(int r, MPI_Datatype *newtype)
//
// MPI_TYPE_CREATE_F90_INTEGER(R, NEWTYPE, IERROR)
// INTEGER R, NEWTYPE, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_SIZEOF(x, size)
// [ IN x] a Fortran variable of numeric intrinsic type (choice)
// [ OUT size] size of machine representation of that type (integer)
// 
// MPI_SIZEOF(X, SIZE, IERROR)
// <type> X
// INTEGER SIZE, IERROR
// 
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_TYPE_MATCH_SIZE(typeclass, size, type)
// [ IN typeclass] generic type specifier (integer)
// [ IN size] size, in bytes, of representation (integer)
// [ OUT type] datatype with correct type, size (handle)
// 
// int MPI_Type_match_size(int typeclass, int size, MPI_Datatype *type)
//
// MPI_TYPE_MATCH_SIZE(TYPECLASS, SIZE, TYPE, IERROR)
// INTEGER TYPECLASS, SIZE, TYPE, IERROR
//
// MPI-2 routine not yet implemented in SGI MPT

// 
// MPI_PCONTROL(level, ...)
// [ IN level] Profiling level
//
// int MPI_Pcontrol(const int level, ...)
//
// MPI_PCONTROL(level)
// INTEGER LEVEL, ...
// 
void
mpi_pcontrol_(
fortran_int *level,
fortran_int *ierr)
{
        *ierr = (fortran_int) PMPI_Pcontrol( (int) *level);
}

// 
// MPI_TEST(request, flag, status)
// [ INOUT request] communication request (handle)
// [ OUT flag] true if operation completed (logical)
// [ OUT status] status object (Status)
//
// int MPI_Test(MPI_Request *request, int *flag, MPI_Status *status)
//
// MPI_TEST(REQUEST, FLAG, STATUS, IERROR)
// LOGICAL FLAG
// INTEGER REQUEST, STATUS(MPI_STATUS_SIZE), IERROR
// 
void mpio_test_(
fortran_int *request,
fortran_int *flag,
fortran_MPI_Status  *status,
fortran_int *ierr )
{
MPI_Status x_status;
int x_flag;
    MPIO_Request req_c;
    
    req_c = PMPIO_Request_f2c( (int) *request);
    *ierr = (fortran_int) PMPIO_Test(&req_c, &x_flag, &x_status);
    *request = (fortran_int) PMPIO_Request_c2f(req_c);

    *flag = (fortran_int) x_flag ;
    copy_status_struct(status, &x_status, MPI_WRAPPER_TO_FORTRAN);
}

// 
// MPIO_WAIT(request, status)
// [ INOUT request] request (handle)
// [ OUT status] status object (Status)
//
// int MPIO_Wait(MPI_Request *request, MPI_Status *status)
//
// MPIO_WAIT(REQUEST, STATUS, IERROR)
// INTEGER REQUEST, STATUS(MPI_STATUS_SIZE), IERROR
// 
void mpio_wait_(
fortran_int *request,
fortran_MPI_Status  *status,
fortran_int *ierr )
{
MPI_Status x_status;
    MPIO_Request req_c;
    
    req_c = PMPIO_Request_f2c( (int) *request);
    *ierr = (fortran_int) PMPIO_Wait(&req_c, &x_status);
    *request = (fortran_int) (fortran_int) PMPIO_Request_c2f(req_c);

    copy_status_struct(status, &x_status, MPI_WRAPPER_TO_FORTRAN);
}

// 
// MPI_WTICK()
//
// double MPI_Wtick(void)
//
// DOUBLE PRECISION MPI_WTICK()
// 
fortran_double
mpi_wtick_()
{
        return (fortran_double) PMPI_Wtick();
}

// 
// MPI_WTIME()
//
// double MPI_Wtime(void)
//
// DOUBLE PRECISION MPI_WTIME()
// 
fortran_double
mpi_wtime_()
{
        return (fortran_double) PMPI_Wtime();
}

//
// NOTES
//
// Missing entry points as displayed by the following script:
// -------------------------------
// #!/bin/sh
// 
// f=all_mpi_entry_points
// echo listing entry points in $f  ...
// nm /usr/lib64/libmpi.so \
//         | cut -f8 -d'|' \
//         | grep _\$ \
//         | grep -v pmpi \
//                 > $f
// 
// echo
// echo searching wrapper.o ...
// for i in `cat $f` ; do
//   nm wrapper.o | grep $i\$ > /dev/null
//   if [ $? -ne 0 ] ; then
//           echo $i : not found
//   fi
// done
// -------------------------------
//
// The above script identified a few missing entry points that
// are SGI extensions (defined in: lib/libmpi/src/adi/printstat.c).
// These include:
//      MPI_SGI_stat_print
//      MPI_SGI_stat_resetall
//      MPI_SGI_stat_reset
//      MPI_SGI_stat_support
//      MPI_SGI_stat_get
// These extentions will be omitted from the wrapper library.
// 
// However, there were a few standard MPI functions that have not
// been implemented yet.  These include:
// 
//          mpi_null_copy_fn_
//          mpi_null_delete_fn_
//          mpi_dup_fn_

#ifdef __linux
#pragma weak mpi_null_copy_fn__ = pmpi_null_copy_fn_
#pragma weak mpi_null_delete_fn__ = pmpi_null_delete_fn_
#pragma weak mpi_dup_fn__ = pmpi_dup_fn_
#endif

//
// MPI_NULL_COPY_FN
//
// INCOMPLETE mpi_null_copy_fn_
//
UNSUPPORTED(mpi_null_copy_fn_)
//

//
// MPI_NULL_DELETE_FN
//
// INCOMPLETE mpi_null_delete_fn_
//
UNSUPPORTED(mpi_null_delete_fn_)
//

//
// MPI_DUP_FN
//
// INCOMPLETE mpi_dup_fn_
//
UNSUPPORTED(mpi_dup_fn_)
//
